/* 2014-08-28: File changed by Sony Corporation */
/*
 *Copyright 2012 Sony Corporation
 *This program is free software; you can redistribute it and/or modify
 *it under the terms of the GNU General Public License, version 2, as
 *published by the Free Software Foundation.
 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *GNU General Public License for more details.
 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <linux/types.h>
#include <linux/list.h>
#include <linux/module.h>

#include <asm/string.h>
#include <asm/atomic.h>

#include "expcd_debug.h"
#include "expcd_drv.h"
#include "expcd_local.h"

//==============================================================================
// local variables
struct list_head oc_callback_list = LIST_HEAD_INIT(oc_callback_list);
struct list_head hp_callback_list = LIST_HEAD_INIT(hp_callback_list);

struct expcd_callback_node {
	struct list_head node;
	void * cb_func;
	void * data;
};

//==============================================================================
// local functions

/* get ExpressCard status of 2 slots */
static int expcdCardStatusGetMain(EXPCDCARDSTATUS *pStatus)
{
	/* param should be asserted before calling this function */

	typedef struct EXPCDCARDSTATUSINFO_t {
		unsigned int pciex;
		unsigned int usb;
		unsigned int ovc;
		unsigned int shdn;
	} EXPCDCARDSTATUSINFO;

	static const EXPCDCARDSTATUSINFO expcdCardStatusInfoTbl[]={
		{	EXPCD_GPIO_SLOT0_CPPE_BIT,
			EXPCD_GPIO_SLOT0_CPUSB_BIT,
			EXPCD_GPIO_SLOT0_OC_BIT,
			EXPCD_GPIO_SLOT0_SHDN_BIT
		},
		{	EXPCD_GPIO_SLOT1_CPPE_BIT,
			EXPCD_GPIO_SLOT1_CPUSB_BIT,
			EXPCD_GPIO_SLOT1_OC_BIT,
			EXPCD_GPIO_SLOT1_SHDN_BIT
		}
	} ;

	unsigned int reg_gpio = 0;
	int cnt;

	expcd_slot_get_status(&reg_gpio);

	expcd_debugl("expcd status = 0x%08x", reg_gpio);

	memset(pStatus, 0, sizeof(EXPCDCARDSTATUS)*EXPCD_SLOT_MAX ) ;

	/* Since GPIO is Low-Level-Assert, reverse all bits first is
	 * a tricky way to do 'and' operation later. */
	reg_gpio = ~reg_gpio;

	for(cnt=0; cnt < EXPCD_SLOT_MAX; cnt++)
	{
		/* slot number */
		pStatus[cnt].slotno = cnt;

		/* card_status */
		if (reg_gpio & expcdCardStatusInfoTbl[cnt].usb) {
			/* as long as EXPCD_GPIO_SLOT0_CPUSB_BIT is Low-Level,
			 * we treat the ExpressCard as a USB device. */
			pStatus[cnt].card_status = EXPCD_USB;
		} else
		if (reg_gpio & expcdCardStatusInfoTbl[cnt].pciex) {
			pStatus[cnt].card_status = EXPCD_PCIEX;
		} else {
			pStatus[cnt].card_status = EXPCD_NON;
		}
		
		/* pow_status */
		if(EXPCD_NON == pStatus[cnt].card_status) {
			/* we treat the slot as power-off when no card in. */
			pStatus[cnt].pow_status = EXPCD_POWOFF;
		} else
		if (reg_gpio & expcdCardStatusInfoTbl[cnt].ovc) {

			//XXX might be another conditions.

			/* over current situation */
			pStatus[cnt].pow_status = EXPCD_POWOCOFF;
		} else {
			/* normal power status */
			if(reg_gpio & expcdCardStatusInfoTbl[cnt].shdn) {
				pStatus[cnt].pow_status = EXPCD_POWOFF;
			} else {
				pStatus[cnt].pow_status = EXPCD_POWON;
			}
		}
	}

	return 0;
}

/* return 1 if ExpressCard is OverCurrent. */
static int expcdIsOverCurrent(int slotno)
{
	EXPCDCARDSTATUS status[EXPCD_SLOT_MAX];
	expcdCardStatusGetMain(status);
	if (EXPCD_POWOCOFF == status[slotno].pow_status)
		return 1;
	return 0;
}

/* run callback functions in the given list. */
static void expcd_callback(struct list_head * func_list) {
	EXPCDCARDSTATUS status[EXPCD_SLOT_MAX];
	pfn_expcd_callback pfn;
	struct expcd_callback_node *cb_node;
	struct expcd_callback_node *tmp_node;

	expcdCardStatusGetMain(status);

	list_for_each_entry_safe(cb_node, tmp_node, func_list, node) {
		if (cb_node->cb_func) {
			pfn = (pfn_expcd_callback)cb_node->cb_func;
			pfn(status);
		}
	}
}


//==============================================================================
// external functions

/*
 * run callback functions of OverCurrent.
*/
void expcd_oc_callback(void) {
	expcd_callback(&oc_callback_list);
}

/*
 * run callback functions of HotPlug.
*/
void expcd_hp_callback(void) {
	expcd_callback(&hp_callback_list);
}

/*
 * Get ExpressCard status.
 * The caller should make sure the param pointer points to a valid mem, that
 * has at least EXPCD_SLOT_MAX elements.
*/
int expcdCardStatusGet(EXPCDCARDSTATUS * pStatus)
{
	if( NULL == pStatus ) {
		return -1;
	}
	return expcdCardStatusGetMain(pStatus);
}

/*
 * Set slot power state.
 * This function does effect only when there is already a card available in
 * the related slot.
 * This is a synchronize function that it may take long time to wait until the
 * slot is attached -- including probing block device and mounting file system.
 *
 * NOTE: A userspace app need not do power control manually when a hotplug
 *       event occurs. The kernel handles hotplug event automatically.
*/
int expcdCardPowCtrl(int slotno, int pow)
{
	EXPCDCARDSTATUS status[EXPCD_SLOT_MAX];
	int result = pow;
	int is_powon;
	int is_pcie;

	if(0 != expcdCardStatusGetMain(&status[0])) {
		result = EXPCD_POWOFF_FAIL;
		return result;
	}

	if(EXPCD_NON == status[slotno].card_status) {
		result = EXPCD_POWREQ_ERR;
		return result;
	}

	if((pow == status[slotno].pow_status)
		|| (EXPCD_POWOCOFF == status[slotno].pow_status))
	{
		result = EXPCD_POWREJ;
		return result;
	}

	is_powon = (pow==EXPCD_POWON);
	is_pcie = ((EXPCD_PCIEX == status[slotno].card_status)
			|| (EXPCD_PCIEX_USB == status[slotno].card_status));
	expcd_power_ctrl(slotno, is_powon, is_pcie);

	return result;
}

/*
 * register a HotPlug callback.
*/
int expcdCardStatusChgRegister(void * pFunc)
{
	struct expcd_callback_node *cb_node;
	struct expcd_callback_node *tmp_node;
	list_for_each_entry_safe(cb_node, tmp_node, &hp_callback_list, node) {
		if (cb_node->cb_func == pFunc) {
			expcd_errorl("the func is already registered.");
			return -2;
		}
	}

	/* do not forget to release the memory when delete the node */
	cb_node = kcalloc(1, sizeof(struct expcd_callback_node), GFP_KERNEL);
	if (!cb_node) {
		expcd_errorl("failed to kcalloc struct expcd_callback_node.");
		return -1;
	}
	cb_node->cb_func = pFunc;
	cb_node->data = NULL;
	list_add(&cb_node->node, &hp_callback_list);
	return 0;
}

/*
 * unregister a HotPlug callback.
*/
int expcdCardStatusChgUnRegister(void *pFunc)
{
	struct expcd_callback_node *cb_node;
	struct expcd_callback_node *tmp_node;
	list_for_each_entry_safe(cb_node, tmp_node, &hp_callback_list, node) {
		if (cb_node->cb_func == pFunc) {
			list_del(&cb_node->node);
			kfree(cb_node);
		}
	}
	return 0;
}

/*
 * register a OverCurrent callback.
*/
int expcdOverCurrentRegister(void *pFunc)
{
	struct expcd_callback_node *cb_node;
	struct expcd_callback_node *tmp_node;
	list_for_each_entry_safe(cb_node, tmp_node, &oc_callback_list, node) {
		if (cb_node->cb_func == pFunc) {
			expcd_errorl("the func is already registered.");
			return -2;
		}
	}

	/* do not forget to release the memory when delete the node */
	cb_node = kcalloc(1, sizeof(struct expcd_callback_node), GFP_KERNEL);
	if (!cb_node) {
		expcd_errorl("failed to kcalloc struct expcd_callback_node.");
		return -1;
	}
	cb_node->cb_func = pFunc;
	cb_node->data = NULL;
	list_add(&cb_node->node, &oc_callback_list);
	return 0;
}
EXPORT_SYMBOL(expcdOverCurrentRegister);

/*
 * unregister a OverCurrent callback.
*/
int expcdOverCurrentUnRegister(void *pFunc)
{
	struct expcd_callback_node *cb_node;
	struct expcd_callback_node *tmp_node;
	list_for_each_entry_safe(cb_node, tmp_node, &oc_callback_list, node) {
		if (cb_node->cb_func == pFunc) {
			list_del(&cb_node->node);
			kfree(cb_node);
		}
	}
	return 0;
}
EXPORT_SYMBOL(expcdOverCurrentUnRegister);

/*
 * generate a swi to tell the system that the ExpressCard is pulled out.
*/
int expcdForcedDetach(int slotno)
{
	extern int cxd9225gg_generate_irq(unsigned int);
	extern atomic_t expcd_force_op;

	atomic_set(&expcd_force_op, EXPCD_OP_PULLOUT);
	if (slotno == 0) {
		cxd9225gg_generate_irq(EXPCD_INT_S0_USB);
		cxd9225gg_generate_irq(EXPCD_INT_S0_PCIE);
	} else
	if (slotno == 1) {
		cxd9225gg_generate_irq(EXPCD_INT_S1_USB);
		cxd9225gg_generate_irq(EXPCD_INT_S1_PCIE);
	}
	return 0;
}

/*
 * generate a swi to tell the system that the ExpressCard is pulled out,
 * when the ExpressCard is not in OverCurrent state.
*/
int expcdErrorForceDetach(int slotno)
{
	if (expcdIsOverCurrent(slotno)) {
		/* Do not detach on OverCurrent state */
		return 0;
	}
	expcdForcedDetach(slotno);
	return 0;
}

#ifndef EXPCD_INCLUDE_AUTOPOW_CONTROL

int expcdAutoPowControl(int pw)
{
	return 0;
}

int expcdAutoPowUseControl(int slotno, int kind)
{
	return 0;
}

int expcdAutoPowUseForceControl(int slotno, int kind)
{
	return 0;
}
#else
#error EXPCD_INCLUDE_AUTOPOW_CONTROL is not implemented in vxworks
#endif

/* not used */
unsigned int expcdDriverVersionGet(void)
{
	expcd_errorl("This function is not implemented !!");
	return 0;
}

/* not used */
int expcdSelfDiagnose(int flag)
{
	expcd_errorl("This function is not implemented !!");
	return 0;
}

/* not used */
void expcdTestModeAllInterruptDisable(void)
{
	expcd_errorl("This function is not implemented !!");
}

