/* 2014-08-28: File changed by Sony Corporation */
/*
 *Copyright 2012 Sony Corporation
 *This program is free software; you can redistribute it and/or modify
 *it under the terms of the GNU General Public License, version 2, as
 *published by the Free Software Foundation.
 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *GNU General Public License for more details.
 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <linux/module.h>
#include <linux/fs.h>
//#include <linux/cdev.h>
#include <linux/delay.h>
#include <linux/sched.h>

//==============================================================================

#define fasync_printk(level, format, arg...) \
	printk(level "[Fasync] %4d | %s() @%d :> " format , \
		current->pid, __func__ , __LINE__ , ## arg)

#define	CONFIG_FASYNC_DEBUG
#ifdef 	CONFIG_FASYNC_DEBUG

#define fasync_debugl(format, arg...) \
	fasync_printk(KERN_DEBUG, format "\n" , ## arg)/*KERN_DEBUG*/

#define fasync_infol(format, arg...) \
	fasync_printk(KERN_NOTICE, format "\n" , ## arg)/*KERN_INFO*/

#define fasync_warningl(format, arg...) \
	fasync_printk(KERN_WARNING, format "\n" , ## arg)

#define fasync_errorl(format, arg...) \
	fasync_printk(KERN_ERR, "Error! " format "\n" , ## arg)

#else

#define fasync_debugl(format, arg...)

#define fasync_infol(format, arg...)

#define fasync_warningl(format, arg...)

#define fasync_errorl(format, arg...) \
	fasync_printk(KERN_ERR, "Error! " format "\n" , ## arg)

#endif

//==============================================================================

/*
 * A wait queue that our process (kernel space) will wait on.
*/
wait_queue_head_t fasync_waitqueue = __WAIT_QUEUE_HEAD_INITIALIZER(fasync_waitqueue);

/*
 * Reference of pending signals, used to sync signal handler callbacks.
 * Before kill_fasync(), this ref should be set as the tatal num of listeners.
 * Userspace signal handlers should call fasync_sighandler_done() to decrease
 * this ref (namely, by a ioctl() func).
 * When this ref reduced to ZERO, we know all user signal handers are called.
*/
static atomic_t fasync_sig_ref = ATOMIC_INIT(0);

/*
 * A list of listeners.
*/
static struct fasync_struct *fasync_list = NULL;

/*
 * Mutex between fasync_helper() and kill_fasync().
*/
DECLARE_MUTEX(fasync_sem);

/*
 * Set this function to a char device's struct file_operations, so that the
 * char device supports fasync.
*/
int fasync_setup(int fd, struct file * filp, int on)
{
	int retval;

	down(&fasync_sem);
	/* if on is 1, the helper will add a fasync with fd and filp. */
	retval = fasync_helper(fd, filp, on, &fasync_list);
	up(&fasync_sem);

	if (retval>0) {
		struct fasync_struct * fa;
		int fa_count = 0;
		for (fa=fasync_list; fa; fa=fa->fa_next, fa_count++) {}

		fasync_debugl("filp=0x%p, fd=%d, on=%d", filp, fd, on);
		fasync_debugl("%s listener=0x%p, total=%d", on?"append":"delete", filp, fa_count);
	}

	/* no change also means success. */
	return ((retval<0)?retval:0);
}
EXPORT_SYMBOL(fasync_setup);

/*
 * Send signal to all listeners those registered in the fasync_list.
 * This is a sync function that returns after all userspace signal handlers
 * have been finished.
 * Params:
 * sig   - signal number
 * async - 0 means to wait for someone to call fasync_sighandler_done();
 *         1 means to send signal and then return immediately.
*/
void fasync_sighandler_call(int sig, int async)
{
	struct fasync_struct * fa;

	static int cnt = 0;

	if (!fasync_list) {
		return;
	}

	down(&fasync_sem);
	{
		int fa_count = 0;
		for (fa = fasync_list; fa; fa=fa->fa_next) {
			/* If fa->fa_file->f_owner.signum is not set,
			 * SIGIO will be sent.
			*/
			fa->fa_file->f_owner.signum = sig;
			fa_count++;
		}
		if (0==fa_count) {
			up(&fasync_sem);
			return;
		}
		fasync_infol("send signal [%d], cnt=%d", sig, cnt++);

		/* mark how many times the signal is to be sent */
		atomic_set(&fasync_sig_ref, fa_count);

		/* send the signal */
		kill_fasync(&fasync_list, SIGIO, POLL_IN);
	}
	up(&fasync_sem);

	if (!async) {
		long ret = wait_event_timeout(fasync_waitqueue,
				(0==atomic_read(&fasync_sig_ref)),
				msecs_to_jiffies(1000));
		if (!ret) {
			fasync_debugl("fasync userspace handler timeout!");
		}
	}
}
EXPORT_SYMBOL(fasync_sighandler_call);

/*
 * This function is used to notify that one of the listeners has finished its
 * userspace signal handler.
*/
void fasync_sighandler_done(void)
{
	fasync_infol("");

	atomic_dec(&fasync_sig_ref);

	wake_up(&fasync_waitqueue);
}
EXPORT_SYMBOL(fasync_sighandler_done);

