/* 2014-08-28: File changed by Sony Corporation */
/*
 * The low performance USB storage driver (ub).
 *
 * Copyright (c) 1999, 2000 Matthew Dharm (mdharm-usb@one-eyed-alien.net)
 * Copyright (C) 2004 Pete Zaitcev (zaitcev@yahoo.com)
 *
 * This work is a part of Linux kernel, is derived from it,
 * and is not licensed separately. See file COPYING for details.
 *
 * TODO (sorted by decreasing priority)
 *  -- Return sense now that rq allows it (we always auto-sense anyway).
 *  -- set readonly flag for CDs, set removable flag for CF readers
 *  -- do inquiry and verify we got a disk and not a tape (for LUN mismatch)
 *  -- verify the 13 conditions and do bulk resets
 *  -- highmem
 *  -- move top_sense and work_bcs into separate allocations (if they survive)
 *     for cache purists and esoteric architectures.
 *  -- Allocate structure for LUN 0 before the first ub_sync_tur, avoid NULL. ?
 *  -- prune comments, they are too volumnous
 *  -- Resove XXX's
 *  -- CLEAR, CLR2STS, CLRRS seem to be ripe for refactoring.
 */
 
/*
 * USB Mass Storage Charactor Device (UMSC) Driver Module.
 */

#ifndef CONFIG_BLK_DEV_UB

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <linux/usb_usual.h>
#include <linux/cdev.h>
#include <linux/mm.h> //for page_address()
#include <linux/scatterlist.h> //for struct scatterlist
#include <linux/uaccess.h>
#include <mach/hardware.h>
#include <mach/cxd9225gg_cmn_debug.h>
#include "umsc.h"
#include <linux/timer.h>
#include <scsi/scsi.h>

/*
 * Debug Tools
 */
#define umsc_printk(level, format, arg...) \
	printk(level "[UMSC] %s() @%d :> " format, \
		__FUNCTION__, __LINE__, ## arg)

#define CONFIG_UMSC_DEBUG
#ifdef  CONFIG_UMSC_DEBUG

#define umsc_debugl(format, arg...) \
	umsc_printk(KERN_DEBUG, format "\n" , ## arg)/*KERN_INFO*/

#define umsc_infol(format, arg...) \
	umsc_printk(KERN_NOTICE, format "\n" , ## arg)/*KERN_NOTICE*/

#define umsc_warningl(format, arg...) \
	umsc_printk(KERN_WARNING, format "\n" , ## arg)

#define umsc_errorl(format, arg...) \
	umsc_printk(KERN_ERR, "Error! " format "\n" , ## arg)

#else /* CONFIG_UMSC_DEBUG */

#define umsc_debugl(format, arg...)

#define umsc_infol(format, arg...)

#define umsc_warningl(format, arg...)

#define umsc_errorl(format, arg...) \
	umsc_printk(KERN_ERR, "Error! " format "\n" , ## arg)
	
#endif /* CONFIG_UMSC_DEBUG */

//==============================================================================
#define DRV_NAME "umsc"

#define UB_MAJOR 180

/*
 * The command state machine is the key model for understanding of this driver.
 *
 * The general rule is that all transitions are done towards the bottom
 * of the diagram, thus preventing any loops.
 *
 * An exception to that is how the STAT state is handled. A counter allows it
 * to be re-entered along the path marked with [C].
 *
 *       +--------+
 *       ! INIT   !
 *       +--------+
 *           !
 *        ub_scsi_cmd_start fails ->--------------------------------------\
 *           !                                                            !
 *           V                                                            !
 *       +--------+                                                       !
 *       ! CMD    !                                                       !
 *       +--------+                                                       !
 *           !                                            +--------+      !
 *         was -EPIPE -->-------------------------------->! CLEAR  !      !
 *           !                                            +--------+      !
 *           !                                                !           !
 *         was error -->------------------------------------- ! --------->\
 *           !                                                !           !
 *  /--<-- cmd->dir == NONE ?                                 !           !
 *  !        !                                                !           !
 *  !        V                                                !           !
 *  !    +--------+                                           !           !
 *  !    ! DATA   !                                           !           !
 *  !    +--------+                                           !           !
 *  !        !                           +---------+          !           !
 *  !      was -EPIPE -->--------------->! CLR2STS !          !           !
 *  !        !                           +---------+          !           !
 *  !        !                                !               !           !
 *  !        !                              was error -->---- ! --------->\
 *  !      was error -->--------------------- ! ------------- ! --------->\
 *  !        !                                !               !           !
 *  !        V                                !               !           !
 *  \--->+--------+                           !               !           !
 *       ! STAT   !<--------------------------/               !           !
 *  /--->+--------+                                           !           !
 *  !        !                                                !           !
 * [C]     was -EPIPE -->-----------\                         !           !
 *  !        !                      !                         !           !
 *  +<---- len == 0                 !                         !           !
 *  !        !                      !                         !           !
 *  !      was error -->--------------------------------------!---------->\
 *  !        !                      !                         !           !
 *  +<---- bad CSW                  !                         !           !
 *  +<---- bad tag                  !                         !           !
 *  !        !                      V                         !           !
 *  !        !                 +--------+                     !           !
 *  !        !                 ! CLRRS  !                     !           !
 *  !        !                 +--------+                     !           !
 *  !        !                      !                         !           !
 *  \------- ! --------------------[C]--------\               !           !
 *           !                                !               !           !
 *         cmd->error---\                +--------+           !           !
 *           !          +--------------->! SENSE  !<----------/           !
 *         STAT_FAIL----/                +--------+                       !
 *           !                                !                           V
 *           !                                V                      +--------+
 *           \--------------------------------\--------------------->! DONE   !
 *                                                                   +--------+
 */

/*
 * This many LUNs per USB device.
 * Every one of them takes a host, see UB_MAX_HOSTS.
 */
#define UB_MAX_LUNS   9

/*
 */

#define UB_PARTS_PER_LUN      8

#define UB_MAX_CDB_SIZE      16		/* Corresponds to Bulk */

#define UB_SENSE_SIZE  18

/*
 */

/* command block wrapper */
struct bulk_cb_wrap {
	__le32	Signature;		/* contains 'USBC' */
	u32	Tag;			/* unique per command id */
	__le32	DataTransferLength;	/* size of data */
	u8	Flags;			/* direction in bit 0 */
	u8	Lun;			/* LUN */
	u8	Length;			/* of of the CDB */
	u8	CDB[UB_MAX_CDB_SIZE];	/* max command */
};

#define US_BULK_CB_WRAP_LEN	31
#define US_BULK_CB_SIGN		0x43425355	/*spells out USBC */
#define US_BULK_FLAG_IN		1
#define US_BULK_FLAG_OUT	0

/* command status wrapper */
struct bulk_cs_wrap {
	__le32	Signature;		/* should = 'USBS' */
	u32	Tag;			/* same as original command */
	__le32	Residue;		/* amount not transferred */
	u8	Status;			/* see below */
};

#define US_BULK_CS_WRAP_LEN	13
#define US_BULK_CS_SIGN		0x53425355	/* spells out 'USBS' */
#define US_BULK_STAT_OK		0
#define US_BULK_STAT_FAIL	1
#define US_BULK_STAT_PHASE	2

/* bulk-only class specific requests */
#define US_BULK_RESET_REQUEST	0xff
#define US_BULK_GET_MAX_LUN	0xfe

/*
 */
struct ub_dev;

#define UB_MAX_REQ_SG	9	/* cdrecord requires 32KB and maybe a header */
#define UB_MAX_SECTORS 64

/*
 * A second is more than enough for a 32K transfer (UB_MAX_SECTORS)
 * even if a webcam hogs the bus, but some devices need time to spin up.
 */
#define UB_URB_TIMEOUT	(HZ*2)
#define UB_DATA_TIMEOUT	(HZ*5)	/* ZIP does spin-ups in the data phase */
#define UB_STAT_TIMEOUT	(HZ*5)	/* Same spinups and eject for a dataless cmd. */
#define UB_CTRL_TIMEOUT	(HZ/2)	/* 500ms ought to be enough to clear a stall */

/*
 * An instance of a SCSI command in transit.
 */
#define UB_DIR_NONE	0
#define UB_DIR_READ	1
#define UB_DIR_ILLEGAL2	2
#define UB_DIR_WRITE	3

#define UB_DIR_CHAR(c)  (((c)==UB_DIR_WRITE)? 'w': \
			 (((c)==UB_DIR_READ)? 'r': 'n'))

enum ub_scsi_cmd_state {
	UB_CMDST_INIT,			/* Initial state */
	UB_CMDST_CMD,			/* Command submitted */
	UB_CMDST_DATA,			/* Data phase */
	UB_CMDST_CLR2STS,		/* Clearing before requesting status */
	UB_CMDST_STAT,			/* Status phase */
	UB_CMDST_CLEAR,			/* Clearing a stall (halt, actually) */
	UB_CMDST_CLRRS,			/* Clearing before retrying status */
	UB_CMDST_SENSE,			/* Sending Request Sense */
	UB_CMDST_DONE			/* Final state */
};

struct ub_scsi_cmd {
	unsigned char cdb[UB_MAX_CDB_SIZE];
	unsigned char cdb_len;

	unsigned char dir;		/* 0 - none, 1 - read, 3 - write. */
	enum ub_scsi_cmd_state state;
	unsigned int tag;
	struct ub_scsi_cmd *next;

	int error;			/* Return code - valid upon done */
	unsigned int act_len;		/* Return size */
	unsigned char key, asc, ascq;	/* May be valid if error==-EIO */

	int stat_count;			/* Retries getting status. */

	unsigned int len;		/* Requested length */
	unsigned int current_sg;
	unsigned int nsg;		/* sgv[nsg] */
	struct scatterlist sgv[UB_MAX_REQ_SG];

	struct ub_lun *lun;
	void (*done)(struct ub_dev *, struct ub_scsi_cmd *);
	void *back;
};

struct ub_request {
	struct usbcard_cmd uc;          /* user space command data */
	unsigned int current_try;
	unsigned int nsg;		/* sgv[nsg] */
	struct scatterlist sgv[UB_MAX_REQ_SG];
};

/*
 */
struct ub_capacity {
	unsigned long nsec;		/* Linux size - 512 byte sectors */
	unsigned int bsize;		/* Linux hardsect_size */
	unsigned int bshift;		/* Shift between 512 and hard sects */
};

/*
 * This is a direct take-off from linux/include/completion.h
 * The difference is that I do not wait on this thing, just poll.
 * When I want to wait (ub_probe), I just use the stock completion.
 *
 * Note that INIT_COMPLETION takes no lock. It is correct. But why
 * in the bloody hell that thing takes struct instead of pointer to struct
 * is quite beyond me. I just copied it from the stock completion.
 */
struct ub_completion {
	unsigned int done;
	spinlock_t lock;
};

static inline void ub_init_completion(struct ub_completion *x)
{
	x->done = 0;
	spin_lock_init(&x->lock);
}

#define UB_INIT_COMPLETION(x)	((x).done = 0)

static void ub_complete(struct ub_completion *x)
{
	unsigned long flags;

	spin_lock_irqsave(&x->lock, flags);
	x->done++;
	spin_unlock_irqrestore(&x->lock, flags);
}

static int ub_is_completed(struct ub_completion *x)
{
	unsigned long flags;
	int ret;

	spin_lock_irqsave(&x->lock, flags);
	ret = x->done;
	spin_unlock_irqrestore(&x->lock, flags);
	return ret;
}

/*
 */
struct ub_scsi_cmd_queue {
	int qlen, qmax;
	struct ub_scsi_cmd *head, *tail;
};

/*
 * The block device instance (one per LUN).
 */
struct ub_lun {
	struct ub_dev *udev;
	struct list_head link;
	int id;				/* Host index */
	int num;			/* LUN number */
	char name[16];

	int changed;			/* Media was changed */
	int removable;
	int readonly;

	struct ub_request urq;

	/* Use Ingo's mempool if or when we have more than one command. */
	/*
	 * Currently we never need more than one command for the whole device.
	 * However, giving every LUN a command is a cheap and automatic way
	 * to enforce fairness between them.
	 */
	int cmda[1];
	struct ub_scsi_cmd cmdv[1];

	struct ub_capacity capacity;
};

/*
 * The USB device instance.
 */
struct ub_dev {
	spinlock_t *lock;
	atomic_t poison;		/* The USB device is disconnected */
	int openc;			/* protected by ub_lock! */
					/* kref is too implicit for our taste */
	int reset;			/* Reset is running */
	unsigned int tagcnt;
	char name[12];
	struct usb_device *dev;
	struct usb_interface *intf;

	int maxlun;

	struct list_head luns;

	unsigned int send_bulk_pipe;	/* cached pipe values */
	unsigned int recv_bulk_pipe;
	unsigned int send_ctrl_pipe;
	unsigned int recv_ctrl_pipe;

	struct tasklet_struct tasklet;

	struct ub_scsi_cmd_queue cmd_queue;
	struct ub_scsi_cmd top_rqs_cmd;	/* REQUEST SENSE */
	unsigned char top_sense[UB_SENSE_SIZE];

	struct ub_completion work_done;
	struct urb work_urb;
	struct timer_list work_timer;
	int last_pipe;			/* What might need clearing */
	__le32 signature;		/* Learned signature */
	struct bulk_cb_wrap work_bcb;
	struct bulk_cs_wrap work_bcs;
	struct usb_ctrlrequest work_cr;

	struct work_struct reset_work;
	wait_queue_head_t reset_wait;

	int sg_stat[6];
};

/*
 */
static void ub_cleanup(struct ub_dev *sc);
static void ub_urb_complete(struct urb *urb);
static void ub_scsi_action(unsigned long _dev);
static void ub_scsi_dispatch(struct ub_dev *sc);
static void ub_scsi_urb_compl(struct ub_dev *sc, struct ub_scsi_cmd *cmd);
static void ub_data_start(struct ub_dev *sc, struct ub_scsi_cmd *cmd);
static void ub_state_done(struct ub_dev *sc, struct ub_scsi_cmd *cmd, int rc);
static int __ub_state_stat(struct ub_dev *sc, struct ub_scsi_cmd *cmd);
static void ub_state_stat(struct ub_dev *sc, struct ub_scsi_cmd *cmd);
static void ub_state_stat_counted(struct ub_dev *sc, struct ub_scsi_cmd *cmd);
static void ub_state_sense(struct ub_dev *sc, struct ub_scsi_cmd *cmd);
static int ub_submit_clear_stall(struct ub_dev *sc, struct ub_scsi_cmd *cmd,
    int stalled_pipe);
static void ub_top_sense_done(struct ub_dev *sc, struct ub_scsi_cmd *scmd);
static int ub_sync_tur(struct ub_dev *sc, struct ub_lun *lun);
static int ub_sync_read_cap(struct ub_dev *sc, struct ub_lun *lun,
    struct ub_capacity *ret);
static int ub_probe_lun(struct ub_dev *sc, int lnum);

/*
 */
#ifdef CONFIG_USB_LIBUSUAL

#define ub_usb_ids  storage_usb_ids
#else

static struct usb_device_id ub_usb_ids[] = {
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_SCSI, US_PR_BULK) },
	{ }
};

MODULE_DEVICE_TABLE(usb, ub_usb_ids);
#endif /* CONFIG_USB_LIBUSUAL */

/*
 * Find me a way to identify "next free minor" for add_disk(),
 * and the array disappears the next day. However, the number of
 * hosts has something to do with the naming and /proc/partitions.
 * This has to be thought out in detail before changing.
 * If UB_MAX_HOST was 1000, we'd use a bitmap. Or a better data structure.
 */

/*
 * Minor == Slot Index
 * Slot0 --> Minor = 0 --> /dev/umsc0
 * Slot1 --> Minor = 1 --> /dev/umsc1
 */
#define UMSC_MINORS   2

/* Return the first LUN of the ub_dev. */
#define umsc_dev_to_lun(p_udev) \
	container_of(p_udev->luns.next, struct ub_lun, link)

#define UMSC_SHMEM_MAX_LEN  (1*1024*1024)       /* 1 MB - 4 MB is good */
static unsigned int umsc_shmem_len = UMSC_SHMEM_MAX_LEN;
static void * umsc_shmem = NULL;                /* kernel logic address */

extern int expcd_slot_check(int slotno, int iface);
extern int expcd_slot_get_power(int slotno);
extern int aldControl(int slot, int color, int on);


#define UB_QLOCK_NUM 5
static spinlock_t ub_qlockv[UB_QLOCK_NUM];
static int ub_qlock_next = 0;

static DEFINE_SPINLOCK(ub_lock);	/* Locks globals and ->openc */



/* : Minors Management
 *
 * Each usb device has a level and a port (from its hub). The level of root
 * hub is 0. Device directly connected with root hub will be level 1. If a
 * hub is on level 1, then all its children are level 2.
 *
 * Here we use a minor-level-port map to record devices which we support.
 * The index of this map is treated as our cdev minor.
 * The element of this map is a byte, of which the bits are like this:
 * | X X X X | X X X | X |
 * | portnum | level | v | <-- 1 bit, indicates the minor being Valid or not.
 *
 * So, we at most support 8 levels, and 16 ports for each level.
 * This is enough for our case.
 *
 * At initialization, we must configure each minor with certain level and
 * portnum, which we want to support. So that, the driver will only support
 * USB ports those are specified explicitly. ---- This feature is good on
 * cxd9225gg case, where we can only support the 2 USB ExpressCard ports without
 * others, even there are more USB ports on the same hub.
*/
struct umsc_chardev {
	struct cdev     chardev;
	struct ub_dev * volatile udevs[UMSC_MINORS];
	unsigned char volatile port_map[UMSC_MINORS];
};
static struct umsc_chardev umsc_cdev = {
	.udevs    = {NULL, NULL},
	.port_map = {0x00, 0x00},
};

/* Return the cdev struct of the inode. */
#define inode_to_umsc(p_inode) \
	container_of((p_inode)->i_cdev, struct umsc_chardev, chardev)

static inline unsigned char umsc_makeup_dev_id(int level, int portnum) {
	return (((portnum & 0xf) << 4) | ((level & 0x7) << 1));
}

/* Configure a valid dev_id for the given minor. */
static inline void umsc_config_port_map(int minor, int level, int portnum)
{
	umsc_cdev.port_map[minor] = umsc_makeup_dev_id(level, portnum);
}

/*
 * Check if the given level+port is supported.
 * return 0 if is supported, otherwise return -1.
*/
static int umsc_id_check(int level, int portnum)
{
	unsigned long spin_flags;
	int ret = -1;

	int i;
	spin_lock_irqsave(&ub_lock, spin_flags);
	for (i=0; i<UMSC_MINORS; i++) {
		if (umsc_cdev.port_map[i] == umsc_makeup_dev_id(level, portnum)) {
			ret = i;
			break;
		}
	}
	spin_unlock_irqrestore(&ub_lock, spin_flags);

	return (ret>=0 ? 0 : -1);
}

/*
 * Get proper minor.
 * level   - level of the usb hub, root hub is Level1. [1,2, ...)
 * portnum - port of the usb hub, first port is Port1. [1,2,3]
 */
static int umsc_id_get(int level, int portnum, void* ptr)
{
	unsigned long spin_flags;
	int ret = -1;

	int i;
	spin_lock_irqsave(&ub_lock, spin_flags);
	for (i=0; i<UMSC_MINORS; i++) {
		if (umsc_cdev.port_map[i] == umsc_makeup_dev_id(level, portnum)) {
			umsc_cdev.port_map[i] |= 0x1;
			umsc_cdev.udevs[i] = ptr;
			ret = i;
			break;
		}
	}
	spin_unlock_irqrestore(&ub_lock, spin_flags);

	return ret;
}

static void umsc_id_put(int id)
{
	unsigned long spin_flags;

	if (id < 0 || id >= UMSC_MINORS) {
		printk(KERN_ERR DRV_NAME ": bad minor %d\n", id);
		return;
	}

	spin_lock_irqsave(&ub_lock, spin_flags);
	if ((umsc_cdev.port_map[id] & 1) == 0) {
		spin_unlock_irqrestore(&ub_lock, spin_flags);
		printk(KERN_ERR DRV_NAME ": freeing free minor %d\n", id);
		return;
	}
	umsc_cdev.port_map[id] &= ~1;
	umsc_cdev.udevs[id] = NULL;
	spin_unlock_irqrestore(&ub_lock, spin_flags);
}

/*
 * This is necessitated by the fact that blk_cleanup_queue does not
 * necesserily destroy the queue. Instead, it may merely decrease q->refcnt.
 * Since our blk_init_queue() passes a spinlock common with ub_dev,
 * we have life time issues when ub_cleanup frees ub_dev.
 */
static spinlock_t *ub_next_lock(void)
{
	unsigned long flags;
	spinlock_t *ret;

	spin_lock_irqsave(&ub_lock, flags);
	ret = &ub_qlockv[ub_qlock_next];
	ub_qlock_next = (ub_qlock_next + 1) % UB_QLOCK_NUM;
	spin_unlock_irqrestore(&ub_lock, flags);
	return ret;
}

/*
 * Downcount for deallocation. This rides on two assumptions:
 *  - once something is poisoned, its refcount cannot grow
 *  - opens cannot happen at this time (del_gendisk was done)
 * If the above is true, we can drop the lock, which we need for
 * blk_cleanup_queue(): the silly thing may attempt to sleep.
 * [Actually, it never needs to sleep for us, but it calls might_sleep()]
 */
static void ub_put(struct ub_dev *sc)
{
	unsigned long flags;

	spin_lock_irqsave(&ub_lock, flags);
	--sc->openc;
	if (sc->openc == 0 && atomic_read(&sc->poison)) {
		spin_unlock_irqrestore(&ub_lock, flags);
		ub_cleanup(sc);
	} else {
		spin_unlock_irqrestore(&ub_lock, flags);
	}
}

/*
 * Final cleanup and deallocation.
 */
static void ub_cleanup(struct ub_dev *sc)
{
	struct list_head *p;
	struct ub_lun *lun;

	while (!list_empty(&sc->luns)) {
		p = sc->luns.next;
		lun = list_entry(p, struct ub_lun, link);
		list_del(p);

		
		umsc_id_put(lun->id);
		kfree(lun);
		lun = NULL;
	}

	usb_set_intfdata(sc->intf, NULL);
	usb_put_intf(sc->intf);
	usb_put_dev(sc->dev);
	kfree(sc);
	sc = NULL;
}

/*
 * The "command allocator".
 */
static struct ub_scsi_cmd *ub_get_cmd(struct ub_lun *lun)
{
	struct ub_scsi_cmd *ret;

	if (lun->cmda[0])
		return NULL;
	ret = &lun->cmdv[0];
	lun->cmda[0] = 1;
	return ret;
}

static void ub_put_cmd(struct ub_lun *lun, struct ub_scsi_cmd *cmd)
{
	if (cmd != &lun->cmdv[0]) {
		printk(KERN_WARNING "%s: releasing a foreign cmd %p\n",
		    lun->name, cmd);
		return;
	}
	if (!lun->cmda[0]) {
		printk(KERN_WARNING "%s: releasing a free cmd\n", lun->name);
		return;
	}
	lun->cmda[0] = 0;
}

/*
 * The command queue.
 */
static void ub_cmdq_add(struct ub_dev *sc, struct ub_scsi_cmd *cmd)
{
	struct ub_scsi_cmd_queue *t = &sc->cmd_queue;

	if (t->qlen++ == 0) {
		t->head = cmd;
		t->tail = cmd;
	} else {
		t->tail->next = cmd;
		t->tail = cmd;
	}

	if (t->qlen > t->qmax)
		t->qmax = t->qlen;
}

static void ub_cmdq_insert(struct ub_dev *sc, struct ub_scsi_cmd *cmd)
{
	struct ub_scsi_cmd_queue *t = &sc->cmd_queue;

	if (t->qlen++ == 0) {
		t->head = cmd;
		t->tail = cmd;
	} else {
		cmd->next = t->head;
		t->head = cmd;
	}

	if (t->qlen > t->qmax)
		t->qmax = t->qlen;
}

static struct ub_scsi_cmd *ub_cmdq_pop(struct ub_dev *sc)
{
	struct ub_scsi_cmd_queue *t = &sc->cmd_queue;
	struct ub_scsi_cmd *cmd;

	if (t->qlen == 0)
		return NULL;
	if (--t->qlen == 0)
		t->tail = NULL;
	cmd = t->head;
	t->head = cmd->next;
	cmd->next = NULL;
	return cmd;
}

#define ub_cmdq_peek(sc)  ((sc)->cmd_queue.head)

/*
 * The request function is our main entry point
 */


#if 1
/*
 * urq - a request to be build.
 * uc - a user space command struct.
 */
static void umsc_build_request(struct ub_request * urq, struct usbcard_cmd * uc)
{
	memset(urq, 0, sizeof(struct ub_request));
	memcpy(&urq->uc, uc, sizeof(struct usbcard_cmd));

	/* how many sg vectors, always be 1 */
	urq->nsg = 1;

	/* setting sg page, offset, length */
	/* 
	 * Add a data_flg to indicate different situations.
	*/
	memset(urq->sgv, 0, sizeof(urq->sgv));
	sg_init_table(&urq->sgv[0], UB_MAX_REQ_SG);

	if (0 != uc->data_len) {
		if (uc->data_flg == UMSC_FLG_PHYS_ADDR) {
			/* directly set dma_address, since it is physical address */
			urq->sgv[0].dma_address = (unsigned long)uc->data_buf;
			urq->sgv[0].length = uc->data_len;
		} else
		if (uc->data_flg == UMSC_FLG_KERN_ADDR) {
			/* translate kernel logical address to PAGE and OFFSET */
			sg_set_page(&urq->sgv[0], virt_to_page(uc->data_buf), uc->data_len,
					((unsigned long)uc->data_buf)&(PAGE_SIZE-1));
		} else
		if (uc->data_flg == UMSC_FLG_SHMM_OFFS) {
			/* first translate shmem offset to kernel logical address */
			uc->data_buf = umsc_shmem + (unsigned long)uc->data_buf;
			/* then translate kernel logical address to PAGE and OFFSET */
			sg_set_page(&urq->sgv[0], virt_to_page(uc->data_buf), uc->data_len,
					((unsigned long)uc->data_buf)&(PAGE_SIZE-1));
		} else
		if (uc->data_flg == UMSC_FLG_USER_ADDR) {
			/* need copy, not supported now */
		} else {
			/* bad data_flg, should not happen */
			umsc_errorl("bad UMSC_FLG [0x%08x]", uc->data_flg);
			urq->nsg = 0;
		}
	}
}

static void umsc_build_command(struct ub_scsi_cmd *cmd, struct ub_request *urq)
{
	if (urq->uc.data_len == 0) {
		cmd->dir = UB_DIR_NONE;
	} else {
		if (urq->uc.data_dir == WRITE) {
			cmd->dir = UB_DIR_WRITE;
		} else {
			cmd->dir = UB_DIR_READ;
		}
	}

	cmd->nsg = urq->nsg;
	memcpy(cmd->sgv, urq->sgv, sizeof(struct scatterlist) * cmd->nsg);

	memcpy(&cmd->cdb, urq->uc.cbd, urq->uc.cbd_len);
	cmd->cdb_len = urq->uc.cbd_len;

	cmd->len = urq->uc.data_len;
}

/*
 * Submit a regular SCSI operation (not an auto-sense)
 * without jumping to a tasklet.
 * See ub_submit_scsi() function.
 *
 * The Iron Law of Good Submit Routine is:
 * Zero return - callback is done, Nonzero return - callback is not done.
 * No exceptions.
 *
 * Host is assumed locked.
 */
static int umsc_sync_submit_scsi(struct ub_dev *sc, struct ub_scsi_cmd *cmd)
{

	if (cmd->state != UB_CMDST_INIT ||
	    (cmd->dir != UB_DIR_NONE && cmd->len == 0)) {
		return -EINVAL;
	}

	ub_cmdq_add(sc, cmd);
	ub_scsi_dispatch(sc);
	return 0;
}

static void umsc_sync_urq_done(struct ub_dev *sc, struct ub_scsi_cmd *cmd)
{
	struct ub_lun *lun = cmd->lun;
	struct ub_request *urq = &lun->urq;
	struct completion *cop = cmd->back;

	if (cmd->error == 0) {
		/* Success */
	} else {
		if (cmd->error == -EIO) {
			memcpy(urq->uc.sense, sc->top_sense, UB_SENSE_SIZE);
			urq->uc.sense_len = UB_SENSE_SIZE;
		}
	}

	/* urq->uc will be return to user space */
	urq->uc.error = cmd->error;

	/* clear urb.transfer_flags before complete() */
	sc->work_urb.transfer_flags &= ~URB_NO_TRANSFER_DMA_MAP;

	complete(cop);

	ub_put_cmd(lun, cmd);

	aldControl(lun->id, 0, 1);
}

/*
 * Translate ub_request to ub_scsi_cmd, and do the command synchronously.
 */
static int umsc_sync_do_request(struct ub_dev* sc, struct ub_request * urq)
{
	struct ub_lun * lun;
	struct ub_scsi_cmd *cmd;
	struct completion compl;
	unsigned long flags;
	int rc;

	init_completion(&compl);

	lun = umsc_dev_to_lun(sc);
	if (!lun) {
		umsc_errorl("");
		return -1;
	}

	cmd = ub_get_cmd(lun);
	if (!cmd) {
		umsc_errorl("");
		return -1;
	}
	memset(cmd, 0, sizeof(struct ub_scsi_cmd));

	umsc_build_command(cmd, urq);

	cmd->state = UB_CMDST_INIT;
	cmd->lun = lun;
	cmd->done = umsc_sync_urq_done;
	cmd->back = &compl;

	aldControl(lun->id, 1, 1);

	spin_lock_irqsave(sc->lock, flags);
	cmd->tag = sc->tagcnt++;

	rc = umsc_sync_submit_scsi(sc, cmd);
	spin_unlock_irqrestore(sc->lock, flags);

	if (rc) {
		umsc_errorl("ub_sync_submit_scsi() failed !!");
		goto err_submit;
	}

	if (wait_for_completion_interruptible(&compl) < 0) {
		umsc_errorl("wake up by signal!");
	}

	if (cmd->error) {
		if (cmd->error==-EIO) {
			switch (cmd->cdb[0]) {
			case READ_10:
			case WRITE_10:
			{
				/* print R/W error info */
				unsigned int block=0, nblks=0;
				block = (((unsigned int)cmd->cdb[2])<<24)
					| (((unsigned int)cmd->cdb[3])<<16)
					| (((unsigned int)cmd->cdb[4])<<8)
					| (((unsigned int)cmd->cdb[5]));
				nblks = (((unsigned int)cmd->cdb[7])<<8)
					| (((unsigned int)cmd->cdb[8]));

				umsc_errorl("CMD=0x%02x: "
					"block=%u, nblks=%u, "
					"data_len=%u, ERR=%d, KEY=0x%02x, "
					"ASC=0x%02x, ASCQ=0x%02x",
					cmd->cdb[0], block, nblks,
					cmd->len, cmd->error, cmd->key,
					cmd->asc, cmd->ascq);
				break;
			}
			case MODE_SENSE:
			{
				unsigned int page_code=0, alloc_len=0;
				page_code = cmd->cdb[2];
				alloc_len = cmd->cdb[4];
				umsc_errorl("CMD=0x%02x: "
					"page_code=0x%0x, alloc_len=0x%0x, "
					"data_len=%u, ERR=%d, KEY=0x%02x, "
					"ASC=0x%02x, ASCQ=0x%02x",
					cmd->cdb[0], page_code, alloc_len,
					cmd->len, cmd->error, cmd->key,
					cmd->asc, cmd->ascq);
				break;
			}
			default:
				break;
			}
		} else {
			umsc_errorl("CMD=0x%02x: ERR=%d, KEY=0x%02x, ASC=0x%02x, ASCQ=0x%02x",
				cmd->cdb[0], cmd->error, cmd->key, cmd->asc, cmd->ascq);
		}
	}

	return cmd->error;
	
err_submit:
	return -1;
}

static int umsc_process_user_cmd(struct ub_dev * sc, struct usbcard_cmd *uc)
{
	struct ub_lun * lun;
	struct ub_request * urq;
	int ret = 0;

	lun = umsc_dev_to_lun(sc);
	if (!lun) {
		umsc_errorl("ub_dev has no LUN !");
		return -1;
	}

	urq = &lun->urq;

	umsc_build_request(urq, uc);

	ret = umsc_sync_do_request(sc, urq);

	return ret;
}
#endif



/*
 * Submit the first URB for the queued command.
 * This function does not deal with queueing in any way.
 */
static int ub_scsi_cmd_start(struct ub_dev *sc, struct ub_scsi_cmd *cmd)
{
	struct bulk_cb_wrap *bcb;
	int rc;

	bcb = &sc->work_bcb;

	/*
	 * ``If the allocation length is eighteen or greater, and a device
	 * server returns less than eithteen bytes of data, the application
	 * client should assume that the bytes not transferred would have been
	 * zeroes had the device server returned those bytes.''
	 *
	 * We zero sense for all commands so that when a packet request
	 * fails it does not return a stale sense.
	 */
	memset(sc->top_sense, 0, sizeof(sc->top_sense));

	/* set up the command wrapper */
	bcb->Signature = cpu_to_le32(US_BULK_CB_SIGN);
	bcb->Tag = cmd->tag;		/* Endianness is not important */
	bcb->DataTransferLength = cpu_to_le32(cmd->len);
	bcb->Flags = (cmd->dir == UB_DIR_READ) ? 0x80 : 0;
	bcb->Lun = (cmd->lun != NULL) ? cmd->lun->num : 0;
	bcb->Length = cmd->cdb_len;

	/* copy the command payload */
	memcpy(bcb->CDB, cmd->cdb, UB_MAX_CDB_SIZE);

	UB_INIT_COMPLETION(sc->work_done);

	sc->last_pipe = sc->send_bulk_pipe;
	usb_fill_bulk_urb(&sc->work_urb, sc->dev, sc->send_bulk_pipe,
	    bcb, US_BULK_CB_WRAP_LEN, ub_urb_complete, sc);

	/* Fill what we shouldn't be filling, because usb-storage did so. */
	sc->work_urb.actual_length = 0;
	sc->work_urb.error_count = 0;
	sc->work_urb.status = 0;

	if ((rc = usb_submit_urb(&sc->work_urb, GFP_ATOMIC)) != 0) {
		/* XXX Clear stalls */
		ub_complete(&sc->work_done);
		return rc;
	}

	sc->work_timer.expires = jiffies + UB_URB_TIMEOUT;
	add_timer(&sc->work_timer);

	cmd->state = UB_CMDST_CMD;
	return 0;
}

/*
 * Timeout handler.
 */
static void ub_urb_timeout(unsigned long arg)
{
	struct ub_dev *sc = (struct ub_dev *) arg;
	unsigned long flags;

	spin_lock_irqsave(sc->lock, flags);
	if (!ub_is_completed(&sc->work_done))
		usb_unlink_urb(&sc->work_urb);
	spin_unlock_irqrestore(sc->lock, flags);
}

/*
 * Completion routine for the work URB.
 *
 * This can be called directly from usb_submit_urb (while we have
 * the sc->lock taken) and from an interrupt (while we do NOT have
 * the sc->lock taken). Therefore, bounce this off to a tasklet.
 */
static void ub_urb_complete(struct urb *urb)
{
	struct ub_dev *sc = urb->context;


	/* 
	 * do not forget to clear the URB_NO_TRANSFER_DMA_MAP flag bit.
	*/
	sc->work_urb.transfer_flags &= ~URB_NO_TRANSFER_DMA_MAP;


	ub_complete(&sc->work_done);
	tasklet_schedule(&sc->tasklet);
}

static void ub_scsi_action(unsigned long _dev)
{
	struct ub_dev *sc = (struct ub_dev *) _dev;
	unsigned long flags;

	spin_lock_irqsave(sc->lock, flags);
	ub_scsi_dispatch(sc);
	spin_unlock_irqrestore(sc->lock, flags);
}

static void ub_scsi_dispatch(struct ub_dev *sc)
{
	struct ub_scsi_cmd *cmd;
	int rc;

	while (!sc->reset && (cmd = ub_cmdq_peek(sc)) != NULL) {
		if (cmd->state == UB_CMDST_DONE) {
			ub_cmdq_pop(sc);
			(*cmd->done)(sc, cmd);
		} else if (cmd->state == UB_CMDST_INIT) {
			if ((rc = ub_scsi_cmd_start(sc, cmd)) == 0)
				break;
			cmd->error = rc;
			cmd->state = UB_CMDST_DONE;
		} else {
			if (!ub_is_completed(&sc->work_done))
				break;
			del_timer(&sc->work_timer);
			ub_scsi_urb_compl(sc, cmd);
		}
	}
}

static void ub_scsi_urb_compl(struct ub_dev *sc, struct ub_scsi_cmd *cmd)
{
	struct urb *urb = &sc->work_urb;
	struct bulk_cs_wrap *bcs;
	int len;
	int rc;

	if (atomic_read(&sc->poison)) {
		ub_state_done(sc, cmd, -ENODEV);
		return;
	}

	if (cmd->state == UB_CMDST_CLEAR) {
		if (urb->status == -EPIPE) {
			/*
			 * STALL while clearning STALL.
			 * The control pipe clears itself - nothing to do.
			 */
			printk(KERN_NOTICE "%s: stall on control pipe\n",
			    sc->name);
			goto Bad_End;
		}

		/*
		 * We ignore the result for the halt clear.
		 */

		/* reset the endpoint toggle */
		usb_settoggle(sc->dev, usb_pipeendpoint(sc->last_pipe),
			usb_pipeout(sc->last_pipe), 0);

		ub_state_sense(sc, cmd);

	} else if (cmd->state == UB_CMDST_CLR2STS) {
		if (urb->status == -EPIPE) {
			printk(KERN_NOTICE "%s: stall on control pipe\n",
			    sc->name);
			goto Bad_End;
		}

		/*
		 * We ignore the result for the halt clear.
		 */

		/* reset the endpoint toggle */
		usb_settoggle(sc->dev, usb_pipeendpoint(sc->last_pipe),
			usb_pipeout(sc->last_pipe), 0);

		ub_state_stat(sc, cmd);

	} else if (cmd->state == UB_CMDST_CLRRS) {
		if (urb->status == -EPIPE) {
			printk(KERN_NOTICE "%s: stall on control pipe\n",
			    sc->name);
			goto Bad_End;
		}

		/*
		 * We ignore the result for the halt clear.
		 */

		/* reset the endpoint toggle */
		usb_settoggle(sc->dev, usb_pipeendpoint(sc->last_pipe),
			usb_pipeout(sc->last_pipe), 0);

		ub_state_stat_counted(sc, cmd);

	} else if (cmd->state == UB_CMDST_CMD) {
		switch (urb->status) {
		case 0:
			break;
		case -EOVERFLOW:
			goto Bad_End;
		case -EPIPE:
			rc = ub_submit_clear_stall(sc, cmd, sc->last_pipe);
			if (rc != 0) {
				printk(KERN_NOTICE "%s: "
				    "unable to submit clear (%d)\n",
				    sc->name, rc);
				/*
				 * This is typically ENOMEM or some other such shit.
				 * Retrying is pointless. Just do Bad End on it...
				 */
				ub_state_done(sc, cmd, rc);
				return;
			}
			cmd->state = UB_CMDST_CLEAR;
			return;
		case -ESHUTDOWN:	/* unplug */
		case -EILSEQ:		/* unplug timeout on uhci */
			ub_state_done(sc, cmd, -ENODEV);
			return;
		default:
			goto Bad_End;
		}
		if (urb->actual_length != US_BULK_CB_WRAP_LEN) {
			goto Bad_End;
		}

		if (cmd->dir == UB_DIR_NONE || cmd->nsg < 1) {
			ub_state_stat(sc, cmd);
			return;
		}

		// udelay(125);		// usb-storage has this
		ub_data_start(sc, cmd);

	} else if (cmd->state == UB_CMDST_DATA) {
		if (urb->status == -EPIPE) {
			rc = ub_submit_clear_stall(sc, cmd, sc->last_pipe);
			if (rc != 0) {
				printk(KERN_NOTICE "%s: "
				    "unable to submit clear (%d)\n",
				    sc->name, rc);
				ub_state_done(sc, cmd, rc);
				return;
			}
			cmd->state = UB_CMDST_CLR2STS;
			return;
		}
		if (urb->status == -EOVERFLOW) {
			/*
			 * A babble? Failure, but we must transfer CSW now.
			 */
			cmd->error = -EOVERFLOW;	/* A cheap trick... */
			ub_state_stat(sc, cmd);
			return;
		}

		if (cmd->dir == UB_DIR_WRITE) {
			/*
			 * Do not continue writes in case of a failure.
			 * Doing so would cause sectors to be mixed up,
			 * which is worse than sectors lost.
			 *
			 * We must try to read the CSW, or many devices
			 * get confused.
			 */
			len = urb->actual_length;
			if (urb->status != 0 ||
			    len != cmd->sgv[cmd->current_sg].length) {
				cmd->act_len += len;

				cmd->error = -EIO;
				ub_state_stat(sc, cmd);
				return;
			}

		} else {
			/*
			 * If an error occurs on read, we record it, and
			 * continue to fetch data in order to avoid bubble.
			 *
			 * As a small shortcut, we stop if we detect that
			 * a CSW mixed into data.
			 */
			if (urb->status != 0)
				cmd->error = -EIO;

			len = urb->actual_length;
			if (urb->status != 0 ||
			    len != cmd->sgv[cmd->current_sg].length) {
				if ((len & 0x1FF) == US_BULK_CS_WRAP_LEN)
					goto Bad_End;
			}
		}

		cmd->act_len += urb->actual_length;

		if (++cmd->current_sg < cmd->nsg) {
			ub_data_start(sc, cmd);
			return;
		}
		ub_state_stat(sc, cmd);

	} else if (cmd->state == UB_CMDST_STAT) {
		if (urb->status == -EPIPE) {
			rc = ub_submit_clear_stall(sc, cmd, sc->last_pipe);
			if (rc != 0) {
				printk(KERN_NOTICE "%s: "
				    "unable to submit clear (%d)\n",
				    sc->name, rc);
				ub_state_done(sc, cmd, rc);
				return;
			}


			cmd->error = -EIO;		/* A cheap trick... */

			cmd->state = UB_CMDST_CLRRS;
			return;
		}

		/* Catch everything, including -EOVERFLOW and other nasties. */
		if (urb->status != 0)
			goto Bad_End;

		if (urb->actual_length == 0) {
			ub_state_stat_counted(sc, cmd);
			return;
		}

		/*
		 * Check the returned Bulk protocol status.
		 * The status block has to be validated first.
		 */

		bcs = &sc->work_bcs;

		if (sc->signature == cpu_to_le32(0)) {
			/*
			 * This is the first reply, so do not perform the check.
			 * Instead, remember the signature the device uses
			 * for future checks. But do not allow a nul.
			 */
			sc->signature = bcs->Signature;
			if (sc->signature == cpu_to_le32(0)) {
				ub_state_stat_counted(sc, cmd);
				return;
			}
		} else {
			if (bcs->Signature != sc->signature) {
				ub_state_stat_counted(sc, cmd);
				return;
			}
		}

		if (bcs->Tag != cmd->tag) {
			/*
			 * This usually happens when we disagree with the
			 * device's microcode about something. For instance,
			 * a few of them throw this after timeouts. They buffer
			 * commands and reply at commands we timed out before.
			 * Without flushing these replies we loop forever.
			 */
			ub_state_stat_counted(sc, cmd);
			return;
		}

		len = le32_to_cpu(bcs->Residue);
		if (len != cmd->len - cmd->act_len) {
			/*
			 * It is all right to transfer less, the caller has
			 * to check. But it's not all right if the device
			 * counts disagree with our counts.
			 */
			goto Bad_End;
		}

		switch (bcs->Status) {
		case US_BULK_STAT_OK:
			break;
		case US_BULK_STAT_FAIL:
			ub_state_sense(sc, cmd);
			return;
		case US_BULK_STAT_PHASE:
			goto Bad_End;
		default:
			printk(KERN_INFO "%s: unknown CSW status 0x%x\n",
			    sc->name, bcs->Status);
			ub_state_done(sc, cmd, -EINVAL);
			return;
		}

		/* Not zeroing error to preserve a babble indicator */
		if (cmd->error != 0) {
			ub_state_sense(sc, cmd);
			return;
		}
		cmd->state = UB_CMDST_DONE;
		ub_cmdq_pop(sc);
		(*cmd->done)(sc, cmd);

	} else if (cmd->state == UB_CMDST_SENSE) {
		ub_state_done(sc, cmd, -EIO);

	} else {
		printk(KERN_WARNING "%s: "
		    "wrong command state %d\n",
		    sc->name, cmd->state);
		ub_state_done(sc, cmd, -EINVAL);
		return;
	}
	return;

Bad_End: /* Little Excel is dead */
	ub_state_done(sc, cmd, -EIO);
}

/*
 * Factorization helper for the command state machine:
 * Initiate a data segment transfer.
 */
static void ub_data_start(struct ub_dev *sc, struct ub_scsi_cmd *cmd)
{
	struct scatterlist *sg = &cmd->sgv[cmd->current_sg];
	int pipe;
	int rc;

	/* 
	 * If we have dma_address, just use it, and set URB_NO_TRANSFER_DMA_MAP
	 * flag. This flag must be clean as soon as command is done (mostly in
	 * the function that complete_fn points to).
	*/
	void *bufaddr = NULL;
	int buflen = sg->length;
	if (sg->dma_address) {
		/* use physical address directly */
		sc->work_urb.transfer_dma = sg->dma_address;
		sc->work_urb.transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
	} else {
		bufaddr = sg_virt(sg),
		sc->work_urb.transfer_flags &= ~URB_NO_TRANSFER_DMA_MAP;
	}


	UB_INIT_COMPLETION(sc->work_done);

	if (cmd->dir == UB_DIR_READ)
		pipe = sc->recv_bulk_pipe;
	else
		pipe = sc->send_bulk_pipe;
	sc->last_pipe = pipe;
	usb_fill_bulk_urb(&sc->work_urb, sc->dev, pipe,
	bufaddr, buflen,
	ub_urb_complete, sc);
	sc->work_urb.actual_length = 0;
	sc->work_urb.error_count = 0;
	sc->work_urb.status = 0;

	if ((rc = usb_submit_urb(&sc->work_urb, GFP_ATOMIC)) != 0) {
		/* XXX Clear stalls */
		ub_complete(&sc->work_done);
		ub_state_done(sc, cmd, rc);
		return;
	}

	sc->work_timer.expires = jiffies + UB_DATA_TIMEOUT;
	add_timer(&sc->work_timer);

	cmd->state = UB_CMDST_DATA;
}

/*
 * Factorization helper for the command state machine:
 * Finish the command.
 */
static void ub_state_done(struct ub_dev *sc, struct ub_scsi_cmd *cmd, int rc)
{

	cmd->error = rc;
	cmd->state = UB_CMDST_DONE;
	ub_cmdq_pop(sc);
	(*cmd->done)(sc, cmd);
}

/*
 * Factorization helper for the command state machine:
 * Submit a CSW read.
 */
static int __ub_state_stat(struct ub_dev *sc, struct ub_scsi_cmd *cmd)
{
	int rc;

	UB_INIT_COMPLETION(sc->work_done);

	sc->last_pipe = sc->recv_bulk_pipe;
	usb_fill_bulk_urb(&sc->work_urb, sc->dev, sc->recv_bulk_pipe,
	    &sc->work_bcs, US_BULK_CS_WRAP_LEN, ub_urb_complete, sc);
	sc->work_urb.actual_length = 0;
	sc->work_urb.error_count = 0;
	sc->work_urb.status = 0;

	if ((rc = usb_submit_urb(&sc->work_urb, GFP_ATOMIC)) != 0) {
		/* XXX Clear stalls */
		ub_complete(&sc->work_done);
		ub_state_done(sc, cmd, rc);
		return -1;
	}

	sc->work_timer.expires = jiffies + UB_STAT_TIMEOUT;
	add_timer(&sc->work_timer);
	return 0;
}

/*
 * Factorization helper for the command state machine:
 * Submit a CSW read and go to STAT state.
 */
static void ub_state_stat(struct ub_dev *sc, struct ub_scsi_cmd *cmd)
{

	if (__ub_state_stat(sc, cmd) != 0)
		return;

	cmd->stat_count = 0;
	cmd->state = UB_CMDST_STAT;
}

/*
 * Factorization helper for the command state machine:
 * Submit a CSW read and go to STAT state with counter (along [C] path).
 */
static void ub_state_stat_counted(struct ub_dev *sc, struct ub_scsi_cmd *cmd)
{

	if (++cmd->stat_count >= 4) {
		ub_state_sense(sc, cmd);
		return;
	}

	if (__ub_state_stat(sc, cmd) != 0)
		return;

	cmd->state = UB_CMDST_STAT;
}

/*
 * Factorization helper for the command state machine:
 * Submit a REQUEST SENSE and go to SENSE state.
 */
static void ub_state_sense(struct ub_dev *sc, struct ub_scsi_cmd *cmd)
{
	struct ub_scsi_cmd *scmd;
	struct scatterlist *sg;
	int rc;

	if (cmd->cdb[0] == REQUEST_SENSE) {
		rc = -EPIPE;
		goto error;
	}

	scmd = &sc->top_rqs_cmd;
	memset(scmd, 0, sizeof(struct ub_scsi_cmd));
	scmd->cdb[0] = REQUEST_SENSE;
	scmd->cdb[4] = UB_SENSE_SIZE;
	scmd->cdb_len = 6;
	scmd->dir = UB_DIR_READ;
	scmd->state = UB_CMDST_INIT;
	scmd->nsg = 1;
	sg = &scmd->sgv[0];
	sg_init_table(sg, UB_MAX_REQ_SG);
	sg_set_page(sg, virt_to_page(sc->top_sense), UB_SENSE_SIZE,
			(unsigned long)sc->top_sense & (PAGE_SIZE-1));
	scmd->len = UB_SENSE_SIZE;
	scmd->lun = cmd->lun;
	scmd->done = ub_top_sense_done;
	scmd->back = cmd;

	scmd->tag = sc->tagcnt++;

	cmd->state = UB_CMDST_SENSE;

	ub_cmdq_insert(sc, scmd);
	return;

error:
	ub_state_done(sc, cmd, rc);
}

/*
 * A helper for the command's state machine:
 * Submit a stall clear.
 */
static int ub_submit_clear_stall(struct ub_dev *sc, struct ub_scsi_cmd *cmd,
    int stalled_pipe)
{
	int endp;
	struct usb_ctrlrequest *cr;
	int rc;

	endp = usb_pipeendpoint(stalled_pipe);
	if (usb_pipein (stalled_pipe))
		endp |= USB_DIR_IN;

	cr = &sc->work_cr;
	cr->bRequestType = USB_RECIP_ENDPOINT;
	cr->bRequest = USB_REQ_CLEAR_FEATURE;
	cr->wValue = cpu_to_le16(USB_ENDPOINT_HALT);
	cr->wIndex = cpu_to_le16(endp);
	cr->wLength = cpu_to_le16(0);

	UB_INIT_COMPLETION(sc->work_done);

	usb_fill_control_urb(&sc->work_urb, sc->dev, sc->send_ctrl_pipe,
	    (unsigned char*) cr, NULL, 0, ub_urb_complete, sc);
	sc->work_urb.actual_length = 0;
	sc->work_urb.error_count = 0;
	sc->work_urb.status = 0;

	if ((rc = usb_submit_urb(&sc->work_urb, GFP_ATOMIC)) != 0) {
		ub_complete(&sc->work_done);
		return rc;
	}

	sc->work_timer.expires = jiffies + UB_CTRL_TIMEOUT;
	add_timer(&sc->work_timer);
	return 0;
}

/*
 */
static void ub_top_sense_done(struct ub_dev *sc, struct ub_scsi_cmd *scmd)
{
	unsigned char *sense = sc->top_sense;
	struct ub_scsi_cmd *cmd;

	/*
	 * Find the command which triggered the unit attention or a check,
	 * save the sense into it, and advance its state machine.
	 */
	if ((cmd = ub_cmdq_peek(sc)) == NULL) {
		printk(KERN_WARNING "%s: sense done while idle\n", sc->name);
		return;
	}
	if (cmd != scmd->back) {
		printk(KERN_WARNING "%s: "
		    "sense done for wrong command 0x%x\n",
		    sc->name, cmd->tag);
		return;
	}
	if (cmd->state != UB_CMDST_SENSE) {
		printk(KERN_WARNING "%s: "
		    "sense done with bad cmd state %d\n",
		    sc->name, cmd->state);
		return;
	}

	/*
	 * Ignoring scmd->act_len, because the buffer was pre-zeroed.
	 */
	cmd->key = sense[2] & 0x0F;
	cmd->asc = sense[12];
	cmd->ascq = sense[13];

	ub_scsi_urb_compl(sc, cmd);
}


/*
 * This is called from a process context.
 */
static void ub_revalidate(struct ub_dev *sc, struct ub_lun *lun)
{

	lun->readonly = 0;	/* XXX Query this from the device */

	lun->capacity.nsec = 0;
	lun->capacity.bsize = 512;
	lun->capacity.bshift = 0;

	if (ub_sync_tur(sc, lun) != 0)
		return;			/* Not ready */
	lun->changed = 0;

	if (ub_sync_read_cap(sc, lun, &lun->capacity) != 0) {
		/*
		 * The retry here means something is wrong, either with the
		 * device, with the transport, or with our code.
		 * We keep this because sd.c has retries for capacity.
		 */
		if (ub_sync_read_cap(sc, lun, &lun->capacity) != 0) {
			lun->capacity.nsec = 0;
			lun->capacity.bsize = 512;
			lun->capacity.bshift = 0;
		}
	}
}


/*
 * Common ->done routine for commands executed synchronously.
 */
static void ub_probe_done(struct ub_dev *sc, struct ub_scsi_cmd *cmd)
{
	struct completion *cop = cmd->back;
	complete(cop);

	if (cmd->lun) {
		aldControl(cmd->lun->id, 0, 1);
	}
}

/*
 * Test if the device has a check condition on it, synchronously.
 */
static int ub_sync_tur(struct ub_dev *sc, struct ub_lun *lun)
{
	struct ub_scsi_cmd *cmd;
	enum { ALLOC_SIZE = sizeof(struct ub_scsi_cmd) };
	unsigned long flags;
	struct completion compl;
	int rc;

	init_completion(&compl);

	rc = -ENOMEM;
	if ((cmd = kzalloc(ALLOC_SIZE, GFP_KERNEL)) == NULL)
		goto err_alloc;

	cmd->cdb[0] = TEST_UNIT_READY;
	cmd->cdb_len = 6;
	cmd->dir = UB_DIR_NONE;
	cmd->state = UB_CMDST_INIT;
	cmd->lun = lun;			/* This may be NULL, but that's ok */
	cmd->done = ub_probe_done;
	cmd->back = &compl;

	spin_lock_irqsave(sc->lock, flags);
	cmd->tag = sc->tagcnt++;


	rc = umsc_sync_submit_scsi(sc, cmd);
	spin_unlock_irqrestore(sc->lock, flags);

	if (rc != 0)
		goto err_submit;

	if (wait_for_completion_interruptible(&compl) < 0) {
		umsc_errorl("wake up by signal!");
	}

	rc = cmd->error;

	if (rc == -EIO && cmd->key != 0)	/* Retries for benh's key */
		rc = cmd->key;

err_submit:
	kfree(cmd);
err_alloc:
	return rc;
}

/*
 * Read the SCSI capacity synchronously (for probing).
 */
static int ub_sync_read_cap(struct ub_dev *sc, struct ub_lun *lun,
    struct ub_capacity *ret)
{
	struct ub_scsi_cmd *cmd;
	struct scatterlist *sg;
	char *p;
	enum { ALLOC_SIZE = sizeof(struct ub_scsi_cmd) + 8 };
	unsigned long flags;
	unsigned int bsize, shift;
	unsigned long nsec;
	struct completion compl;
	int rc;

	init_completion(&compl);

	rc = -ENOMEM;
	if ((cmd = kzalloc(ALLOC_SIZE, GFP_KERNEL)) == NULL)
		goto err_alloc;
	p = (char *)cmd + sizeof(struct ub_scsi_cmd);

	cmd->cdb[0] = READ_CAPACITY;//READ_CAPACITY is 0x25;
	cmd->cdb_len = 10;
	cmd->dir = UB_DIR_READ;
	cmd->state = UB_CMDST_INIT;
	cmd->nsg = 1;
	sg = &cmd->sgv[0];
	sg_init_table(sg, UB_MAX_REQ_SG);
	sg_set_page(sg, virt_to_page(p), 8, (unsigned long)p & (PAGE_SIZE-1));
	cmd->len = 8;
	cmd->lun = lun;
	cmd->done = ub_probe_done;
	cmd->back = &compl;

	spin_lock_irqsave(sc->lock, flags);
	cmd->tag = sc->tagcnt++;

	rc = umsc_sync_submit_scsi(sc, cmd);
	spin_unlock_irqrestore(sc->lock, flags);

	if (rc != 0)
		goto err_submit;

	if (wait_for_completion_interruptible(&compl) < 0) {
		umsc_errorl("wake up by signal!");
	}

	if (cmd->error != 0) {
		rc = -EIO;
		goto err_read;
	}
	if (cmd->act_len != 8) {
		rc = -EIO;
		goto err_read;
	}

	/* sd.c special-cases sector size of 0 to mean 512. Needed? Safe? */
	nsec = be32_to_cpu(*(__be32 *)p) + 1;
	bsize = be32_to_cpu(*(__be32 *)(p + 4));
	switch (bsize) {
	case 512:	shift = 0;	break;
	case 1024:	shift = 1;	break;
	case 2048:	shift = 2;	break;
	case 4096:	shift = 3;	break;
	default:
		rc = -EDOM;
		goto err_inv_bsize;
	}

	ret->bsize = bsize;
	ret->bshift = shift;
	ret->nsec = nsec << shift;
	rc = 0;

err_inv_bsize:
err_read:
err_submit:
	kfree(cmd);
err_alloc:
	return rc;
}

/*
 */
static void ub_probe_urb_complete(struct urb *urb)
{
	struct completion *cop = urb->context;
	complete(cop);
}

static void ub_probe_timeout(unsigned long arg)
{
	struct completion *cop = (struct completion *) arg;
	complete(cop);
}



#if 1
/*
 * Get number of LUNs by the way of Bulk GetMaxLUN command.
 */
static int ub_sync_getmaxlun(struct ub_dev *sc)
{
	int ifnum = sc->intf->cur_altsetting->desc.bInterfaceNumber;
	unsigned char *p;
	enum { ALLOC_SIZE = 1 };
	struct usb_ctrlrequest *cr;
	struct completion compl;
	struct timer_list timer;
	int nluns;
	int rc;

	init_completion(&compl);

	rc = -ENOMEM;
	if ((p = kmalloc(ALLOC_SIZE, GFP_KERNEL)) == NULL)
		goto err_alloc;
	*p = 55;

	cr = &sc->work_cr;
	cr->bRequestType = USB_DIR_IN | USB_TYPE_CLASS | USB_RECIP_INTERFACE;
	cr->bRequest = US_BULK_GET_MAX_LUN;
	cr->wValue = cpu_to_le16(0);
	cr->wIndex = cpu_to_le16(ifnum);
	cr->wLength = cpu_to_le16(1);

	usb_fill_control_urb(&sc->work_urb, sc->dev, sc->recv_ctrl_pipe,
	    (unsigned char*) cr, p, 1, ub_probe_urb_complete, &compl);
	sc->work_urb.actual_length = 0;
	sc->work_urb.error_count = 0;
	sc->work_urb.status = 0;

	if ((rc = usb_submit_urb(&sc->work_urb, GFP_KERNEL)) != 0)
		goto err_submit;

	init_timer(&timer);
	timer.function = ub_probe_timeout;
	timer.data = (unsigned long) &compl;
	timer.expires = jiffies + UB_CTRL_TIMEOUT;
	add_timer(&timer);

	if (wait_for_completion_interruptible(&compl) < 0) {
		umsc_errorl("wake up by signal!");
	}

	del_timer_sync(&timer);
	usb_kill_urb(&sc->work_urb);

	if ((rc = sc->work_urb.status) < 0)
		goto err_io;

	if (sc->work_urb.actual_length != 1) {
		nluns = 0;
	} else {
		if ((nluns = *p) == 55) {
			nluns = 0;
		} else {
  			/* GetMaxLUN returns the maximum LUN number */
			nluns += 1;
			if (nluns > UB_MAX_LUNS)
				nluns = UB_MAX_LUNS;
		}
	}

	kfree(p);

	sc->maxlun = nluns;

	return nluns;

err_io:
err_submit:
	kfree(p);
err_alloc:
	return rc;
}
#endif



/*
 * Get the pipe settings.
 */
static int ub_get_pipes(struct ub_dev *sc, struct usb_device *dev,
    struct usb_interface *intf)
{
	struct usb_host_interface *altsetting = intf->cur_altsetting;
	struct usb_endpoint_descriptor *ep_in = NULL;
	struct usb_endpoint_descriptor *ep_out = NULL;
	struct usb_endpoint_descriptor *ep;
	int i;

	/*
	 * Find the endpoints we need.
	 * We are expecting a minimum of 2 endpoints - in and out (bulk).
	 * We will ignore any others.
	 */
	for (i = 0; i < altsetting->desc.bNumEndpoints; i++) {
		ep = &altsetting->endpoint[i].desc;

		/* Is it a BULK endpoint? */
		if ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
				== USB_ENDPOINT_XFER_BULK) {
			/* BULK in or out? */
			if (ep->bEndpointAddress & USB_DIR_IN) {
				if (ep_in == NULL)
					ep_in = ep;
			} else {
				if (ep_out == NULL)
					ep_out = ep;
			}
		}
	}

	if (ep_in == NULL || ep_out == NULL) {
		printk(KERN_NOTICE "%s: failed endpoint check\n",
		    sc->name);
		return -ENODEV;
	}

	/* Calculate and store the pipe values */
	sc->send_ctrl_pipe = usb_sndctrlpipe(dev, 0);
	sc->recv_ctrl_pipe = usb_rcvctrlpipe(dev, 0);
	sc->send_bulk_pipe = usb_sndbulkpipe(dev,
		ep_out->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
	sc->recv_bulk_pipe = usb_rcvbulkpipe(dev,
		ep_in->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);

	return 0;
}

/*
 * Probing is done in the process context, which allows us to cheat
 * and not to build a state machine for the discovery.
 */
static int ub_probe(struct usb_interface *intf,
    const struct usb_device_id *dev_id)
{
	struct ub_dev *sc;
	int rc;
	int i;



	/* :
	 * This module only support USB ExpressCard in port 1 and 2 of
	 * level 1.
	*/
	do {
		struct usb_device * pud = interface_to_usbdev(intf);
		if (-1==umsc_id_check(pud->level, pud->portnum)) {
			return -1;
		}
	} while (0);
	/* 
	 * Reset device is needed because there is a problem if we do
	 * `rmmod umscdev.ko && insmod umscdev.ko` without physical
	 * disconnect.
	 * This approach is more effective than ub_probe_clear_stall(). See
	 * the code block below.
	*/
	if (usb_reset_device(interface_to_usbdev(intf))) {
		umsc_errorl("usb_reset_device() failed!");
	}


	rc = -ENOMEM;
	if ((sc = kzalloc(sizeof(struct ub_dev), GFP_KERNEL)) == NULL)
		goto err_core;
	sc->lock = ub_next_lock();
	INIT_LIST_HEAD(&sc->luns);
	usb_init_urb(&sc->work_urb);
	tasklet_init(&sc->tasklet, ub_scsi_action, (unsigned long)sc);
	atomic_set(&sc->poison, 0);


	init_timer(&sc->work_timer);
	sc->work_timer.data = (unsigned long) sc;
	sc->work_timer.function = ub_urb_timeout;

	ub_init_completion(&sc->work_done);
	sc->work_done.done = 1;		/* A little yuk, but oh well... */

	sc->dev = interface_to_usbdev(intf);
	sc->intf = intf;
	// sc->ifnum = intf->cur_altsetting->desc.bInterfaceNumber;
	usb_set_intfdata(intf, sc);
	usb_get_dev(sc->dev);
	/*
	 * Since we give the interface struct to the block level through
	 * disk->driverfs_dev, we have to pin it. Otherwise, block_uevent
	 * oopses on close after a disconnect (kernels 2.6.16 and up).
	 */
	usb_get_intf(sc->intf);

	snprintf(sc->name, 12, DRV_NAME "(%d.%d)",
	    sc->dev->bus->busnum, sc->dev->devnum);

	/* XXX Verify that we can handle the device (from descriptors) */

	if (ub_get_pipes(sc, sc->dev, intf) != 0)
		goto err_dev_desc;

	/*
	 * At this point, all USB initialization is done, do upper layer.
	 * We really hate halfway initialized structures, so from the
	 * invariants perspective, this ub_dev is fully constructed at
	 * this point.
	 */

	/*
	 * This is needed to clear toggles. It is a problem only if we do
	 * `rmmod ub && modprobe ub` without disconnects, but we like that.
	 */


	/*
	 * The way this is used by the startup code is a little specific.
	 * A SCSI check causes a USB stall. Our common case code sees it
	 * and clears the check, after which the device is ready for use.
	 * But if a check was not present, any command other than
	 * TEST_UNIT_READY ends with a lockup (including REQUEST_SENSE).
	 *
	 * If we neglect to clear the SCSI check, the first real command fails
	 * (which is the capacity readout). We clear that and retry, but why
	 * causing spurious retries for no reason.
	 *
	 * Revalidation may start with its own TEST_UNIT_READY, but that one
	 * has to succeed, so we clear checks with an additional one here.
	 * In any case it's not our business how revaliadation is implemented.
	 */
	for (i = 0; i < 3; i++) {  /* Retries for the schwag key from KS'04 */
		if ((rc = ub_sync_tur(sc, NULL)) <= 0) break;
		if (rc != 0x6) break;
		msleep(10);
	}


	sc->maxlun = ub_sync_getmaxlun(sc);
	if ((rc = ub_probe_lun(sc, 0)) != 0) {
		umsc_errorl("ub_probe_lun() failed !!");
		goto err_dev_desc;
	}

	return 0;

err_dev_desc:
	usb_set_intfdata(intf, NULL);
	usb_put_intf(sc->intf);
	usb_put_dev(sc->dev);
	kfree(sc);
err_core:
	return rc;
}

#include "umsc.h"

static int ub_probe_lun(struct ub_dev *sc, int lnum)
{
	struct ub_lun *lun;

	int rc;

	rc = -ENOMEM;
	if ((lun = kzalloc(sizeof(struct ub_lun), GFP_KERNEL)) == NULL)
		goto err_alloc;
	lun->num = lnum;

	rc = -ENOSR;


	/* print device generic infomation */
	umsc_debugl("USB Dev Level=%d, Port=%d", sc->dev->level, sc->dev->portnum);
	umsc_debugl("USB Dev devnum=%d", sc->dev->devnum);
	umsc_debugl("USB Dev state=%d", sc->dev->state);
	umsc_debugl("USB Dev speed=%d", sc->dev->speed);
	umsc_debugl("USB Dev product=%s", sc->dev->product);
	umsc_debugl("USB Dev manufacturer=%s", sc->dev->manufacturer);
	umsc_debugl("USB Dev serial=%s", sc->dev->serial);
	umsc_debugl("USB Dev maxlun=%d", sc->maxlun);

	{//just print the parents
		struct usb_device *ud = sc->dev->parent;
		for (; ud; ud = ud->parent) {
			umsc_debugl("Parent: Level=%d, Port=%d", ud->level, ud->portnum);
		}
	}

	if ((lun->id = umsc_id_get(sc->dev->level, sc->dev->portnum, sc)) == -1) {
		umsc_errorl("Get minor failed! USB Dev Level=%d, Port=%d", sc->dev->level, sc->dev->portnum);
		goto err_id;
	}
	umsc_debugl("USB Dev minor=%d", lun->id);

	lun->udev = sc;


	snprintf(lun->name, 16, DRV_NAME "%d(%d.%d.%d)",
	    lun->id, sc->dev->bus->busnum, sc->dev->devnum, lun->num);
	umsc_debugl("lun->name = %s", lun->name);

	lun->removable = 1;		/* XXX Query this from the device */
	lun->changed = 1;		/* ub_revalidate clears only */
	ub_revalidate(sc, lun);

	umsc_infol("Slot[%d]: nsec=0x%lx, bsize=0x%x, bshift=%d",
			lun->id, lun->capacity.nsec,
			lun->capacity.bsize, lun->capacity.bshift);


	
	list_add(&lun->link, &sc->luns);

	/* light the led */
	aldControl(lun->id, 0, 1);

	return 0;



err_id:
	kfree(lun);
err_alloc:
	return rc;
}

static void ub_disconnect(struct usb_interface *intf)
{
	struct ub_dev *sc = usb_get_intfdata(intf);
	struct ub_lun *lun;
	unsigned long flags;

	lun = umsc_dev_to_lun(sc);

	/* shutdown the led */
	aldControl(lun->id, 0, 0);


	/*
	 * Prevent ub_bd_release from pulling the rug from under us.
	 * XXX This is starting to look like a kref.
	 * XXX Why not to take this ref at probe time?
	 */
	spin_lock_irqsave(&ub_lock, flags);
	sc->openc++;
	spin_unlock_irqrestore(&ub_lock, flags);

	/*
	 * Fence stall clearnings, operations triggered by unlinkings and so on.
	 * We do not attempt to unlink any URBs, because we do not trust the
	 * unlink paths in HC drivers. Also, we get -84 upon disconnect anyway.
	 */
	atomic_set(&sc->poison, 1);


	/*
	 * Blow away queued commands.
	 *
	 * Actually, this never works, because before we get here
	 * the HCD terminates outstanding URB(s). It causes our
	 * SCSI command queue to advance, commands fail to submit,
	 * and the whole queue drains. So, we just use this code to
	 * print warnings.
	 */
	spin_lock_irqsave(sc->lock, flags);
	{
		struct ub_scsi_cmd *cmd;
		int cnt = 0;
		while ((cmd = ub_cmdq_peek(sc)) != NULL) {
			cmd->error = -ENOTCONN;
			cmd->state = UB_CMDST_DONE;
			ub_cmdq_pop(sc);
			(*cmd->done)(sc, cmd);
			cnt++;
		}
		if (cnt != 0) {
			printk(KERN_WARNING "%s: "
			    "%d was queued after shutdown\n", sc->name, cnt);
		}
	}
	spin_unlock_irqrestore(sc->lock, flags);

	/*
	 * Unregister the upper layer.
	 */


	/*
	 * Testing for -EINPROGRESS is always a bug, so we are bending
	 * the rules a little.
	 */
	spin_lock_irqsave(sc->lock, flags);
	if (sc->work_urb.status == -EINPROGRESS) {	/* janitors: ignore */
		printk(KERN_WARNING "%s: "
		    "URB is active after disconnect\n", sc->name);
	}
	spin_unlock_irqrestore(sc->lock, flags);

	/*
	 * There is virtually no chance that other CPU runs times so long
	 * after ub_urb_complete should have called del_timer, but only if HCD
	 * didn't forget to deliver a callback on unlink.
	 */
	del_timer_sync(&sc->work_timer);

	/*
	 * At this point there must be no commands coming from anyone
	 * and no URBs left in transit.
	 */

	ub_put(sc);
}

static struct usb_driver ub_driver = {

	.name = 	"umsc",
	.probe =	ub_probe,
	.disconnect =	ub_disconnect,
	.id_table =	ub_usb_ids,
};

#if 1
static int umsc_chardev_open(struct inode * node, struct file * f)
{
	int minor = MINOR(node->i_rdev);
	struct umsc_chardev * umsc = NULL;
	struct ub_dev * sc = NULL;
	struct ub_lun * lun = NULL;

	umsc = inode_to_umsc(node);

	/* sync for usb probe */
	if (expcd_slot_check(minor, 0) && expcd_slot_get_power(minor)) {
		int slp_cnt;
		for (slp_cnt=0; slp_cnt<40; slp_cnt++) {
			sc = umsc->udevs[minor];
			if (sc) {
				break;
			} else {
				msleep(100);
			}
		}
	}

	if (sc) {
		lun = umsc_dev_to_lun(sc);
	}
	if (lun){
		f->private_data = sc; /* so this sc must has a LUN */
		umsc_infol("");

	} else {
		umsc_debugl("No device found in Slot %d", minor);
		return -ENXIO;
	}
	return 0;
}

static int umsc_chardev_release(struct inode * node, struct file * f)
{
	umsc_infol("");
	f->private_data = NULL;
	return 0;
}

static int umsc_chardev_ioctl(struct inode * node, struct file * f,
					unsigned int cmd, unsigned long arg)
{
	struct ub_dev * sc;
	struct ub_lun * lun;
	int ret = 0;

	sc = (struct ub_dev *)f->private_data;
	if (!sc) { return -ENXIO; }
	lun = umsc_dev_to_lun(sc);
	if (!lun) { return -ENXIO; }

	switch (cmd) {
	case UMSC_IOCTL_GETBASEINFO:
	{
		struct usbcard_stinfo stinfo = {
			sizeof(struct usbcard_info),
			sizeof(struct usbcard_cmd),
			sizeof(struct usbcard_cap),
		};
		ret = copy_to_user((void __user *)arg, (const void *)&stinfo, sizeof(struct usbcard_stinfo));
		break;
	}
	case UMSC_IOCTL_GETDEVINFO:
	{
		struct usbcard_info ucdinfo;
		ucdinfo.devnum = sc->dev->devnum;
		ucdinfo.portnum = sc->dev->portnum;
		ucdinfo.level = sc->dev->level;
		ucdinfo.state = (unsigned char)sc->dev->state;
		ucdinfo.speed = (unsigned char)sc->dev->speed;
		memcpy(ucdinfo.product, sc->dev->product, UMSC_STR_SIZE);
		memcpy(ucdinfo.manufacturer, sc->dev->manufacturer, UMSC_STR_SIZE);
		memcpy(ucdinfo.serial, sc->dev->serial, UMSC_STR_SIZE);
		memcpy(&ucdinfo.desc_dev, &sc->dev->descriptor, sizeof(struct usb_device_descriptor));
		memcpy(&ucdinfo.desc_cfg, &sc->dev->actconfig->desc, sizeof(struct usb_config_descriptor));
		memcpy(&ucdinfo.desc_inf, &sc->intf->cur_altsetting->desc, sizeof(struct usb_interface_descriptor));
		ucdinfo.maxlun = sc->maxlun;
		ucdinfo.luns[0].nsec = lun->capacity.nsec;
		ucdinfo.luns[0].bsize = lun->capacity.bsize;
		ucdinfo.luns[0].bshift = lun->capacity.bshift;

		ret = copy_to_user((void __user *)arg, &ucdinfo, sizeof(struct usbcard_info));
		break;
	}
	case UMSC_IOCTL_GETSHMEMLEN:
	{
		int __user * plen = (int *)arg;
		ret = copy_to_user(plen, &umsc_shmem_len, sizeof(umsc_shmem_len));
		break;
	}
	case UMSC_IOCTL_GETSHMEMPHY:
	{
		unsigned long phy_shmem = __pa(umsc_shmem);
		ret = copy_to_user((void __user *)arg, &phy_shmem, sizeof(unsigned long));
		break;
	}
	case UMSC_IOCTL_GETCAPACITY:
	{
		struct ub_capacity * __user pcap = (struct ub_capacity *)arg;
		ret = copy_to_user(pcap, &lun->capacity, sizeof(struct ub_capacity));
		break;
	}
	case UMSC_IOCTL_CMD:
	{
		struct usbcard_cmd uc;
		if (0==arg) { return -EINVAL; }

		/* copy user space command struct */
		if (copy_from_user(&uc, (const void __user *)arg, sizeof(struct usbcard_cmd))) {
			return -EFAULT;
		}

		/* process the user space command */
		ret = umsc_process_user_cmd(f->private_data, &uc);
		if (ret) { break; }

		/* return the command result */
		ret = copy_to_user((void __user *)arg, &uc, sizeof(struct usbcard_cmd));

		break;
	}
	default:
		break;
	}
	return ret;
}

static int umsc_chardev_mmap(struct file *f, struct vm_area_struct *vma)
{
	/* calculate physical memory page frame number */
	unsigned long off = vma->vm_pgoff << PAGE_SHIFT; /* virt mem offset */
	unsigned long vsz = vma->vm_end - vma->vm_start; /* virt mem size */
	unsigned long phy = __pa(umsc_shmem);            /* shared mem phys addr */
	unsigned long pfn = (phy + off) >> PAGE_SHIFT;   /* phys page frame number */
	unsigned long psz = umsc_shmem_len - off;        /* phys mem size */

	umsc_debugl("shmem=0x%08lx, phyaddr=0x%08lx, len=%d", (unsigned long)umsc_shmem, phy, umsc_shmem_len);
	umsc_debugl("off=0x%08lx, vsize=%lu; pfn=0x%05lx, psize=%lu", off, vsz, pfn, psz);

        /* check the request-map-size. */
	if (off > psz || vsz > psz || (off+vsz) > psz) {
		umsc_errorl("off=0x%08lx, vsize=%lu; phyaddr=0x%08lx, phylen=%u", off, vsz, phy, umsc_shmem_len);
		return -EINVAL;
	}

	/* set map flags */
        vma->vm_flags |= VM_LOCKED;

        /* set map protection, make it uncacheable */
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

        /* map physical memory (page frame number) to user space */
	remap_pfn_range(vma, vma->vm_start, pfn, vsz, vma->vm_page_prot);
	return 0;
}

static int umsc_major = CXD9225GG_MAJOR_ID_UMSC;
static struct file_operations umsc_fops ={
	.owner   = THIS_MODULE,
	.open    = umsc_chardev_open,
	.release = umsc_chardev_release,
	.llseek  = generic_file_llseek,
	.mmap    = umsc_chardev_mmap,
	.ioctl	 = umsc_chardev_ioctl,
};

int umsc_cdev_register(void)
{
	int ret = -1;
	dev_t dev_id;

	/* register and init char device */
	dev_id = MKDEV(umsc_major, 0);
	ret = register_chrdev_region(dev_id, UMSC_MINORS, DRV_NAME);
	if (ret) {
		umsc_errorl("register_chrdev_region() failed !!");
		return ret;
	}
	cdev_init(&umsc_cdev.chardev, &umsc_fops);
	ret = cdev_add(&umsc_cdev.chardev, dev_id, UMSC_MINORS);
	if (ret) {
		umsc_errorl("cdev_add() failed !!");
		return ret;
	}

	return 0;
}

void umsc_cdev_unregister(void)
{
	dev_t dev_id;

	dev_id = MKDEV(umsc_major, 0);
	unregister_chrdev_region(dev_id, UMSC_MINORS);
	cdev_del(&umsc_cdev.chardev);
}
#endif

static int __init ub_init(void)
{
	int rc;
	int i;

	for (i = 0; i < UB_QLOCK_NUM; i++)
		spin_lock_init(&ub_qlockv[i]);



	umsc_config_port_map(0, 1, 1);
	umsc_config_port_map(1, 1, 2);

	umsc_shmem = (void*)__get_free_pages(GFP_KERNEL, get_order(umsc_shmem_len));
	if (!umsc_shmem) {
		rc = -1;
		return rc;
	}


	if ((rc = usb_register(&ub_driver)) != 0) {
		umsc_errorl("usb_register() failed !!");
		return rc;
	}
	if((rc = umsc_cdev_register()) != 0) {
		umsc_errorl("umsc_cdev_register() failed !!");
		return rc;
	}

	return 0;


	return rc;
}

static void __exit ub_exit(void)
{
	umsc_cdev_unregister();


	usb_deregister(&ub_driver);

	free_page((unsigned long)umsc_shmem);



}

module_init(ub_init);
module_exit(ub_exit);

MODULE_LICENSE("GPL");

#endif //CONFIG_BLK_DEV_UB

