/* 2014-08-28: File changed by Sony Corporation */
/*
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/freezer.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/mutex.h>
#include <linux/utsname.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include <linux/blkdev.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <linux/usb_usual.h>
#include <linux/timer.h>
#include <linux/scatterlist.h>
#include <scsi/scsi.h>
#include <mach/cxd9225gg_types.h>

#define US_BULK_CB_WRAP_LEN		31
#define US_BULK_CB_SIGN			0x43425355	/*spells out USBC */
#define US_BULK_FLAG_IN			1
#define US_BULK_FLAG_OUT		0

#define US_BULK_CS_WRAP_LEN		13
#define US_BULK_CS_SIGN			0x53425355	/* spells out 'USBS' */
#define US_BULK_STAT_OK			0
#define US_BULK_STAT_FAIL		1
#define US_BULK_STAT_PHASE		2

/* bulk-only class specific requests */
#define US_BULK_RESET_REQUEST	0xff
#define US_BULK_GET_MAX_LUN		0xfe

/* usb_stor_bulk_transfer_xxx() return codes, in order of severity */
#define USB_STOR_XFER_GOOD		0	/* good transfer                 */
#define USB_STOR_XFER_SHORT		1	/* transferred less than expected */
#define USB_STOR_XFER_STALLED	2	/* endpoint stalled              */
#define USB_STOR_XFER_LONG		3	/* device tried to send too much */
#define USB_STOR_XFER_ERROR		4	/* transfer died in the middle   */

/* Dynamic flag definitions: used in set_bit() etc. */
#define US_FLIDX_URB_ACTIVE		18  /* 0x00040000  current_urb is in use  */
#define US_FLIDX_SG_ACTIVE		19  /* 0x00080000  current_sg is in use   */
#define US_FLIDX_ABORTING		20  /* 0x00100000  abort is in progress   */
#define US_FLIDX_DISCONNECTING	21  /* 0x00200000  disconnect in progress */
#define ABORTING_OR_DISCONNECTING	((1UL << US_FLIDX_ABORTING) | \
									 (1UL << US_FLIDX_DISCONNECTING))
#define US_FLIDX_RESETTING		22  /* 0x00400000  device reset in progress */
#define US_FLIDX_TIMED_OUT		23  /* 0x00800000  SCSI midlayer timed out  */

#define USB_STOR_STRING_LEN		32

/* Transport return codes */
#define USB_STOR_TRANSPORT_GOOD	   	0  /* Transport good, command good	   */
#define USB_STOR_TRANSPORT_FAILED   1  /* Transport good, command failed   */
#define USB_STOR_TRANSPORT_NO_SENSE 2  /* Command failed, no auto-sense    */
#define USB_STOR_TRANSPORT_ERROR    3  /* Transport bad (i.e. device dead) */

#define US_IOBUF_SIZE		64	/* Size of the DMA-mapped I/O buffer */
#define US_SENSE_SIZE		18	/* Size of the autosense data buffer */

#define MAX_COMMAND_SIZE	16

/* return value for loading firmware nandflash	*/
#define USB_FIRMWARE_SUCCESS			0
#define USB_FIRMWARE_WRITE_FAILED		1
#define USB_FIRMWARE_READ_FAILED		2
#define USB_FIRMWARE_CHECK_ERROR		3
#define USB_FIRMWARE_ERASE_FAILED		4
#define USB_FIRMWARE_RENUM_FAILED		5
#define USB_FIRMWARE_BLOCK_FIND_FAILED	6

/* NAND CBW CMD*/
#define SCSI_CMD                0xC8    /* NX2LP Manufacturing/Validation command support */
#define CBW_NAND_READ_PAGE		0x00	/* Read NAND Page */
#define CBW_NAND_WRITE_PAGE		0x01	/* Write NAND page */
#define CBW_NAND_ERASE_BLK 		0x02	/* Erase Single Block */
#define CBW_NAND_ERASE_ALL		0x05	/* Erase All Blocks */
#define CBW_8051_RENUM          0x0b    /* re-enumeration */

/* 2112 bytes*/
#define CBD_DATA_XFER_LENGTH	0x0840

/* Get a minor range for your devices from the usb maintainer */
#define USB_CY_MINOR_BASE		192
#define ECC_NUM					6

/* size defined */
#define BULKOUT_BUFFER_SIZE		64
#define NAND_PAGE_SIZE			2112		/* page size of nandflash */
#define NAND_BLOCK_SIZE			64			/* block size of nandflash */
#define NAND_PAGE_DATA_SIZE 	512

/* Vendor ID and Product ID of CY7C68034 */
#define USB_CY_VENDOR_ID			0x04B4
#define USB_CY_PRODUCT_ID_STAGE1	0x8613
#define USB_CY_PRODUCT_ID_STAGE2	0x4617

#define CY_IOC_MAGIC			'['
#define CY_IOC_FW_DL_OK_TEST 	_IOR(CY_IOC_MAGIC, 1 ,char)
#define CY_IOC_FW_LEN			_IOW(CY_IOC_MAGIC, 2 ,unsigned int)
#define CY_IOC_DLFW_CHECK		_IOW(CY_IOC_MAGIC, 3 ,char)
#define CY_IOC_MAXNR 			3

/* debug log */
#define USB_NFW "usb-nfw: "
//#define CY_DEBUG
#ifdef CY_DEBUG
#define US_DEBUGP(x...)		printk( KERN_ALERT USB_NFW x )
#define US_DEBUGPX(x...)	printk( x )
#define US_DEBUG(x) x
#else
#define US_DEBUGP(x...)
#define US_DEBUGPX(x...)
#define US_DEBUG(x)
#endif

/*
 * Bulk only data structures
 */
/* command block wrapper */
struct bulk_cb_wrap {
	__le32	Signature;			/* contains 'USBC' */
	__u32	Tag;				/* unique per command id */
	__le32	DataTransferLength;	/* size of data */
	__u8	Flags;				/* direction in bit 0 */
	__u8	Lun;				/* LUN normally 0 */
	__u8	Length;				/* of of the CDB */
	__u8	CDB[16];			/* max command */
};

/* command status wrapper */
struct bulk_cs_wrap {
	__le32	Signature;		/* should = 'USBS' */
	__u32	Tag;			/* same as original command */
	__le32	Residue;		/* amount not transferred */
	__u8	Status;			/* see below */
	__u8	Filler[18];
};

struct scsi_cmnd_fake {
	unsigned char 			cmd_len;
	enum dma_data_direction sc_data_direction;
	/* These elements define the operation we are about to perform */
	unsigned char 			cmnd[MAX_COMMAND_SIZE];
	unsigned 				request_bufflen;	/* Actual request size */
	unsigned char			*request_buffer;	/* Actual requested buffer */
	int 					resid;				/* Number of bytes requested to be transferred
												 * less actual number transferred (0 if not supported) */
	int 					lun;
	int 					result;				/* Status code from device */
};

struct us_data;

typedef int (*trans_cmnd)(struct scsi_cmnd_fake *, struct us_data*);
typedef int (*trans_reset)(struct us_data*);
typedef void (*proto_cmnd)(struct scsi_cmnd_fake*, struct us_data*);
typedef void (*extra_data_destructor)(void *);	/* extra data destructor */

/* we allocate one of these for every device that we remember */
struct us_data {
	/* The device we're working with
	 * It's important to note:
	 *    (o) you must hold dev_mutex to change pusb_dev
	 */
	struct mutex			dev_mutex;	 		/* protect pusb_dev */
	struct usb_device		*pusb_dev;	 		/* this usb_device */
	struct usb_interface	*pusb_intf;	 		/* this interface */
	unsigned long			flags;		 		/* from filter initially */
	unsigned int			send_bulk_pipe;	 	/* cached pipe values */
	unsigned int			recv_bulk_pipe;
	unsigned int			send_ctrl_pipe;
	unsigned int			recv_ctrl_pipe;

	/* information about the device */
	char					*transport_name;
	char					*protocol_name;
	__le32					bcs_signature;
	u8						subclass;
	u8						protocol;
	u8						max_lun;
	u8						ifnum;		 		/* interface number */

	/* function pointers for this device */
	trans_cmnd				transport;	 		/* transport function */
	trans_reset				transport_reset; 	/* transport device reset */
	proto_cmnd				proto_handler;	 	/* protocol handler */

	/* SCSI interfaces */
	struct scsi_cmnd_fake	*srb;		 		/* current srb */
	unsigned int			tag;		 		/* current dCBWTag */
	
	/* control and bulk communications data */
	struct urb				*current_urb;	 	/* USB requests	 */
	struct usb_ctrlrequest	*cr;		 		/* control requests	 */
	unsigned char			*iobuf;		 		/* I/O buffer */
	unsigned char			*sensebuf;	 		/* sense data buffer */
	dma_addr_t				cr_dma;		 		/* buffer DMA addresses */
	dma_addr_t				iobuf_dma;
	/* mutual exclusion and synchronization structures */
	wait_queue_head_t		delay_wait;			/* wait during scan, reset */

	/* block check */
	__u32 					good_blk_addr; 		/* address of good block for writing */
};

/* Structure to hold all of our device specific stuff */
struct usb_cy {
	struct usb_device		*udev;				/* the usb device for this device */
	struct usb_interface	*interface;			/* the interface for this device */
	struct usb_anchor		submitted;			/* in case we need to retract our submissions */
	unsigned char			*bulk_out_buffer;	/* the buffer to send data */
	size_t					bulk_out_size;		/* the size of the send buffer */
	__u8					bulk_out_endpointAddr;	/* the address of the bulk out endpoint */
	spinlock_t				err_lock;				/* lock for errors */
	struct kref				kref;
	struct mutex			io_mutex;				/* synchronize I/O with disconnect */
	unsigned int 			send_bulk_pipe;			/* cached pipe values */
};

/* function prototype	*/
static void	usb_nfw_show_command(struct scsi_cmnd_fake *srb);
static void usb_nfw_blocking_completion(struct urb *urb);
static void usb_nfw_blocking_completion(struct urb *urb);
static int 	interpret_urb_result(struct us_data *us, unsigned int pipe,
								unsigned int length, int result, unsigned int partial);
static int 	usb_nfw_clear_halt(struct us_data *us, unsigned int pipe);
static void usb_nfw_invoke_transport(struct scsi_cmnd_fake *srb, struct us_data *us);
static int 	package_cbd_cbw_command(struct us_data *us ,struct scsi_cmnd_fake *srb,void (*done)(struct scsi_cmnd_fake *));
static int 	usb_nfw_port_reset(struct us_data *us);
static int	usb_nfw_probe(struct usb_interface *intf, const struct usb_device_id *id);

static unsigned char	*fw_buf = NULL;			/* Firmware buffer */
static unsigned int		fw_len = 0;				/* Firmware length */
static unsigned int 	fw_length = 0;			/* Firmware length(temp) */
static int 				bulk_reset_flags = 0;
static char 			dlfw_flag = 0; 			/* flag indicate if fw download succeed */
static struct semaphore			dlfw_flag_sem;	/* limiting the number of writes in progress */
static struct usb_driver 		cy_driver;
static struct scsi_cmnd_fake	*srb;

// save stage2_probe arguments for stage2_write
struct us_data *us = NULL;
u16 pre_vendor_id = 0;
u16 pre_product_id = 0;

static struct usb_device_id cy_table[] = {
	{ USB_DEVICE(USB_CY_VENDOR_ID,USB_CY_PRODUCT_ID_STAGE1) },
	{ USB_DEVICE(USB_CY_VENDOR_ID,USB_CY_PRODUCT_ID_STAGE2) },
	{},
};
MODULE_DEVICE_TABLE(usb, cy_table);
#define to_cy_dev(d) container_of(d, struct usb_cy, kref)

/*
 * ECC computation.
 */
static unsigned char parity[256];
static unsigned char ecc2[256];
static void nand_init_ecc(void)
{
	int i, j, a;

	parity[0] = 0;
	for (i = 1; i < 256; i++)
		parity[i] = (parity[i&(i-1)] ^ 1);

	for (i = 0; i < 256; i++) {
		a = 0;
		for (j = 0; j < 8; j++) {
			if (i & (1<<j)) {
				if ((j & 1) == 0)
					a ^= 0x04;
				if ((j & 2) == 0)
					a ^= 0x10;
				if ((j & 4) == 0)
					a ^= 0x40;
			}
		}
		ecc2[i] = ~(a ^ (a<<1) ^ (parity[i] ? 0xa8 : 0));
	}
}

/* compute 3-byte ecc on 256 bytes */
static void nand_compute_ecc(unsigned char *data, unsigned char *ecc)
{
	int i, j, a;
	unsigned char par, bit, bits[8];

	par = 0;
	for (j = 0; j < 8; j++)
		bits[j] = 0;

	/* collect 16 checksum bits */
	for (i = 0; i < 256; i++) {
		par ^= data[i];
		bit = parity[data[i]];
		for (j = 0; j < 8; j++)
			if ((i & (1<<j)) == 0)
				bits[j] ^= bit;
	}

	/* put 4+4+4 = 12 bits in the ecc */
	a = (bits[3] << 6) + (bits[2] << 4) + (bits[1] << 2) + bits[0];
	ecc[0] = ~(a ^ (a<<1) ^ (parity[par] ? 0xaa : 0));

	a = (bits[7] << 6) + (bits[6] << 4) + (bits[5] << 2) + bits[4];
	ecc[1] = ~(a ^ (a<<1) ^ (parity[par] ? 0xaa : 0));

	ecc[2] = ecc2[par];
}

/***********************************************************************
 * Device probing and disconnecting
 ***********************************************************************/

/* Dissociate from the USB device */
static void dissociate_dev(struct us_data *us)
{
	/* Free the device-related DMA-mapped buffers */
	if (us->cr)
		usb_buffer_free(us->pusb_dev, sizeof(*us->cr), us->cr,
				us->cr_dma);
				
	if (us->iobuf)
		usb_buffer_free(us->pusb_dev, US_IOBUF_SIZE, us->iobuf,
				us->iobuf_dma);
	
	if (us->srb) {
		//us->srb->result = DID_NO_CONNECT << 16;
		if(us->srb->request_buffer)
			kfree(us->srb->request_buffer);
		us->srb->request_buffer = NULL;
		/*free srb*/
		kfree(us->srb);
		us->srb = NULL;
	}

	/* Remove our private data from the interface */
	usb_set_intfdata(us->pusb_intf, NULL);
}

/* Second stage of disconnect processing: deallocate all resources */
static void release_everything(struct us_data *us)
{
	set_bit(US_FLIDX_DISCONNECTING, &us->flags);
	usb_free_urb(us->current_urb);

	dissociate_dev(us);
	kfree(us);
}

static void usb_nfw_show_command(struct scsi_cmnd_fake *srb)
{
	int i;

	US_DEBUGP("Command %d (%d bytes)\n", srb->cmnd[0], srb->cmd_len);
	US_DEBUGP("");
	for (i = 0; i < srb->cmd_len && i < 16; i++)
		US_DEBUGPX(" %02x", srb->cmnd[i]);
	US_DEBUGPX("\n");
}

/* queue a command */
/* This is called with usb_load_nandflash_firmware() held */
static int package_cbd_cbw_command(struct us_data *us ,struct scsi_cmnd_fake *srb,
			void (*done)(struct scsi_cmnd_fake *))
{
	US_DEBUGP("%s called\n", __FUNCTION__);
	/* fail the command if we are disconnecting */
	if (test_bit(US_FLIDX_DISCONNECTING, &us->flags)) {
		US_DEBUGP("Error: Fail command during disconnect\n");
		return (-1);
	}

	/* enqueue the command and wake up the control thread */
	us->srb = srb;
	/*print  scsi request log*/
	US_DEBUG(usb_nfw_show_command(us->srb));
	/*/* send the command to the transport layer ,Bulk-only handler usb_nfw_invoke_transport()*/
	us->proto_handler(us->srb, us);
	
	return 0;
}

/* init srb */
static int usb_nfw_srb_init(struct scsi_cmnd_fake *srb, unsigned char opcode, __be32 addr)
{
	int retval = 0;
	__be16 xfer_len = cpu_to_be16(0x0004);
	
	srb->cmnd[0] = SCSI_CMD;	// command
	srb->cmnd[6] = 0x00;		// banknr
	srb->cmd_len = 10;			// command length
	srb->lun =0;
	memcpy(srb->cmnd+2, &addr, 4);	//addr

	switch (opcode)
	{
		case CBW_NAND_READ_PAGE :	// read
			srb->cmnd[1] = CBW_NAND_READ_PAGE ;	// op code
			memcpy(srb->cmnd+7, &xfer_len, 2);	// Transfer Length
			srb->sc_data_direction = DMA_FROM_DEVICE;
			srb->request_bufflen = NAND_PAGE_SIZE;
			break;
		case CBW_NAND_WRITE_PAGE:	// write
			srb->cmnd[1] = CBW_NAND_WRITE_PAGE;
			memcpy(srb->cmnd+7, &xfer_len, 2);
			srb->sc_data_direction = DMA_TO_DEVICE;
			srb->request_bufflen = NAND_PAGE_SIZE;
			break;
		case CBW_NAND_ERASE_BLK:	// erase one block
			srb->cmnd[1] = CBW_NAND_ERASE_BLK;
			memcpy(srb->cmnd+7, &xfer_len, 2);
			srb->sc_data_direction = DMA_TO_DEVICE;
			srb->request_bufflen = 0;
			break;
		case CBW_NAND_ERASE_ALL:	// erase all blocks
			srb->cmnd[1] = CBW_NAND_ERASE_ALL;
			srb->sc_data_direction = DMA_TO_DEVICE;
			srb->request_bufflen = 0;
			memset(&(srb->cmnd[2]), 0, 8);
			break;
		case CBW_8051_RENUM :		// renumerate
			srb->cmnd[1] = CBW_8051_RENUM;
			srb->sc_data_direction = DMA_TO_DEVICE;
			srb->request_bufflen = 0;
			break;
		default:
			printk(KERN_ALERT "Error: wrong op code!\n");
			retval = -EPERM;
			break;
	}

	return retval;
}

/* Check page read and written */
static int page_check(struct us_data *us, struct scsi_cmnd_fake *srb,__be32 addr, unsigned char *page)
{
	int i = 0;

	memset(srb->request_buffer, 0, NAND_PAGE_SIZE);
	
	// read page
	usb_nfw_srb_init(srb, CBW_NAND_READ_PAGE, addr);
	package_cbd_cbw_command(us,srb,NULL);
	if (USB_STOR_TRANSPORT_GOOD != srb->result) {
		printk(KERN_ALERT "Error: failed to read page!\n");
		
		return USB_FIRMWARE_READ_FAILED;
	}

	// check page
	for (i = 0 ; i < NAND_PAGE_SIZE ; i++) {
		if (srb->request_buffer[i] != page[i]) {
			printk(KERN_ALERT "Error: page check error!\n");
			
			return USB_FIRMWARE_CHECK_ERROR;	
		}	
	}

	return USB_FIRMWARE_SUCCESS;
}

/* Write configuration page 0 to nandflash */
static int usb_nfw_wrc_cfg0(struct us_data *us, struct scsi_cmnd_fake *srb, int cfg0_times)
{
	int retval = USB_FIRMWARE_SUCCESS;
	
	/* we lost much time in __be32 or __le32, SCSI cmd is __be32,CBD cmd is __le32	*/
	__be32 addr = cpu_to_be32(0x0);
	__le16 addr_cfg0_0 = (__le16)cpu_to_le32(us->good_blk_addr + 0x0020);
	__le16 addr_cfg0_1 = (__le16)cpu_to_le32(us->good_blk_addr + 0x0000);

	unsigned char cfg0_0[NAND_PAGE_SIZE] = {
		/* [0x00] .. [0x1F] */
		0x53, 0x4D, 0x54, 0x44, 0x4D, 0x47, 0x20, 0x00,
		0x12, 0x00, 0x01, 0x06, 0x0A, 0x00, 0x03, 0x02,
		0x53, 0x4D, 0x54, 0x44, 0x4D, 0x47, 0x20, 0x00,
		0x12, 0x00, 0x01, 0x06, 0x0A, 0x00, 0x03, 0x02,
		[0x20 ... 0x1FF] = 0x00,
		[0x200] = 0x00, [0x201] = 0x01, [0x202 ... 0x7FF] = 0x00,
		[0x800] = 0xFF, [0x801] = 0x01, [0x802 ... 0x83F] = 0xFF,
		[0x831] = 0x01,
	};
	unsigned char cfg0_1[NAND_PAGE_SIZE] = {
		/* [0x00] .. [0x1F] */
		0x53, 0x4D, 0x54, 0x44, 0x4D, 0x47, 0x00, 0x00,
		0x12, 0x00, 0x01, 0x06, 0x0A, 0x00, 0x03, 0x02,
		0x53, 0x4D, 0x54, 0x44, 0x4D, 0x47, 0x00, 0x00,
		0x12, 0x00, 0x01, 0x06, 0x0A, 0x00, 0x03, 0x02,
		[0x20 ... 0x1FF] = 0x00,
		[0x200] = 0x00, [0x201] = 0x01, [0x202 ... 0x7FF] = 0x00,
		[0x800] = 0xFF, [0x801] = 0x01, [0x802 ... 0x83F] = 0xFF,
		[0x831] = 0x01,
	};

	/* next block addr */
	// cfg0_0
	memcpy(cfg0_0+0x06, &addr_cfg0_0, 2);
	memcpy(cfg0_0+0x16, &addr_cfg0_0, 2);
	// cfg0_1	
	memcpy(cfg0_1+0x06, &addr_cfg0_1, 2);
	memcpy(cfg0_1+0x16, &addr_cfg0_1, 2);

	nand_init_ecc();
	nand_compute_ecc(&cfg0_0[0],&cfg0_0[0x20A]);
	nand_compute_ecc(&cfg0_0[256],&cfg0_0[0x20D]);
	nand_compute_ecc(&cfg0_1[0],&cfg0_1[0x20A]);
	nand_compute_ecc(&cfg0_1[256],&cfg0_1[0x20D]);
	
	/*	write page 0 configure	*/
	if (1 == cfg0_times)	{
		/* we lost much time in __be32 or __le32, SCSI cmd is __be32,CBD cmd is __le32	*/
		addr = cpu_to_be32(us->good_blk_addr + 0x00);
		memcpy(srb->request_buffer, &cfg0_0[0], NAND_PAGE_SIZE);
	}
	else if (2 == cfg0_times) {
		/* we lost much time in __be32 or __le32, SCSI cmd is __be32,CBD cmd is __le32	*/
		addr = cpu_to_be32(us->good_blk_addr + 0x20);
		memcpy(srb->request_buffer, &cfg0_1[0], NAND_PAGE_SIZE);
	}

	usb_nfw_srb_init(srb, CBW_NAND_WRITE_PAGE, addr);
	/* write cfpage0_0 and cfpage0_1*/
	package_cbd_cbw_command(us, srb, NULL);
	if (USB_STOR_TRANSPORT_GOOD != srb->result) {
		printk(KERN_ALERT "Error: failed to write configuration page0 to nandflash!\n");

		return USB_FIRMWARE_WRITE_FAILED;
	}

	/* check page0 configure */
	// check ecc and data
	if (1 == cfg0_times) {
		retval = page_check(us,srb,addr,cfg0_0);
		if (USB_FIRMWARE_SUCCESS != retval) {
			printk(KERN_ALERT "Error: configuration page 0 check error!\n");
		}
	}
	else if (2 == cfg0_times) {
		retval = page_check(us,srb,addr,cfg0_1);
		if (USB_FIRMWARE_SUCCESS != retval) {
			printk(KERN_ALERT "Error: configuration page 0 check error!\n");
		}
	}
	
	return retval;
}

/* Write configuration page 1 to nandflash */
static int usb_nfw_wrc_cfg1(struct us_data *us, struct scsi_cmnd_fake *srb, int cfg1_times)
{
	int retval = USB_FIRMWARE_SUCCESS;
	__be32 addr = cpu_to_be32(0x0);

	unsigned char cfg1[NAND_PAGE_SIZE]={
		[0]=0x4C,[1] = 0x05, [2] = 0xB5, [3] = 0x07,
		 [4 ... 0x0F] = 0x00,
		/* [0x10 ... 0x27] */
		0x53, 0x6F, 0x6E, 0x79, 0x20, 0x20, 0x20, 0x20,
		0x53, 0x74, 0x6F, 0x72, 0x61, 0x67, 0x65, 0x20,
		0x44, 0x72, 0x69, 0x76, 0x65, 0x20, 0x20, 0x20,	
		[0x28 ... 0x2F] = 0x00,
	    /* [0x30 ... 0x33] */
		0x04, 0x03, 0x09, 0x04,
		[0x34 ... 0x5F] = 0x00,
	  	/* [0x60 ... 0x68] */
		0x0A, 0x03, 0x53, 0x00, 0x6F, 0x00, 0x6E, 0x00,
		0x79,
		[0x69 ... 0x8F] = 0x00,
		/* [0x90 ... 0xAF] */
		0x1E, 0x03, 0x53, 0x00, 0x6F, 0x00, 0x6E, 0x00,
		0x79, 0x00, 0x20, 0x00, 0x43, 0x00, 0x61, 0x00,
		0x6D, 0x00, 0x63, 0x00, 0x6F, 0x00, 0x72, 0x00,
		0x64, 0x00, 0x65, 0x00, 0x72, 0x00, 0x00, 0x00,
		[0xB0 ...0xBF] = 0x00,
		/* [0xC0 ... 0xDF] */
		0x20, 0x03, 0x30, 0x00, 0x30, 0x00, 0x30, 0x00,
		0x30, 0x00, 0x30, 0x00, 0x30, 0x00, 0x30, 0x00,
		0x30, 0x00, 0x30, 0x00, 0x30, 0x00, 0x30, 0x00,
		0x30, 0x00, 0x30, 0x00, 0x30, 0x00, 0x30, 0x00,
		[0xE0 ... 0x1FF] = 0x00,

		[0x200] = 0x00, [0x201] = 0x01, [0x202 ... 0x7FF] = 0x00,
		[0x800] = 0xFF, [0x801] = 0x01, [0x802 ...0x83F] = 0xFF,
		[0x831] = 0x01,	
	};

	nand_init_ecc();
	// calc ecc
	nand_compute_ecc(&cfg1[0],&cfg1[0x20A]);
	nand_compute_ecc(&cfg1[256],&cfg1[0x20D]);
	memcpy(srb->request_buffer,&cfg1[0],NAND_PAGE_SIZE);
		
	if (1 == cfg1_times)
		addr = cpu_to_be32(us->good_blk_addr + 0x01);
	else if (2 == cfg1_times)
		addr = cpu_to_be32(us->good_blk_addr + 0x21);

	usb_nfw_srb_init(srb, CBW_NAND_WRITE_PAGE, addr);
	package_cbd_cbw_command(us, srb, NULL);
	if (USB_STOR_TRANSPORT_GOOD != srb->result) {
		printk(KERN_ALERT "Error: failed to write configuration page 1 to nandflash!\n");
		
		return USB_FIRMWARE_WRITE_FAILED;
	}

	/* check page1 configure */
	// check ecc and data
	retval = page_check(us,srb,addr,cfg1);
	if (USB_FIRMWARE_SUCCESS != retval) {
		printk(KERN_ALERT "Error: configuration page 1 check error!\n");
	}		
	
	return retval;
}

/* Write firmware to nandflash */
static int usb_nfw_wrc_fw(struct us_data *us, struct scsi_cmnd_fake *srb, unsigned int fw_times)
{
	int retval = USB_FIRMWARE_SUCCESS;
	int i = 0;
	unsigned int fw_packets_num;
	__be32 addr = cpu_to_be32(0x0);
	unsigned char fw_page[NAND_PAGE_SIZE] = {0};

	nand_init_ecc();

	// send FW packet(18 = 8938/512 = 17.46)
	fw_packets_num = fw_len / NAND_PAGE_DATA_SIZE;
	if (fw_packets_num % NAND_PAGE_DATA_SIZE != 0)
		fw_packets_num += 1;

	for (i = 0; i < fw_packets_num; i++) {	
		if (1 == fw_times)
			addr = cpu_to_be32(us->good_blk_addr + 0x0000002 + i);
		if (2 == fw_times)
			addr = cpu_to_be32(us->good_blk_addr + 0x0000022 + i);

		// copy FW, most 512bypes
		memset(srb->request_buffer,0x00,NAND_PAGE_SIZE);
		if ((fw_packets_num != fw_len/NAND_PAGE_DATA_SIZE) && (fw_packets_num-1 == i))
			memcpy(srb->request_buffer, fw_buf+NAND_PAGE_DATA_SIZE*i, fw_len-NAND_PAGE_DATA_SIZE*i);
		else
			memcpy(srb->request_buffer, fw_buf+NAND_PAGE_DATA_SIZE*i, NAND_PAGE_DATA_SIZE);
		
		memset(srb->request_buffer+0x00200, 0xff, 0x640);

		/* Fix data 0x01 */
		srb->request_buffer[0x00201] = 0x01;
		srb->request_buffer[0x00801] = 0x01;
		srb->request_buffer[0x00831] = 0x01;
			
		// prepare ECC for each fw transfer
		nand_compute_ecc(srb->request_buffer, srb->request_buffer+0x20A);
		nand_compute_ecc(srb->request_buffer+256, srb->request_buffer+0x20D);

		// write a fw packet
		usb_nfw_srb_init(srb, CBW_NAND_WRITE_PAGE, addr);
		package_cbd_cbw_command(us,srb,NULL);
		if (USB_STOR_TRANSPORT_GOOD != srb->result) {
			printk(KERN_ALERT "Error: failed to write firmware page to nandflash!\n");
			
			retval = USB_FIRMWARE_WRITE_FAILED;
			/* break loop */
			break;
		}

	 	/* check firmware configure */
		// check ecc and data, if any page goes wrong, exit
		memcpy(fw_page,srb->request_buffer,NAND_PAGE_SIZE);
		retval = page_check(us,srb,addr,fw_page);
		if (USB_FIRMWARE_SUCCESS != retval) {
			printk(KERN_ALERT "Error: firmware page check error!\n");
			
			return retval;
		}
	}
	
	return retval;
}

/* Write configuration pages and firware to nand flash */
static int usb_load_nandflash_firmware(struct us_data *us)
{
	int ret = 0;
	int retval = USB_FIRMWARE_SUCCESS;
	int per_page_times = 1;
	__be32 addr;	

	// alloc and init srb
	srb = kzalloc(sizeof(*srb),GFP_KERNEL);
	if (!srb) {
		printk(KERN_ALERT "Error: failed to alloc memory for srb!\n");
		retval = -ENOMEM;
		goto exit;	
	}
	srb->request_bufflen = CBD_DATA_XFER_LENGTH;
	srb->request_buffer= kzalloc(sizeof(*(srb->request_buffer))*CBD_DATA_XFER_LENGTH,GFP_KERNEL);
	if (!srb->request_buffer) {
		printk(KERN_ALERT "Error: failed to alloc memory for srb's request_buffer!\n");
		retval = -ENOMEM;
		goto exit;	
	}
	
	us->good_blk_addr = 0x0;

	// erase all blocks
	addr = cpu_to_be32(0x00000000);		// in fact, erase all cmd don't need addr
	usb_nfw_srb_init(srb, CBW_NAND_ERASE_ALL, addr);
	package_cbd_cbw_command(us,srb,NULL);
	if (USB_STOR_TRANSPORT_GOOD != srb->result) {
		printk(KERN_ALERT "Error: failed to erase all blocks!\n");
		
		retval = USB_FIRMWARE_ERASE_FAILED;
		goto exit;
	}

	// write cfg 0, cfg 1 and firmware
	while (per_page_times <= 2)
	{
		// write cfg 0
		ret = usb_nfw_wrc_cfg0(us, srb, per_page_times);
		if (USB_FIRMWARE_SUCCESS != ret) {
			printk(KERN_ALERT "Error: failed to write configuration page 0 to nandflash!\n");
			retval = ret;
			goto exit;
		}
		
		// write cfg 1
		ret = usb_nfw_wrc_cfg1(us, srb, per_page_times);
		if (USB_FIRMWARE_SUCCESS != ret) {
			printk(KERN_ALERT "Error: failed to write configuration page 1 to nandflash!\n");
			retval = ret;
			goto exit;
		}

		// write fw
		ret = usb_nfw_wrc_fw(us, srb, per_page_times);
		if (USB_FIRMWARE_SUCCESS != ret) {
			printk(KERN_ALERT "Error: failed to write firmware page to nandflash!\n");
			retval = ret;
			goto exit;
		}

		// 2 times
		per_page_times++;
	}

	// re-numerate
	addr = cpu_to_be32(0x00000000);
	usb_nfw_srb_init(srb, CBW_8051_RENUM , addr);
	package_cbd_cbw_command(us,srb,NULL);
	if (USB_STOR_TRANSPORT_GOOD != srb->result) {
		printk(KERN_ALERT "Error: failed to renumerate after writing cfg0, cfg1 and FW pages to nandflash!\n");
		
		retval = USB_FIRMWARE_RENUM_FAILED;
	}

exit:
	// set fw download success flag
	if (USB_FIRMWARE_SUCCESS == retval)
		dlfw_flag = 1;
	else
		dlfw_flag = 0;
	
	return retval;
}

/* This is the common part of the URB message submission code
 *
 * All URBs from the usb-storage driver involved in handling a queued scsi
 * command _must_ pass through this function (or something like it) for the
 * abort mechanisms to work properly.
 */
static int usb_nfw_msg_common(struct us_data *us, int timeout)
{
	struct completion urb_done;
	long timeleft;
	int status;

	/* don't submit URBs during abort/disconnect processing */
	if (us->flags & ABORTING_OR_DISCONNECTING) {
		printk(KERN_ALERT "Error: can't submit URBs during abort/disconnect processing!\n");
		
		return -EIO;
	}

	/* set up data structures for the wakeup system */
	init_completion(&urb_done);

	/* fill the common fields in the URB */
	us->current_urb->context = &urb_done;
	us->current_urb->actual_length = 0;
	us->current_urb->error_count = 0;
	us->current_urb->status = 0;

	/* we assume that if transfer_buffer isn't us->iobuf then it
	 * hasn't been mapped for DMA.  Yes, this is clunky, but it's
	 * easier than always having the caller tell us whether the
	 * transfer buffer has already been mapped. */
	us->current_urb->transfer_flags = URB_NO_SETUP_DMA_MAP;
	if (us->current_urb->transfer_buffer == us->iobuf)
		/*use DMA*/
		us->current_urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
	us->current_urb->transfer_dma = us->iobuf_dma;
	us->current_urb->setup_dma = us->cr_dma;

	/* submit the URB */
	status = usb_submit_urb(us->current_urb, GFP_NOIO);
	if (status) {
		/* something went wrong */
		printk(KERN_ALERT "Error: failed to submit urb!\n");
		
		return status;
	}

	/* since the URB has been submitted successfully, it's now okay
	 * to cancel it */
	set_bit(US_FLIDX_URB_ACTIVE, &us->flags);

	/* did an abort/disconnect occur during the submission? */
	if (us->flags & ABORTING_OR_DISCONNECTING) {

		/* cancel the URB, if it hasn't been cancelled already */
		if (test_and_clear_bit(US_FLIDX_URB_ACTIVE, &us->flags)) {
			US_DEBUGP("-- cancelling URB\n");
			usb_unlink_urb(us->current_urb);
		}
	}

	/* wait for the completion of the URB */
	timeleft = wait_for_completion_interruptible_timeout(
			&urb_done, timeout ? : MAX_SCHEDULE_TIMEOUT);

	clear_bit(US_FLIDX_URB_ACTIVE, &us->flags);

	if (timeleft <= 0) {
		US_DEBUGP("%s -- cancelling URB\n",
			  timeleft == 0 ? "Timeout" : "Signal");
		usb_kill_urb(us->current_urb);
	}

	/* return the URB status */
	return us->current_urb->status;
}

/*
 * Interpret the results of a URB transfer
 *
 * This function prints appropriate debugging messages, clears halts on
 * non-control endpoints, and translates the status to the corresponding
 * USB_STOR_XFER_xxx return code.
 */
static int interpret_urb_result(struct us_data *us, unsigned int pipe,
		unsigned int length, int result, unsigned int partial)
{
	US_DEBUGP("Status code %d; transferred %u/%u\n",
			result, partial, length);
	switch (result) {
		/* no error code; did we send all the data? */
		case 0:
			if (partial != length) {
				US_DEBUGP("-- short transfer\n");
				return USB_STOR_XFER_SHORT;
			}
	
			US_DEBUGP("-- transfer complete\n");
			return USB_STOR_XFER_GOOD;

		/* stalled */
		case -EPIPE:
			/* for control endpoints, (used by CB[I]) a stall indicates
			 * a failed command */
			if (usb_pipecontrol(pipe)) {
				US_DEBUGP("-- stall on control pipe\n");
				return USB_STOR_XFER_STALLED;
			}

			/* for other sorts of endpoint, clear the stall */
			US_DEBUGP("clearing endpoint halt for pipe 0x%x\n", pipe);
			if (usb_nfw_clear_halt(us, pipe) < 0)
				return USB_STOR_XFER_ERROR;
			return USB_STOR_XFER_STALLED;

		/* babble - the device tried to send more than we wanted to read */
		case -EOVERFLOW:
			printk(KERN_ALERT "Error: babble!\n");
			return USB_STOR_XFER_LONG;

		/* the transfer was cancelled by abort, disconnect, or timeout */
		case -ECONNRESET:
			printk(KERN_ALERT "Error: transfer cancelled!\n");
			return USB_STOR_XFER_ERROR;

		/* short scatter-gather read transfer */
		case -EREMOTEIO:
			printk(KERN_ALERT "Error: short read transfer!\n");
			return USB_STOR_XFER_SHORT;

		/* abort or disconnect in progress */
		case -EIO:
			printk(KERN_ALERT "Error: abort or disconnect in progress!\n");
			return USB_STOR_XFER_ERROR;

		/* the catch-all error case */
		default:
			printk(KERN_ALERT "Error: unknown error!\n");
			return USB_STOR_XFER_ERROR;
	}
}

/*
 * Transfer one buffer via bulk pipe, without timeouts, but allowing early
 * termination.  Return codes are USB_STOR_XFER_xxx.  If the bulk pipe
 * stalls during the transfer, the halt is automatically cleared.
 */
static int usb_nfw_bulk_transfer_buf(struct us_data *us, unsigned int pipe,
	void *buf, unsigned int length, unsigned int *act_len)
{
	int result;

	US_DEBUGP("%s: xfer %u bytes\n", __FUNCTION__, length);

	/* fill and submit the URB */
	usb_fill_bulk_urb(us->current_urb, us->pusb_dev, pipe, buf, length,
		      usb_nfw_blocking_completion, NULL);
	result = usb_nfw_msg_common(us, 0);

	/* store the actual length of the data transferred */
	if (act_len)
		*act_len = us->current_urb->actual_length;
	return interpret_urb_result(us, pipe, length, result,
			us->current_urb->actual_length);
}

static int usb_nfw_Bulk_transport(struct scsi_cmnd_fake *srb, struct us_data *us)
{
	/*use dma*/
	struct bulk_cb_wrap *bcb = (struct bulk_cb_wrap *) us->iobuf;
	struct bulk_cs_wrap *bcs = (struct bulk_cs_wrap *) us->iobuf;
	/*dCBWDataTransferLength */
	unsigned int transfer_length = srb->request_bufflen;
	unsigned int residue;
	int result;
	int fake_sense = 0;
	unsigned int cswlen;
	unsigned int cbwlen = US_BULK_CB_WRAP_LEN;

	/* set up the command wrapper */
	/*DCBWSignature =USBC*/
	bcb->Signature = cpu_to_le32(US_BULK_CB_SIGN);
	/*dCBWDataTransferLength*/
	bcb->DataTransferLength = cpu_to_le32(transfer_length);
	/*BmCBWFlags*/
	bcb->Flags = srb->sc_data_direction == DMA_FROM_DEVICE ? 1 << 7 : 0;
	bcb->Tag = ++us->tag;
	bcb->Lun = srb->lun;
	bcb->Length = srb->cmd_len;

	/* CDB :copy the command payload */
	memset(bcb->CDB, 0, sizeof(bcb->CDB));
	memcpy(bcb->CDB, srb->cmnd, bcb->Length);

	/* send it to out endpoint */
	US_DEBUGP("Bulk Command S 0x%x T 0x%x L %d F %d Trg %d LUN %d CL %d\n",
			le32_to_cpu(bcb->Signature), bcb->Tag,
			le32_to_cpu(bcb->DataTransferLength), bcb->Flags,
			(bcb->Lun >> 4), (bcb->Lun & 0x0F),
			bcb->Length);
	/*Stage 1:  CBW stage */
	/*use DMA,urb->transfer_dma*/
	result = usb_nfw_bulk_transfer_buf(us, us->send_bulk_pipe,
				bcb, cbwlen, NULL);
	US_DEBUGP("Bulk command transfer result=%d\n", result);
	if (result != USB_STOR_XFER_GOOD) {
		printk(KERN_ALERT "Error: usb_nfw_bulk_transfer_buf error!\n");
		
		return USB_STOR_TRANSPORT_ERROR;
	}

	/*Stage 2: Data stage */
	/* DATA STAGE */
	/* send/receive data payload, if there is any */

	/* Some USB-IDE converter chips need a 100us delay between the
	 * command phase and the data phase.  Some devices need a little
	 * more than that, probably because of clock rate inaccuracies. */
	if (unlikely(us->flags & US_FL_GO_SLOW))
		udelay(125);

	if (transfer_length) {
		unsigned int pipe = srb->sc_data_direction == DMA_FROM_DEVICE ?
				us->recv_bulk_pipe : us->send_bulk_pipe;
		/*no use DMA,urb->transfer_buffer*/
		result = usb_nfw_bulk_transfer_buf(us, pipe,
				srb->request_buffer, transfer_length, &srb->resid);
		US_DEBUGP("Bulk data transfer result 0x%x\n", result);
		if (result == USB_STOR_XFER_ERROR) {
			printk(KERN_ALERT "Error: usb_nfw_bulk_transfer_buf error!\n");
			
			return USB_STOR_TRANSPORT_ERROR;
		}

		/* If the device tried to send back more data than the
		 * amount requested, the spec requires us to transfer
		 * the CSW anyway.  Since there's no point retrying the
		 * the command, we'll return fake sense data indicating
		 * Illegal Request, Invalid Field in CDB.
		 */
		if (result == USB_STOR_XFER_LONG)
			fake_sense = 1;
	}

	/* See flow chart on pg 15 of the Bulk Only Transport spec for
	 * an explanation of how this code works.
	 */
	/*Stage 3: CSW stage */
	/* get CSW for device status */
	US_DEBUGP("Attempting to get CSW...\n");
	result = usb_nfw_bulk_transfer_buf(us, us->recv_bulk_pipe,
				bcs, US_BULK_CS_WRAP_LEN, &cswlen);

	/* Some broken devices add unnecessary zero-length packets to the
	 * end of their data transfers.  Such packets show up as 0-length
	 * CSWs.  If we encounter such a thing, try to read the CSW again.
	 */
	if (result == USB_STOR_XFER_SHORT && cswlen == 0) {
		US_DEBUGP("Received 0-length CSW; retrying...\n");
		/*use DMA,urb->transfer_dma*/
		result = usb_nfw_bulk_transfer_buf(us, us->recv_bulk_pipe,
				bcs, US_BULK_CS_WRAP_LEN, &cswlen);
	}

	/* did the attempt to read the CSW fail? */
	if (result == USB_STOR_XFER_STALLED) {

		/* get the status again */
		US_DEBUGP("Attempting to get CSW (2nd try)...\n");
		result = usb_nfw_bulk_transfer_buf(us, us->recv_bulk_pipe,
				bcs, US_BULK_CS_WRAP_LEN, NULL);
	}

	/* if we still have a failure at this point, we're in trouble */
	US_DEBUGP("Bulk status result = %d\n", result);
	if (result != USB_STOR_XFER_GOOD) {
		printk(KERN_ALERT "Error: usb_nfw_bulk_transfer_buf error!\n");
	
		return USB_STOR_TRANSPORT_ERROR;
	}

	/* check bulk status */
	residue = le32_to_cpu(bcs->Residue);
	US_DEBUGP("Bulk Status S 0x%x T 0x%x R %u Stat 0x%x\n",
			le32_to_cpu(bcs->Signature), bcs->Tag,
			residue, bcs->Status);
	if (bcs->Tag != us->tag || bcs->Status > US_BULK_STAT_PHASE) {
		printk(KERN_ALERT "Error: Bulk logical error!\n");
		return USB_STOR_TRANSPORT_ERROR;
	}

	/* Some broken devices report odd signatures, so we do not check them
	 * for validity against the spec. We store the first one we see,
	 * and check subsequent transfers for validity against this signature.
	 */
	if (!us->bcs_signature) {
		us->bcs_signature = bcs->Signature;
		if (us->bcs_signature != cpu_to_le32(US_BULK_CS_SIGN))
			US_DEBUGP("Learnt BCS signature 0x%08X\n",
					le32_to_cpu(us->bcs_signature));
	} else if (bcs->Signature != us->bcs_signature) {
		US_DEBUGP("Signature mismatch: got %08X, expecting %08X\n",
			  le32_to_cpu(bcs->Signature),
			  le32_to_cpu(us->bcs_signature));
		return USB_STOR_TRANSPORT_ERROR;
	}

	/* try to compute the actual residue, based on how much data
	 * was really transferred and what the device tells us */
	if (residue) {
		if (!(us->flags & US_FL_IGNORE_RESIDUE)) {
			residue = min(residue, transfer_length);
			srb->resid = max(srb->resid, (int) residue);
		}
	}
	/* finished working on this command */
	//us->srb = NULL;
	/* based on the status code, we report good or bad */
	switch (bcs->Status) {
		case US_BULK_STAT_OK:
			/* device babbled -- return fake sense data */
			if (fake_sense) {
				return USB_STOR_TRANSPORT_NO_SENSE;
			}
			/* command good -- note that data could be short */
			return USB_STOR_TRANSPORT_GOOD;

		case US_BULK_STAT_FAIL:
			/* command failed */
			return USB_STOR_TRANSPORT_FAILED;

		case US_BULK_STAT_PHASE:
			/* phase error -- note that a transport reset will be
			 * invoked by the invoke_transport() function
			 */
			return USB_STOR_TRANSPORT_ERROR;
	}

	/* we should never get here, but if we do, we're in trouble */
	return USB_STOR_TRANSPORT_ERROR;
}


/* Issue a USB port reset to the device.  The caller must not hold
 * us->dev_mutex.
 */
static int usb_nfw_port_reset(struct us_data *us)
{
	int result, rc_lock;

	result = rc_lock =
		usb_lock_device_for_reset(us->pusb_dev, us->pusb_intf);
	if (result < 0)
		printk(KERN_ALERT "Error: unable to lock device for reset: %d!\n", result);
	else {
		/* Were we disconnected while waiting for the lock? */
		if (test_bit(US_FLIDX_DISCONNECTING, &us->flags)) {
			result = -EIO;
			printk(KERN_ALERT "Error: No reset during disconnect!\n");
		} else {
			result = usb_reset_device(us->pusb_dev);
			US_DEBUGP("usb_reset_composite_device returns %d\n",
					result);
		}
		if (rc_lock)
			usb_unlock_device(us->pusb_dev);
	}
	
	return result;
}

/* Invoke the transport and basic error-handling/recovery methods
 *
 * This is used by the protocol layers to actually send the message to
 * the device and receive the response.
 */
static void usb_nfw_invoke_transport(struct scsi_cmnd_fake *srb, struct us_data *us)
{

	int result;

	/* send the command to the transport layer */
	srb->resid = 0;
	/* real invoke usb_nfw_Bulk_transport()*/
	result = us->transport(srb, us);
	US_DEBUGP("---%s %d result =%d\n",__FUNCTION__,__LINE__,result);
	/* if there is a transport error, reset and don't auto-sense */
	if (result == USB_STOR_TRANSPORT_ERROR) {
		printk(KERN_ALERT "Error: transport indicates error, resetting!\n");
		/*we use usb_nfw_Bulk_transport() return value, for check */
		srb->result = USB_STOR_TRANSPORT_ERROR;//DID_ERROR << 16;
		goto Handle_Errors;
	}
	//save result
	srb->result = result;

	return;

	/* Error and abort processing: try to resynchronize with the device
	 * by issuing a port reset.  If that fails, try a class-specific
	 * device reset. */
Handle_Errors:

	/* Set the RESETTING bit, and clear the ABORTING bit so that
	 * the reset may proceed. */

	set_bit(US_FLIDX_RESETTING, &us->flags);
	clear_bit(US_FLIDX_ABORTING, &us->flags);


	/* We must release the device lock because the pre_reset routine
	 * will want to acquire it. */
	mutex_unlock(&us->dev_mutex);
	result = usb_nfw_port_reset(us);
	mutex_lock(&us->dev_mutex);

	if (result < 0) {
		/*NOTE: need reset*/
		bulk_reset_flags = 1;
		us->transport_reset(us);
	}
	clear_bit(US_FLIDX_RESETTING, &us->flags);
}

/* This is the completion handler which will wake us up when an URB
 * completes.
 */
static void usb_nfw_blocking_completion(struct urb *urb)
{
	struct completion *urb_done_ptr = (struct completion *)urb->context;

	complete(urb_done_ptr);
}

/*
 * Transfer one control message, with timeouts, and allowing early
 * termination.  Return codes are usual -Exxx, *not* USB_STOR_XFER_xxx.
 */
static int usb_nfw_control_msg(struct us_data *us, unsigned int pipe,
		 u8 request, u8 requesttype, u16 value, u16 index,
		 void *data, u16 size, int timeout)
{
	int status;

	US_DEBUGP("%s: rq=%02x rqtype=%02x value=%04x index=%02x len=%u\n",
			__FUNCTION__, request, requesttype,
			value, index, size);

	/* fill in the devrequest structure */
	us->cr->bRequestType = requesttype;
	us->cr->bRequest = request;
	us->cr->wValue = cpu_to_le16(value);
	us->cr->wIndex = cpu_to_le16(index);
	us->cr->wLength = cpu_to_le16(size);

	/* fill and submit the URB */
	/*use control DMA ,urb->setup_dma*/
	usb_fill_control_urb(us->current_urb, us->pusb_dev, pipe,
			 (unsigned char*) us->cr, data, size,
			 usb_nfw_blocking_completion, NULL);
	status = usb_nfw_msg_common(us, timeout);

	/* return the actual length of the data transferred if no error */
	if (status == 0)
		status = us->current_urb->actual_length;
		
	return status;
}

/* This is a version of usb_clear_halt() that allows early termination and
 * doesn't read the status from the device -- this is because some devices
 * crash their internal firmware when the status is requested after a halt.
 *
 * A definitive list of these 'bad' devices is too difficult to maintain or
 * make complete enough to be useful.  This problem was first observed on the
 * Hagiwara FlashGate DUAL unit.  However, bus traces reveal that neither
 * MacOS nor Windows checks the status after clearing a halt.
 *
 * Since many vendors in this space limit their testing to interoperability
 * with these two OSes, specification violations like this one are common.
 */
static int usb_nfw_clear_halt(struct us_data *us, unsigned int pipe)
{
	int result;
	int endp = usb_pipeendpoint(pipe);

	if (usb_pipein (pipe))
		endp |= USB_DIR_IN;

	result = usb_nfw_control_msg(us, us->send_ctrl_pipe,
		USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT,
		USB_ENDPOINT_HALT, endp,
		NULL, 0, 3*HZ);

	/* reset the endpoint toggle */
	if (result >= 0)
		usb_settoggle(us->pusb_dev, usb_pipeendpoint(pipe),
				usb_pipeout(pipe), 0);

	US_DEBUGP("%s: result = %d\n", __FUNCTION__, result);
	
	return result;
}

/* This issues a Bulk-only Reset to the device in question, including
 * clearing the subsequent endpoint halts that may occur.
 */
static int usb_nfw_Bulk_reset(struct us_data *us)
{

	int result;
	int result2;
	US_DEBUGP("%s called\n", __FUNCTION__);
	if (test_bit(US_FLIDX_DISCONNECTING, &us->flags)) {
		printk(KERN_ALERT "Error: No reset during disconnect!\n");
		return -EIO;
	}

	result = usb_nfw_control_msg(us, us->send_ctrl_pipe,
			US_BULK_RESET_REQUEST,
			USB_TYPE_CLASS | USB_RECIP_INTERFACE,
			0, us->ifnum, 0,0,5*HZ);
	if (result < 0) {
		printk(KERN_ALERT "Error: Soft reset failed: %d!\n", result);
		return result;
	}

 	/* Give the device some time to recover from the reset,
 	 * but don't delay disconnect processing. */
 	wait_event_interruptible_timeout(us->delay_wait,
 			test_bit(US_FLIDX_DISCONNECTING, &us->flags),
 			HZ*6);
	if (test_bit(US_FLIDX_DISCONNECTING, &us->flags)) {
		US_DEBUGP("Reset interrupted by disconnect\n");
		return -EIO;
	}

	US_DEBUGP("Soft reset: clearing bulk-in endpoint halt\n");
	result = usb_nfw_clear_halt(us, us->recv_bulk_pipe);

	US_DEBUGP("Soft reset: clearing bulk-out endpoint halt\n");
	result2 = usb_nfw_clear_halt(us, us->send_bulk_pipe);

	/* return a result code based on the result of the clear-halts */
	if (result >= 0)
		result = result2;
	if (result < 0)
		printk(KERN_ALERT "Error: Soft reset failed!\n");
	else
		US_DEBUGP("Soft reset done\n");
	return result;
}

static int cy_ioctl_stage2(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int err = 0;
	int copy_length;
	int retval = 0;
	
	// check cmd type
	if (_IOC_TYPE(cmd) != CY_IOC_MAGIC)	{
		printk(KERN_ALERT "Error: Invalid cmd magic number\n");

		return -EINVAL;
	}
	// check cmd nr
	if (_IOC_NR(cmd) > CY_IOC_MAXNR) {
		printk(KERN_ALERT "Error: Invalid cmd number\n");

		return -EINVAL;
	}
	// check cmd dir
	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
	else if (_IOC_DIR(cmd) & _IOC_WRITE)
		err = !access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
	if (err) {
		printk(KERN_ALERT "Error: Invalid cmd direction\n");

		return -EINVAL;
	}
	
	// run func according to cmd
	switch(cmd)	{
		case CY_IOC_FW_LEN:
			if (!arg) {
				printk(KERN_ALERT "Error: userspace pointer is NULL!\n");
			
				return -EFAULT;
			}
			// copy FW length to buffer from userspace
			copy_length = copy_from_user(&fw_length, (unsigned int __user*)arg, sizeof(unsigned int));
			if (copy_length) {
				printk(KERN_ALERT "Error: failed to copy data from userspace! data left for copying: %d\n", copy_length);
			
				return -EFAULT;
			}
			
			// fw length = fw_length-1, the last byte is used for storing dlfw_flag
			fw_length--;

			break;
		case CY_IOC_DLFW_CHECK:
			if(!fw_length) {
				printk(KERN_ALERT "Error: fw length is 0!\n");

				return -EFAULT;
			}
			
			if (!arg) {
				printk(KERN_ALERT "Error: userspace pointer is NULL!\n");
			
				return -EFAULT;
			}

			fw_buf = kmalloc(fw_length, GFP_KERNEL);
			if (!fw_buf) {
				printk(KERN_ALERT "Error: failed to alloc memory for fw_buf");

				return -ENOMEM;
			}

			// copy FW to buffer from userspace
			copy_length = copy_from_user(fw_buf, (char __user*)arg, fw_length);
			if (copy_length) {
				printk(KERN_ALERT "Error: failed to copy data from userspace! data left for copying: %d\n", copy_length);
			
				retval = -EFAULT;
				goto exit;	
			}

			fw_len = fw_length;

			//call usb_nfw_probe instead of renumerate to download firmware
			retval = usb_load_nandflash_firmware(us);
			if (USB_FIRMWARE_SUCCESS != retval) {
				printk(KERN_ALERT "Error: failed to programming nandflash!\n");
				release_everything(us);
			}

			// copy FW download OK flag to buffer from userspace
			copy_length = copy_to_user((void __user *)(arg+fw_length), &dlfw_flag, sizeof(char));
			if (copy_length)	{
				printk(KERN_ALERT "Error: failed to copy data to userspace! data left for copying: %d\n", copy_length);
			
				retval = -EFAULT;
				goto exit;	
			}

			break;
		default:
			printk(KERN_ALERT "Error: invalid cmd\n");
			
			return -EINVAL;
	}
exit:
	// Whether fail or succeed, we can free fw_buf, because it won't be used
	if (fw_buf) {
		kfree(fw_buf);
		fw_buf = NULL;
		fw_len = 0;
	}	
	return retval;
}

static int cy_open_stage2(struct inode *inode, struct file *file)
{
	return 0;
}

static int cy_release_stage2(struct inode *inode, struct file *file)
{
	return 0;
}

static ssize_t cy_read_stage2(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	return 0;
}

static const struct file_operations cy_fops_stage2 = {
	.owner =	THIS_MODULE,
	.read = 	cy_read_stage2,
	.ioctl =	cy_ioctl_stage2,
	.open =		cy_open_stage2,
	.release =	cy_release_stage2,
};

static struct usb_class_driver cy_class_stage2 = {
	.name =			"cy%d",
	.fops =			&cy_fops_stage2,
	.minor_base =	USB_CY_MINOR_BASE,
};

/* Probe to see if we can drive a newly-connected USB device */
static int usb_nfw_probe(struct usb_interface *intf,
			 const struct usb_device_id *id)
{
	struct usb_endpoint_descriptor *ep;
	struct usb_endpoint_descriptor *ep_in = NULL;	
	struct usb_endpoint_descriptor *ep_out = NULL;	
	int i;
	int retval = 0;

	us = kzalloc(sizeof(*us), GFP_KERNEL);
	if (!us) {
		printk(KERN_ALERT "Error: failed to allocate memory for usb data struct\n");

		return -ENOMEM;
	}

	memset(us, 0, sizeof(struct us_data));
	mutex_init(&(us->dev_mutex));
	init_waitqueue_head(&us->delay_wait);

	/* Associate the us_data structure with the USB device */
	/* Fill in the device-related fields */
	us->pusb_dev = interface_to_usbdev(intf);
	us->pusb_intf = intf;
	us->ifnum = intf->cur_altsetting->desc.bInterfaceNumber;
	US_DEBUGP("Vendor: 0x%04x, Product: 0x%04x, Revision: 0x%04x\n",
			le16_to_cpu(us->pusb_dev->descriptor.idVendor),
			le16_to_cpu(us->pusb_dev->descriptor.idProduct),
			le16_to_cpu(us->pusb_dev->descriptor.bcdDevice));
	US_DEBUGP("Interface Subclass: 0x%02x, Protocol: 0x%02x\n",
			intf->cur_altsetting->desc.bInterfaceSubClass,
			intf->cur_altsetting->desc.bInterfaceProtocol);
	/* Store our private data in the interface */
	usb_set_intfdata(intf, us);

	/* Allocate the device-related DMA-mapped buffers */
	us->cr = usb_buffer_alloc(us->pusb_dev, sizeof(*us->cr),
			GFP_KERNEL, &us->cr_dma);
	
	us->sensebuf = kmalloc(US_SENSE_SIZE, GFP_KERNEL);
	if (!us->cr) {
		printk(KERN_ALERT "Error: failed to alloc memory for usb_ctrlrequest\n");
		goto BadDevice;
	}

	us->iobuf = usb_buffer_alloc(us->pusb_dev, US_IOBUF_SIZE,
			GFP_KERNEL, &us->iobuf_dma);
	if (!us->iobuf) {
		printk(KERN_ALERT "Error: failed to alloc memory for I/O buffer\n");
		goto BadDevice;
	}
	
	if (!us->sensebuf) {
		printk(KERN_ALERT "Error: failed to alloc sense buffer\n");
		goto BadDevice;
	}

	/* get device info */
	/* SCSI */
	us->subclass = us->pusb_intf->cur_altsetting->desc.bInterfaceSubClass;
	/*Bulk-only */
	us->protocol =  us->pusb_intf->cur_altsetting->desc.bInterfaceProtocol ;
	us->flags = USB_US_ORIG_FLAGS(id->driver_info);

	if (us->flags & US_FL_IGNORE_DEVICE) {
		printk(KERN_ALERT "Error: device ignored\n");
		goto BadDevice;
	}

	/*
	 * This flag is only needed when we're in high-speed, so let's
	 * disable it if we're in full-speed
	 */
	if (us->pusb_dev->speed != USB_SPEED_HIGH)
		us->flags &= ~US_FL_GO_SLOW;
	/* Get the transport, protocol, and pipe settings */	
	if (us->protocol != US_PR_BULK){		
		printk(KERN_ALERT "Error: device protocol mismatched\n");		
		goto BadDevice;		
	} else{
		us->transport_name = "Bulk-only";
		us->transport = usb_nfw_Bulk_transport;
		us->transport_reset = usb_nfw_Bulk_reset;
	}
	/* Get the protocol settings */
	if (us->subclass != US_SC_SCSI){		
		printk(KERN_ALERT "Error: device subclass mismatched\n");		
		goto BadDevice;
		
	} else{
		us->protocol_name = "Transparent SCSI";
		us->proto_handler = usb_nfw_invoke_transport;
	}
	
	/* Get the pipe settings */
	/*
	 * Find the first endpoint of each type we need.
	 * We are expecting a minimum of 2 endpoints - in and out (bulk).
	 * An optional interrupt-in is OK (necessary for CBI protocol).
	 * We will ignore any others.
	 */
	for (i = 0; i < us->pusb_intf->cur_altsetting->desc.bNumEndpoints; i++) {
		ep = &us->pusb_intf->cur_altsetting->endpoint[i].desc;
		
		/* Is it a BULK endpoint? */
		if (usb_endpoint_xfer_bulk(ep)) {
			/* BULK in or out? */
			if (usb_endpoint_dir_in(ep)) {
				if (!ep_in)
					ep_in = ep;
			} else {
				if (!ep_out)
					ep_out = ep;
			}
		}
	}

	if (!ep_in || !ep_out ) {
		printk(KERN_ALERT "Error: failed to get endpoint in or endpoint out\n");
		goto BadDevice;	
	}
	/* Calculate and store the pipe values */
	us->send_ctrl_pipe = usb_sndctrlpipe(us->pusb_dev, 0);
	us->recv_ctrl_pipe = usb_rcvctrlpipe(us->pusb_dev, 0);
	us->send_bulk_pipe = usb_sndbulkpipe(us->pusb_dev,
		ep_out->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
	us->recv_bulk_pipe = usb_rcvbulkpipe(us->pusb_dev, 
		ep_in->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
	
	/* Acquire all the other resources and add the host */
	us->current_urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!us->current_urb) {
		printk(KERN_ALERT "Error: failed to alloc memory for URB\n");
		goto BadDevice;	
	}
	
	us->max_lun = 0;

	/* we can register the device now, as it is ready */
	retval = usb_register_dev(intf, &cy_class_stage2);
	if (retval) {
		/* something prevented us from registering this driver */
		printk(KERN_ALERT "Error: failed to get a minor for this device.");

		usb_set_intfdata(intf, NULL);
		goto BadDevice;
	}
	
	// if previous ProductID is 0x8613, then download firmware, 
	// if current ProductID is 0x4617, download firmware by ioctl
	if((USB_CY_VENDOR_ID == pre_vendor_id) && 
		(USB_CY_PRODUCT_ID_STAGE1 == pre_product_id)) {
		/* programing nand flash firmware */	
		if (USB_FIRMWARE_SUCCESS != usb_load_nandflash_firmware(us)) {
			printk(KERN_ALERT "Error: failed to programming nandflash!\n");
			goto  BadDevice;
		}
	}

	return 0;

	/* We come here if there are any problems */
BadDevice:
	printk(KERN_ALERT "Error: failed to probe device!\n");
	release_everything(us);
	
	/*please ,future fixme	*/
	return -ENOMEM;
}

/* Handle a disconnect event from the USB core */
static void usb_nfw_disconnect(struct usb_interface *intf)
{
	struct us_data *us = usb_get_intfdata(intf);

	US_DEBUGP("usb_nfw_disconnect() called\n");
	set_bit(US_FLIDX_DISCONNECTING, &us->flags);
	
	/* Stop the current URB transfer */
	if (test_and_clear_bit(US_FLIDX_URB_ACTIVE, &us->flags)) {
		US_DEBUGP("-- cancelling URB\n");
		usb_unlink_urb(us->current_urb);
	}

	/* give back our minor */
	usb_deregister_dev(intf, &cy_class_stage2);

	release_everything(us);
}

static void cy_delete(struct kref *kref)
{
	struct usb_cy *dev = to_cy_dev(kref);

	usb_put_dev(dev->udev);
	kfree(dev);
}

int cy_mem_write(struct usb_cy *dev, const unsigned char *buffer, size_t length)
{
	unsigned int len = length;		// length of FW not sent
	unsigned int to_write;			// sent FW len for one msg
	unsigned int total_writed = 0;	// sent FW len
	unsigned int actual_length;
	int ret;
	int retval;
	int i = 1;

	while (len > 0) {
		to_write = min((unsigned int)len, (unsigned int)dev->bulk_out_size);

		ret = usb_bulk_msg(dev->udev, dev->send_bulk_pipe, (void*)(buffer+total_writed),
				to_write, &actual_length, 3*HZ);
		if (ret || actual_length!=to_write) {
			printk(KERN_ERR "Error: failed to write data through endpoint_out!\n");
			
			retval = -EFAULT;
			goto exit;
		}
		len -= to_write;
		total_writed += to_write;

		i++;
	}
	retval = total_writed;
		
exit:
	return retval;
}

static ssize_t cy_write(struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
	int ret1 = 0;
	int ret2 = 0;
	int retval = 0;
	unsigned int actual_length;
	int copy_length;
	struct usb_cy *dev = file->private_data;
	int const_val_zero = 0;
	int to_write = 0;

	if (!buffer) {
		printk(KERN_ALERT "Error: userspace pointer is NULL!\n");
			
		return -EFAULT;
	}
	
	if (!count) {
		printk(KERN_ALERT "Error: firmware length is 0!\n");
			
		return -EFAULT;
	}
	
	fw_buf = kmalloc(count, GFP_KERNEL);
	if (!fw_buf) {
		printk(KERN_ALERT "Error: failed to alloc memory for firmware buffer!");

		return -ENOMEM;
	}

	dev->bulk_out_buffer = fw_buf;

	// copy FW to buffer from userspace
	copy_length = copy_from_user(dev->bulk_out_buffer, buffer, count);
	if (copy_length)	{
		printk(KERN_ALERT "Error: failed to copy data from userspace! data left for copying: %d\n", copy_length);
			
		retval = -EFAULT;
		goto exit;
	}

	// Write FW to CY ram through EP1OUT
	ret1 = cy_mem_write(dev, dev->bulk_out_buffer, count);
	if (ret1 != count) {	// if sent FW len != actural FW len
		printk(KERN_ALERT "Error: failed to write firmware to the ram of flash controller, sent FW len:%d != actural FW len:%d\n", ret1, count);

		retval = -EFAULT;
		goto exit;
	}
	
	fw_len = count;
			
	// Write 0 through EP1OUT
	ret2 = usb_bulk_msg(dev->udev, dev->send_bulk_pipe, &const_val_zero,
			to_write, &actual_length, 3*HZ);
	if (ret2 || actual_length!=to_write) {
		printk(KERN_ALERT "Error: failed to write 0 to EP1OUT to enumerate!\n");
			
		retval = -EFAULT;
		goto exit;
	}
	retval = ret1;

	return retval;
   
exit:
	// we can't free fw_buf when succeed, because it will be used in stage 2
	if (fw_buf) {
		kfree(fw_buf);
		fw_buf = NULL;
		fw_len = 0;
	}
		
	return retval;
}

static int cy_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int err = 0;
	int copy_length;

	// check cmd type
	if (_IOC_TYPE(cmd) != CY_IOC_MAGIC)	{
		printk(KERN_ALERT "Error: invalid cmd magic number!\n");

		return -EINVAL;
	}
	// check cmd nr
	if (_IOC_NR(cmd) > CY_IOC_MAXNR) {
		printk(KERN_ALERT "Error: invalid cmd number!\n");

		return -EINVAL;
	}
	// check cmd dir
	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
	else if (_IOC_DIR(cmd) & _IOC_WRITE)
		err = !access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
	if (err) {
		printk(KERN_ALERT "Error: invalid cmd direction!\n");

		return -EINVAL;
	}
	// run func according to cmd
	switch(cmd)	{
		case CY_IOC_FW_DL_OK_TEST:
			if (!arg) {
				printk(KERN_ALERT "Error: userspace pointer is NULL!\n");
			
				return -EFAULT;
			}
			
			// get semaphore here, release after usb_nfw_probe()
			if (down_interruptible(&dlfw_flag_sem)) {
				printk(KERN_ALERT "ERROR: down_interruptible!\n");
				
				return -ERESTARTSYS;
			}
			
			// copy FW download OK flag to buffer from userspace
			copy_length = copy_to_user((void __user *)arg, &dlfw_flag, sizeof(char));
			if (copy_length)	{
				printk(KERN_ERR "Error: failed to copy data to userspace! data left for copying: %d!\n", copy_length);
			
				return -EFAULT;
			}

			// release the semaphore for ioctl to access flag
			up(&dlfw_flag_sem);

			break;
		default:
			printk(KERN_ALERT "Error: invalid cmd!\n");
			
			return -EINVAL;
	}

	return 0;
}

static int cy_open(struct inode *inode, struct file *file)
{
	struct usb_cy *dev;
	struct usb_interface *interface;
	int subminor;
	int retval = 0;

	subminor = iminor(inode);

	interface = usb_find_interface(&cy_driver, subminor);
	if (!interface) {
		printk(KERN_ALERT "Error: failed to find device for minor %d!", subminor);
		
		retval = -ENODEV;
		goto exit;
	}

	dev = usb_get_intfdata(interface);
	if (!dev) {
		printk(KERN_ALERT "Error: dev is NULL!\n");
		
		retval = -ENODEV;
		goto exit;
	}

	/* increment our usage count for the device */
	kref_get(&dev->kref);

	/* lock the device to allow correctly handling errors
	 * in resumption */
	mutex_lock(&dev->io_mutex);

	/* save our object in the file's private structure */
	file->private_data = dev;
	
	mutex_unlock(&dev->io_mutex);

exit:
	return retval;
}

static int cy_release(struct inode *inode, struct file *file)
{
	struct usb_cy *dev;
	int retval = 0;
	
	dev = (struct usb_cy *)file->private_data;
	if (dev == NULL) {
		printk(KERN_ALERT "Error: dev is NULL!\n");
		
		retval = -ENODEV;
		goto exit;
	}

	/* decrement the count on our device */
	kref_put(&dev->kref, cy_delete);

exit:
	return retval;
}

static const struct file_operations cy_fops = {
	.owner =	THIS_MODULE,
	.write =	cy_write,
	.ioctl =	cy_ioctl,
	.open =		cy_open,
	.release =	cy_release,
};

static struct usb_class_driver cy_class = {
	.name =			"cy%d",
	.fops =			&cy_fops,
	.minor_base =	USB_CY_MINOR_BASE,
};

static int cy_stage1_probe(struct usb_interface *interface,
    const struct usb_device_id *dev_id)
{
	struct usb_cy *dev;
	struct usb_host_interface *iface_desc;
	struct usb_host_interface *alt;
	struct usb_endpoint_descriptor *endpoint;
	size_t buffer_size;
	int i;
	int ret;
	int retval;

	/* allocate memory for our device state and initialize it */
	dev = kzalloc(sizeof(*dev), GFP_KERNEL);
	if (!dev) {
		printk(KERN_ALERT "Error: failed to alloc memory for dev!");
		
		goto error;
	}
	kref_init(&dev->kref);
	mutex_init(&dev->io_mutex);
	spin_lock_init(&dev->err_lock);
	init_usb_anchor(&dev->submitted);

	dev->udev = usb_get_dev(interface_to_usbdev(interface));
	dev->interface = interface;

	/* set the cur_setting to altsetting[1] */
	alt = interface->altsetting;
	ret = usb_set_interface(dev->udev, alt->desc.bInterfaceNumber, 1);
	if (ret < 0) {
		printk(KERN_ALERT "Error: failed to change cur_altseting!\n");
			
		goto error;
	}
		
	/* set up the endpoint information */
	/* use only the first bulk-in and bulk-out endpoints */
	iface_desc = interface->cur_altsetting;
	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;

		if (!dev->bulk_out_endpointAddr &&
		    usb_endpoint_is_bulk_out(endpoint)) {
			/* we found a bulk out endpoint */
			buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
			dev->bulk_out_size = 64;
			dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
		}
	}

	if (!dev->bulk_out_endpointAddr) {
		printk(KERN_ALERT "Error: failed to find bulk-out endpoint!\n");

		goto error;
	}

	/* save our data pointer in this interface device */
	usb_set_intfdata(interface, dev);

	/* Calculate and store the pipe values */
	dev->send_bulk_pipe = usb_sndbulkpipe(dev->udev,
		dev->bulk_out_endpointAddr & USB_ENDPOINT_NUMBER_MASK);

	/* we can register the device now, as it is ready */
	retval = usb_register_dev(interface, &cy_class);
	if (retval) {
		/* something prevented us from registering this driver */
		printk(KERN_ALERT "Error: failed to get a minor for this device!\n");

		usb_set_intfdata(interface, NULL);
	
		goto error;
	}
	
	return retval;
	
error:
	/* this frees allocated memory */
	if (dev)
		kref_put(&dev->kref, cy_delete);

	printk(KERN_ALERT "Error: failed to probe device!\n");
	
	return retval;
}

static int cy_probe(struct usb_interface *interface,
    const struct usb_device_id *dev_id)
{
	int retval = 0;

	if (USB_CY_PRODUCT_ID_STAGE1 == dev_id->idProduct) {
		// lock the semaphore for access fw_dl_ok flag
		init_MUTEX_LOCKED(&dlfw_flag_sem);
		
		retval = cy_stage1_probe(interface,dev_id);
	} else if(USB_CY_PRODUCT_ID_STAGE2 == dev_id->idProduct) {
		if(bulk_reset_flags){
			/* clear flag, bulk-only reset ,terminate process */
			bulk_reset_flags = 0;

			// release the semaphore for ioctl to access flag avoid deadlock
			if(USB_CY_PRODUCT_ID_STAGE1 == pre_product_id)
				up(&dlfw_flag_sem);
				
			return -1;
		}
		/* stage2 nand flash write start from here */
		retval = usb_nfw_probe(interface, dev_id);

		/* release the semaphore for ioctl to access flag
		 * if we erase all blocks and renumerate, the following semaphore won't be initialized
		 */
		if(USB_CY_PRODUCT_ID_STAGE1 == pre_product_id)
			up(&dlfw_flag_sem);
	}

	pre_vendor_id = dev_id->idVendor;
	pre_product_id  = dev_id->idProduct;

	return retval;
}

static void cy_stage1_disconnect(struct usb_interface * interface)
{
	struct usb_cy *dev;

	dev = usb_get_intfdata(interface);
	usb_set_intfdata(interface, NULL);

	/* give back our minor */
	usb_deregister_dev(interface, &cy_class);

	/* prevent more I/O from starting */
	mutex_lock(&dev->io_mutex);
	dev->interface = NULL;
	mutex_unlock(&dev->io_mutex);

	usb_kill_anchored_urbs(&dev->submitted);

	/* decrement our usage count */
	kref_put(&dev->kref, cy_delete);
}

static void cy_disconnect(struct usb_interface *interface)
{
	struct usb_device *udevice = interface_to_usbdev(interface);
	
	if (USB_CY_PRODUCT_ID_STAGE1 == udevice->descriptor.idProduct)
 		cy_stage1_disconnect(interface);
	else if (USB_CY_PRODUCT_ID_STAGE2 == udevice->descriptor.idProduct) {
		usb_nfw_disconnect(interface);

		/*free fw_buf*/
		if(fw_buf) {
			kfree(fw_buf);
			fw_buf = NULL;
			fw_len = 0;
		}
		fw_length = 0;
	}
}

static int cy_pre_reset(struct usb_interface *iface)
{
	struct us_data *us = usb_get_intfdata(iface);

	/* Make sure no command runs during the reset */
	mutex_lock(&us->dev_mutex);
	
	return 0;
}

static int cy_post_reset(struct usb_interface *iface)
{
	struct us_data *us = usb_get_intfdata(iface);

	mutex_unlock(&us->dev_mutex);
	
	return 0;
}

static struct usb_driver cy_driver = {
	.name =			"cy",
	.probe =		cy_probe,
	.disconnect =	cy_disconnect,
	.pre_reset =	cy_pre_reset,
	.post_reset =	cy_post_reset,
	.id_table =		cy_table,
};

static int __init cy_init(void)
{
	int retval;
	retval = usb_register(&cy_driver);
	if (retval != 0) {
		printk(KERN_ALERT "Error: failed to register driver cy.\n");
	}
	return retval;
}

static void __exit cy_exit(void)
{
	usb_deregister(&cy_driver);
}

module_init(cy_init);
module_exit(cy_exit);

MODULE_LICENSE("GPL");
