/* 2014-08-28: File changed by Sony Corporation */
/*
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/dma-mapping.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/cdev.h>
#include <linux/mutex.h>
#include <linux/mm.h>
#include <linux/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <linux/slab.h>

#include "bio_data_pool.h"

#ifdef CONFIG_SNSC_HSS
#include <linux/snsc_hss.h>
#endif /* #ifdef CONFIG_SNSC_HSS */



/* trace-debug switch for bio data pool module.  */
#define CONFIG_BIO_DATA_DEBUG


/* trace function of bio data pool module. */
#ifdef CONFIG_BIO_DATA_DEBUG
    #define bio_pool_debug(fmt, args...)      \
          printk(KERN_DEBUG "%s: " fmt, __func__, ## args)

    #define bio_pool_warning(fmt, args...)      \
          printk(KERN_ERR "%s: " fmt, __func__, ## args)

#else

  #define bio_pool_debug(fmt, args...)
  #define bio_pool_warning(fmt, args...)

#endif

  #define bio_pool_error(fmt, args...)      \
          printk(KERN_ERR "%s: " fmt, __func__, ## args)

/*  */
static struct cdev *g_bio_pool_dev = NULL;

/*  */
static unsigned long g_bio_pool_base = 0;

/*  */
#ifdef BIO_POOL_MALLOC

static unsigned long g_mifmc_pool_base = 0;     /*  */

#else

static unsigned long g_bio_pool_phy_start = 0;  /*  */

#endif

#ifdef CONFIG_SNSC_HSS
struct hss_bio_d_info {
    struct hss_node *node;
};
struct hss_bio_d_info   *p_hss_bio_d_info;
#endif  /* #ifdef CONFIG_SNSC_HSS */

static int bio_pool_open (struct inode *inode, struct file *filp)
{
    bio_pool_debug(" entry test. \n");

    return 0;
}


/**
 *  @fn hfdma_ioctl
 *
 *  @brief  io control device by system call
 *
 *  @param
 *
 *  @return CXD9225GG_STATUS status
 */
static int bio_pool_ioctl (struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
    void __user *usermem = (void __user *) arg;
    bio_buf_desc_t buf_desc;
    unsigned long size;
    unsigned long len;
    int ret;

    bio_pool_debug(" entry log:  cmd:0x%0x, arg:0x%0x. \n", cmd, (unsigned int)arg);


    /* validate IOCTL cmd id and cmd */
    if( BIO_IO_MAGIC != _IOC_TYPE(cmd) )
    {
        bio_pool_error(" TK1_IO_MAGIC != _IOC_TYPE(cmd). \n");
        return -EINVAL;
    }

    /*  */
    if( BIO_IO_MAX_NR <= _IOC_NR(cmd) )
    {
        bio_pool_error(" TK1_IO_MAX_NR <= _IOC_NR(cmd). \n");
        return -EINVAL;
    }

    /* if this cmd has associated data (in or out), make sure we can access it */
    size = _IOC_SIZE(cmd);
    if(size)
    {
        ret = !access_ok( VERIFY_WRITE | VERIFY_READ, usermem, size );
        if(ret)
        {
            bio_pool_error(" arg-usermem can NOT be access_ok. \n");
            return -EINVAL;
        }
    }

    /* command-handling. */
    switch (cmd)
    {
        /*  */
        case BIO_IO_CMD_PHY_ADDR:
            bio_pool_debug("cmd--BIO_IO_CMD_PHY_ADDR. \n");

        #ifdef BIO_POOL_MALLOC
            /*  */
            size = __pa(g_bio_pool_base);
        #else
            size = g_bio_pool_phy_start;
        #endif

            len = copy_to_user(usermem, &size, sizeof(unsigned long));
            {
                ret = 0;
            }

            break;

        /*  */
        case BIO_IO_CMD_MEM_SIZE:
            bio_pool_debug("cmd--BIO_IO_CMD_MEM_SIZE. \n");

            /*  */
            size = BIO_DATA_POOL_SIZE;

            len = copy_to_user(usermem, &size, sizeof(unsigned long));
            {
                ret = 0;
            }

            break;

        /*  */
        case BIO_IO_CMD_FLUSH:
            bio_pool_debug("cmd--BIO_IO_CMD_FLUSH. \n");

            len = copy_from_user(&buf_desc, usermem, sizeof(bio_buf_desc_t));

            {
                ret = 0;
            }

            break;

/*  */
#ifdef BIO_POOL_MALLOC

    /*  */
    case BIO_IO_CMD_PHY_ADDR_MIF:
        bio_pool_debug("cmd--BIO_IO_CMD_PHY_ADDR_MIF. \n");

        /*  */
        size = __pa(g_mifmc_pool_base);

        len = copy_to_user(usermem, &size, sizeof(unsigned long));
        {
            ret = 0;
        }

        break;

    /*  */
    case BIO_IO_CMD_MEM_SIZE_MIF:
        bio_pool_debug("cmd--BIO_IO_CMD_MEM_SIZE_MIF. \n");

        /*  */
        size = MIFMC_MEM_POOL_SIZE;

        len = copy_to_user(usermem, &size, sizeof(unsigned long));
        {
            ret = 0;
        }

        break;

#endif

    default:

        bio_pool_error("cmd:0x%0x has no matched handler. \n", (unsigned int)cmd);

        ret = -1;
        break;
    }

    bio_pool_debug("exit test. \n");

    return ret;
}

static int bio_pool_release (struct inode *inode, struct file *filp)
{
    bio_pool_debug(" entry test. \n");

    return 0;
}


/*  */
static int bio_pool_mmap(struct file *file, struct vm_area_struct *vma)
{
    unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
    unsigned long size = vma->vm_end - vma->vm_start;

    bio_pool_debug("entry test, size:0x%0x, offset:0x%0x. \n",
                   (unsigned int)size, (unsigned int)offset);

    /*support reserved memory*/
    if(offset == 0)
    {
        /* check the request-map-size. */
        if(size > BIO_DATA_POOL_SIZE)
        {
            bio_pool_error("size:0x%0x > BIO_DATA_POOL_SIZE. \n", (unsigned int)size);
            return -ENXIO;
        }

        vma->vm_flags |= VM_LOCKED;

        /* make it uncacheable... */
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

    #ifdef BIO_POOL_MALLOC
        /* map physical memory to user space */
        if(remap_pfn_range(vma, vma->vm_start, __pa(g_bio_pool_base) >> PAGE_SHIFT, size, vma->vm_page_prot))
    #else
        if(remap_pfn_range(vma, vma->vm_start, g_bio_pool_phy_start >> PAGE_SHIFT, size, vma->vm_page_prot))
    #endif
        {
            bio_pool_error("remap_pfn_range failed. \n");
            return -ENXIO;
        }

    }
#ifdef BIO_POOL_MALLOC
    else if (offset == (unsigned long)BIO_DATA_POOL_SIZE)
    {
        /* check the request-map-size. */
        if(size > (unsigned long)MIFMC_MEM_POOL_SIZE)
        {
            bio_pool_error("size:0x%0x > MIFMC_MEM_POOL_SIZE. \n", (unsigned int)size);
            return -ENXIO;
        }

        vma->vm_flags |= VM_LOCKED;

        /* make it uncacheable... */
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

        /* map physical memory to user space */
        if(remap_pfn_range(vma, vma->vm_start, __pa(g_mifmc_pool_base) >> PAGE_SHIFT, size, vma->vm_page_prot))
        {
            bio_pool_error("remap_pfn_range failed for mifmc_mem. \n");
            return -ENXIO;
        }
    }
#endif
    else //if()
    {
        bio_pool_error("offset != 0, so need another remap. \n");
        return -1;
    }

    bio_pool_debug("exit succeed. \n");

    return 0;
}


struct file_operations bio_pool_fops =
{
    .owner      = THIS_MODULE,
    .mmap       = bio_pool_mmap,
    .llseek     = NULL,
    .read       = NULL,
    .write      = NULL,
    .ioctl      = bio_pool_ioctl,
    .open       = bio_pool_open,
    .release    = bio_pool_release,
};


#ifdef CONFIG_SNSC_HSS
static int bio_d_hss_node_suspend(struct hss_node *node)
{
    /* TENTATIVE : no procedure */
    return 0;
}

static int bio_d_hss_node_resume(struct hss_node *node)
{
    /* TENTATIVE : no procedure */
    return 0;
}

static struct hss_node_ops hss_bio_d_node_ops = {
    .suspend = bio_d_hss_node_suspend,
    .resume = bio_d_hss_node_resume,
};

static int bio_d_hss_node_init(void)
{
    struct hss_node *node;
    int error = 0;

    /* alloc for emif_dl info */
    p_hss_bio_d_info = kzalloc(sizeof(*p_hss_bio_d_info), GFP_KERNEL);
    if (!p_hss_bio_d_info) {
        return -ENOMEM;
    }

    /*
     * alloc the HSS node
     */
    node = hss_alloc_node(BIO_D_HSS_NODE_NAME);
    if (!node) {
        kfree(p_hss_bio_d_info);
        return -ENOMEM;
    }

    /*
     * save struct to HSS
     */
    hss_set_private(node, p_hss_bio_d_info);
    p_hss_bio_d_info->node = node;

    /*
     * Register the node.  If you want to get the callbacks to be called
     * before the specified node's one, specify the third parameter with
     * the name of the parent node. NULL means for no parent needed.
     */

    /* register HSS Platform for suspend/resume */
    error = hss_register_node(node, &hss_bio_d_node_ops, BIO_D_HSS_NODE_PARENT);

    if (error) {
        hss_free_node(node);
        kfree(p_hss_bio_d_info);
    }

    return error;
}

static void bio_d_hss_node_exit(void)
{
	hss_unregister_node(p_hss_bio_d_info->node);
	hss_free_node(p_hss_bio_d_info->node);
	kfree(p_hss_bio_d_info);
}
#endif /* #ifdef CONFIG_SNSC_HSS */


/*  */
static int __init bio_module_init(void)
{
    int ret;

    bio_pool_debug("entry test. \n");

#ifdef CONFIG_SNSC_HSS
    /* register driver to hss */
    ret = bio_d_hss_node_init();
    if(ret != 0) {
        return ret;
    }
#endif /* #ifdef CONFIG_SNSC_HSS */


#ifdef BIO_POOL_MALLOC

#if 0
    /*  */
    g_bio_pool_base = __get_free_pages(GFP_KERNEL, get_order(BIO_DATA_POOL_SIZE));
    if (0 == g_bio_pool_base)
    {
        bio_pool_error("__get_free_pages for bio-pool failed.\n");
        return -ENOMEM;
    }

    /*  */
    g_mifmc_pool_base = __get_free_pages(GFP_KERNEL, get_order(MIFMC_MEM_POOL_SIZE));
    if (0 == g_mifmc_pool_base)
    {
        bio_pool_error("__get_free_pages for mifmc-pool failed.\n");
        return -ENOMEM;
    }
#else
    g_bio_pool_base   = BIO_DATA_POOL_START_ADDR;
    g_mifmc_pool_base = BIO_DATA_POOL_START_ADDR + BIO_DATA_POOL_SIZE;
#endif

#else

    /* map physical memory to kernel space
     * start phy addr:  (252M - )
     * size:
     */
    g_bio_pool_base = (unsigned long)ioremap(PHYS_OFFSET + (unsigned long)(0x0FC00000) - BIO_DATA_POOL_SIZE,
                                            BIO_DATA_POOL_SIZE);
    if(0 == g_bio_pool_base)
    {
            bio_pool_error("ioremap failed. \n");
            return -ENOMEM;
    }

    /*  */
    SetPageReserved(virt_to_page(g_bio_pool_base));

    /*  */
    g_bio_pool_phy_start = PHYS_OFFSET + (unsigned long)(0x0FC00000) - BIO_DATA_POOL_SIZE;

#endif



    g_bio_pool_dev = cdev_alloc();
    if (NULL == g_bio_pool_dev)
    {
        bio_pool_error("cdev_alloc for bio-pool failed.\n");
        free_pages(g_bio_pool_base, get_order(BIO_DATA_POOL_SIZE));
        g_bio_pool_base = 0;
        return -ENOMEM;
    }

    /*  */
    cdev_init(g_bio_pool_dev, &bio_pool_fops);
    g_bio_pool_dev->dev = MKDEV(BIO_DATA_POOL_MAJOR, 0);
    g_bio_pool_dev->owner = bio_pool_fops.owner;

    /* add the c-dev to system. */
    ret = cdev_add(g_bio_pool_dev, g_bio_pool_dev->dev, 1);
    if (ret)
    {
        bio_pool_error("cdev_add bio-pool to system failed.\n");

        free_pages(g_bio_pool_base, get_order(BIO_DATA_POOL_SIZE));
        g_bio_pool_base = 0;

        cdev_del(g_bio_pool_dev);
        g_bio_pool_dev = NULL;

        return -ENOMEM;
    }

    bio_pool_debug("module init succeed.\n");

    return 0;
}

static void __exit bio_module_exit(void)
{
    cdev_del(g_bio_pool_dev);
    g_bio_pool_dev = NULL;

    /* release the pages. */
    free_pages(g_bio_pool_base, get_order(BIO_DATA_POOL_SIZE));

    /* iounmap for reserved memory. */
    if (g_bio_pool_base)
    {
        /*  */
        ClearPageReserved(virt_to_page(g_bio_pool_base));

        iounmap((void *)g_bio_pool_base);

        g_bio_pool_base = 0;
    }

#ifdef CONFIG_SNSC_HSS
    /* unregister driver from hss */
    bio_d_hss_node_exit();
#endif /* #ifdef CONFIG_SNSC_HSS */

    bio_pool_debug("module exit.\n");

    return;
}

unsigned long bio_pool_virt2phy(unsigned long virt_addr)
{
#ifdef BIO_POOL_MALLOC
    if( ( (virt_addr >= g_bio_pool_base) && ((virt_addr - g_bio_pool_base) < BIO_DATA_POOL_SIZE) )
        ||
        ( (virt_addr >= g_mifmc_pool_base) && ((virt_addr - g_mifmc_pool_base) < MIFMC_MEM_POOL_SIZE) )
       )
    {
        return __pa(virt_addr);
    }
    else
    {
        return 0;
    }
#else
    if( (virt_addr >= g_bio_pool_base) && ((virt_addr - g_bio_pool_base) < BIO_DATA_POOL_SIZE) )
    {
        return (virt_addr - g_bio_pool_base + g_bio_pool_phy_start);
    }
    else
    {
        return 0;
    }
#endif
}
EXPORT_SYMBOL(bio_pool_virt2phy);
unsigned long bio_pool_phy2virt(unsigned long phy_addr)
{
#ifdef BIO_POOL_MALLOC
    if( ( (phy_addr >= __pa(g_bio_pool_base)) && ((phy_addr - __pa(g_bio_pool_base)) < BIO_DATA_POOL_SIZE) )
        ||
        ( (phy_addr >= __pa(g_mifmc_pool_base)) && ((phy_addr - __pa(g_mifmc_pool_base)) < MIFMC_MEM_POOL_SIZE) )
       )
    {
        return (unsigned long)__va(phy_addr);
    }
    else
    {
        return 0;
    }
#else
    if( (phy_addr >= g_bio_pool_phy_start) && ((phy_addr - g_bio_pool_phy_start) < BIO_DATA_POOL_SIZE) )
    {
        return (phy_addr - g_bio_pool_phy_start + g_bio_pool_base);
    }
    else
    {
        return 0;
    }
#endif
}
EXPORT_SYMBOL(bio_pool_phy2virt);
MODULE_LICENSE("GPL");

module_init(bio_module_init);
module_exit(bio_module_exit);


