/* 2014-08-28: File changed by Sony Corporation */
/*
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <mach/hardware.h>
#include <asm/mach/irq.h>

#include "irq.h"

#ifdef  __IRQ_DEBUG__
#define irq_printl(level, format, args...) \
		printk(level "%s() %d : " format "\n", __func__, __LINE__, ##args)
#else
#define irq_printl(level, format, args...)
#endif


/* generate an interrupt by setting SWI registers.
 * SWC registers are cleared in eoi function. */
int cxd9225gg_generate_irq(unsigned int irq)
{
    unsigned int reg_offset = (irq / CXD9225GG_INT_CH_BLK);
    unsigned int bit_set = 1 << (irq % CXD9225GG_INT_CH_BLK);

    /* Parameter check */
    if ((irq < 0) || (irq >= CXD9225GG_INT_SOURCES_NUM)) {
        return -1;
    }

    /* Generate S/W interrupt */
    writel(bit_set, CXD9225GG_INTC_SWI_VIRT(reg_offset));

	return 0;
}
EXPORT_SYMBOL(cxd9225gg_generate_irq);

/* cancel an interrupt which is generated by setting SWI registers.
 * this function should be called in eoi functions */
int cxd9225gg_cancel_irq(unsigned int irq)
{
	unsigned int reg_offset = (irq / CXD9225GG_INT_CH_BLK);
	unsigned int bit_set = 1 << (irq % CXD9225GG_INT_CH_BLK);

	/* Parameter check */
	if ((irq < 0) || (irq >= CXD9225GG_INT_SOURCES_NUM)) {
		return -1;
	}

	/* Cancel S/W interrupt */
    writel(bit_set, CXD9225GG_INTC_SWC_VIRT(reg_offset));

	return 0;
}
EXPORT_SYMBOL(cxd9225gg_cancel_irq);

static void cxd9225gg_mask_irq(unsigned int irq)
{
	u32 *reg;
	u32 regval;

	reg = (u32 *)CXD9225GG_INTC_IEC_VIRT((irq / (sizeof(u32) * 8)));
	regval = (u32)1 << (irq % (sizeof(u32) * 8));

	writel(regval, reg);
}

static void cxd9225gg_unmask_irq(unsigned int irq)
{
	u32 *reg;
	u32 regval;

	reg = (u32 *)CXD9225GG_INTC_IEN_VIRT((irq / (sizeof(u32) * 8)));
	regval = (u32)1 << (irq % (sizeof(u32) * 8));
	writel(regval, reg);
}

static void cxd9225gg_ack_irq_level(unsigned int irq)
{
	/* ack the interrupt */
	writel(0, CXD9225GG_INTC_HVA_VIRT);

	/* clear swi */
	cxd9225gg_cancel_irq(irq);
}

static void cxd9225gg_ack_irq_edge(unsigned int irq)
{
	u32 *reg;

	/* ack the interrupt */
	reg = (u32 *)CXD9225GG_INTC_PIC_VIRT((irq / (sizeof(u32) * 8)));
	writel((1 << (irq % (sizeof(u32) * 8))), reg);

	writel(0, CXD9225GG_INTC_HVA_VIRT);

	/* clear swi */
	cxd9225gg_cancel_irq(irq);
}

void cxd9225gg_set_irq_bit_clear(unsigned int irq)
{
    cxd9225gg_ack_irq_edge(irq);
}

/* return irq pulse type, level(0) or edge(1).*/
int cxd9225gg_get_irq_pulse(unsigned int irq)
{
	u32 *reg;
	u32 bit, val;
	int edge;

	reg = (u32 *)CXD9225GG_INTC_PLS_VIRT((irq / (sizeof(u32) * 8)));
	val = readl(reg);

	bit = (u32)1 << (irq % (sizeof(u32) * 8));

	edge = (val & bit) ? 1 : 0;

	return edge;
}

/* select irq pulse as level(0) or edge(1).*/
void cxd9225gg_set_irq_pulse(unsigned int irq, int edge)
{
	u32 *reg;
	u32 bit, val;

	edge = edge ? 1 : 0;

	reg = (u32 *)CXD9225GG_INTC_PLS_VIRT((irq / (sizeof(u32) * 8)));
	val = readl(reg);

	bit = (u32)1 << (irq % (sizeof(u32) * 8));

	if (edge) {
		val |= bit;
	} else {
		val &= ~bit;
	}

	irq_printl(KERN_INFO, "edge=%d, bit=0x%08x, val=0x%08x", edge, bit, val);
	writel(val, reg);
}

/*
 * rising_or_high:
 * for edge:  0 = falling, 1 = rising, 2 = both
 * for level: 0 = low, 1 = high
 */
void cxd9225gg_set_irq_ctrl(unsigned int irq, int rising_or_high)
{
	u32 *reg;
	u32 val;
	int shift;
	int edge = 0;

	edge = cxd9225gg_get_irq_pulse(irq);

	if (edge) {
		reg = (u32 *)CXD9225GG_INTC_EDGC_VIRT((irq / (sizeof(u32) * 4)));
	} else {
		reg = (u32 *)CXD9225GG_INTC_LVLC_VIRT((irq / (sizeof(u32) * 4)));
	}

	/* set Level Control or Edge Control register */
	shift = (irq % (sizeof(u32) * 4)) * 2;
	val = readl(reg);
	val &= ~((u32)3 << shift);

	/*
	 * edge:  01 = rising edge, 10 = falling edge
	 * level: 01 = high active, 10 = low active
	 */
	if (rising_or_high == 1 )
		val |= (1 << shift);
	else if ( rising_or_high == 0 )
		val |= (2 << shift);
	else if ( rising_or_high == 2 )
		val |= (3 << shift);

	irq_printl(KERN_INFO, "edge=%d, reg=0x%p, rising_or_high=%d",
			edge, reg, rising_or_high);
	writel(val, reg);
}

int cxd9225gg_set_irq_type(unsigned int irq, unsigned int type)
{
	/* : set the interrupt type */

	u32 *reg;
	u32 bit, val;
	int shift;
	int edge = 0;
	int rising_or_high = 0;

	switch (type) {
	case IRQ_TYPE_EDGE_RISING:
		edge = 1;
		rising_or_high = 1;
		break;
	case IRQ_TYPE_EDGE_FALLING:
		edge = 1;
		rising_or_high = 0;
		break;
	case IRQ_TYPE_EDGE_BOTH:
		edge = 1;
		rising_or_high = 2;
		break;
	case IRQ_TYPE_LEVEL_HIGH:
		edge = 0;
		rising_or_high = 1;
		break;
	case IRQ_TYPE_LEVEL_LOW:
		edge = 0;
		rising_or_high = 0;
		break;
	default:
		/* it is a bad type */
		return -EINVAL;
	};

	/* set Pulse Select register */
	reg = (u32 *)CXD9225GG_INTC_PLS_VIRT((irq / (sizeof(u32) * 8)));
	val = readl(reg);

	bit = (u32)1 << (irq % (sizeof(u32) * 8));

	if (edge) {
		val |= bit;
		writel(val, reg);

		reg = (u32 *)CXD9225GG_INTC_EDGC_VIRT((irq / (sizeof(u32) * 4)));
	} else {
		val &= ~bit;
		writel(val, reg);

		reg = (u32 *)CXD9225GG_INTC_LVLC_VIRT((irq / (sizeof(u32) * 4)));
	}

	/* set Level Control or Edge Control register */
	shift = (irq % (sizeof(u32) * 4)) * 2;
	val = readl(reg);
	val &= ~((u32)3 << shift);

	/*
	 * edge:  01 = rising edge, 10 = falling edge
	 * level: 01 = high active, 10 = low active
	 */
	if (rising_or_high == 1 )
		val |= (1 << shift);
	else if ( rising_or_high == 0 )
		val |= (2 << shift);
	else if ( rising_or_high == 2 )
		val |= (3 << shift);

	irq_printl(KERN_INFO, "edge=%d, reg=0x%p, rising_or_high=%d",
			edge, reg, rising_or_high);

	writel(val, reg);

	return 0;
}
EXPORT_SYMBOL(cxd9225gg_get_irq_pulse);
EXPORT_SYMBOL(cxd9225gg_set_irq_pulse);
EXPORT_SYMBOL(cxd9225gg_set_irq_ctrl);
EXPORT_SYMBOL(cxd9225gg_set_irq_type);
EXPORT_SYMBOL(cxd9225gg_set_irq_bit_clear);

static struct irq_chip cxd9225gg_irq_chip_edge = {
	.name    = "INTC-edge",
	.mask    = cxd9225gg_mask_irq,
	.unmask  = cxd9225gg_unmask_irq,
	.enable  = cxd9225gg_unmask_irq,
	.disable = cxd9225gg_mask_irq,
	.eoi     = cxd9225gg_ack_irq_edge,

	/* : allow changing irq detection type dynamically */
	.set_type = cxd9225gg_set_irq_type,
};

static struct irq_chip cxd9225gg_irq_chip_level = {
	.name    = "INTC-level",
	.mask    = cxd9225gg_mask_irq,
	.unmask  = cxd9225gg_unmask_irq,
	.enable  = cxd9225gg_unmask_irq,
	.disable = cxd9225gg_mask_irq,
	.eoi     = cxd9225gg_ack_irq_level,
};

struct irq_info {
	int detect;
	int valid;
};

static struct irq_info __initdata irq_info_array[NR_IRQS] = {
	[INT_TIMER0]    = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_TIMER1]    = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_TIMER2]    = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_TIMER3]    = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_TIMER4]    = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_TIMER5]    = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_TIMER6]    = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_TIMER7]    = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_TIMER8]    = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_UART0]     = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},
	[INT_PCI_INTA]  = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},
	[INT_PCI_INTB]  = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},
	[INT_PCI_INTC]  = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},
	[INT_PCI_INTD]  = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},

	[INT_SYSCON]  = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},
	[INT_SYSCON_ERR]  = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},

	[INT_STCC]  = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},

	/*
	* Hardware Error Report Interrupt.
	* SYSCON needs to monitor all these Error-Report IRQs individually.
	* So if any module wants to request these IRQs, it must request them
	* using IRQF_SHARED flag.
	*/
	[INT_X2H_ERRINT] =    {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_H2X_ERRINT] =    {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_DVC_ERR] =       {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_PCIEX0_ERR] =    {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},
	[INT_PCIEX0_AXIERR] = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},
	[INT_PCIEX1_ERR] =    {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},
	[INT_PCIEX1_AXIERR] = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},
	[INT_BSIF_ERR]	=     {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},

	/* for BSIF */
	[INT_BSIF]  = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},

	/* : the following 6 GPINTs are For ExpressCard Control */
	/* for EXPCD slot0 USB */
	[INT_GPIO0]    = {CXD9225GG_IRQDETECT_EDGE_BOTH, 1},
	/* for EXPCD slot0 PCIEX */
	[INT_GPIO1]    = {CXD9225GG_IRQDETECT_EDGE_BOTH, 1},
	/* for EXPCD slot0 OverCurrent */
	[INT_GPIO2]    = {CXD9225GG_IRQDETECT_EDGE_FALLING, 1},
	/* for EXPCD slot1 USB */
	[INT_GPIO4]    = {CXD9225GG_IRQDETECT_EDGE_BOTH, 1},
	/* for EXPCD slot1 PCIEX */
	[INT_GPIO5]    = {CXD9225GG_IRQDETECT_EDGE_BOTH, 1},
	/* for EXPCD slot1 OverCurrent */
	[INT_GPIO6]    = {CXD9225GG_IRQDETECT_EDGE_FALLING, 1},

	[INT_GPIO16]    = {CXD9225GG_IRQDETECT_EDGE_FALLING, 1},

	//for IRQ test buttom
	[INT_GPIO19]    = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},

	[INT_HFDMA_CH0] = {CXD9225GG_IRQDETECT_LEVEL_HIGH, 1},
	[INT_HFDMA_CH1] = {CXD9225GG_IRQDETECT_LEVEL_HIGH, 1},
	[INT_HFDMA_CH2] = {CXD9225GG_IRQDETECT_LEVEL_HIGH, 1},
	[INT_HFDMA_CH3] = {CXD9225GG_IRQDETECT_LEVEL_HIGH, 1},

	/* add for dv*/
	[INT_DVC_INT0] = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_DVC_INT1] = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_DVC_INT2] = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_DVC_ERR]  = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},

	/* add for avsel*/
	[INT_GPIO8]  = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},
	[INT_CSI]    = {CXD9225GG_IRQDETECT_LEVEL_HIGH, 1},     // add for csi Int
	[INT_GPIO17] = {CXD9225GG_IRQDETECT_EDGE_FALLING, 1},   // add for FPGA err Int

	[INT_GPIO18] = {CXD9225GG_IRQDETECT_EDGE_FALLING, 1},

	/* for PCIE Host Controller Int */
	[INT_PCIEX0_INTA] = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},
	[INT_PCIEX1_INTA] = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},
	[INT_PCIEX2_INTA] = {CXD9225GG_IRQDETECT_LEVEL_HIGH,  1},

	/*  multi-CPU communation */
	[INT_CROSS_INT0]  = {CXD9225GG_IRQDETECT_EDGE_RISING, 1}, /**CPU1 SEND*/
	[INT_CROSS_INT1]  = {CXD9225GG_IRQDETECT_EDGE_RISING, 1}, /**CPU1 RECV*/

	/* for USB Mass Storage Device */
	[INT_USBDEV]     = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
	[INT_USBDEV_EPC] = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},

    [INT_USBDEV] = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
    [INT_USBDEV_EPC] = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},

    /* : for CSI GPIO request */
	[INT_GPIO12] = {CXD9225GG_IRQDETECT_EDGE_FALLING, 1},

	[INT_IIC] = {CXD9225GG_IRQDETECT_EDGE_RISING, 1},
};

static int __init cxd9225gg_irq_valid(unsigned int irq)
{
	return irq_info_array[irq].valid;
}

static int __init cxd9225gg_irq_level(unsigned int irq)
{
	int ret = 0;

	switch (irq_info_array[irq].detect) {
	case CXD9225GG_IRQDETECT_EDGE_RISING:
	case CXD9225GG_IRQDETECT_EDGE_FALLING:
	case CXD9225GG_IRQDETECT_EDGE_BOTH:
		ret = 0;
		break;
	case CXD9225GG_IRQDETECT_LEVEL_HIGH:
	case CXD9225GG_IRQDETECT_LEVEL_LOW:
		ret = 1;
		break;
	};
	return ret;
}

static void __init cxd9225gg_set_irq_detection(unsigned int irq)
{

	switch (irq_info_array[irq].detect) {
	case CXD9225GG_IRQDETECT_EDGE_RISING:
		cxd9225gg_set_irq_type(irq, IRQ_TYPE_EDGE_RISING);
		break;
	case CXD9225GG_IRQDETECT_EDGE_FALLING:
		cxd9225gg_set_irq_type(irq, IRQ_TYPE_EDGE_FALLING);
		break;
	case CXD9225GG_IRQDETECT_EDGE_BOTH:
		cxd9225gg_set_irq_type(irq, IRQ_TYPE_EDGE_BOTH);
		break;
	case CXD9225GG_IRQDETECT_LEVEL_HIGH:
		cxd9225gg_set_irq_type(irq, IRQ_TYPE_LEVEL_HIGH);
		break;
	case CXD9225GG_IRQDETECT_LEVEL_LOW:
		cxd9225gg_set_irq_type(irq, IRQ_TYPE_LEVEL_LOW);
		break;
	};
}

void __init cxd9225gg_init_irq(void)
{
	unsigned int i;

	/* disable user mode */
	writel(0, CXD9225GG_INTC_UEN_VIRT);

	for (i = 0; i < 4; i++) {
		/* clear edge interrupts */
		writel(~0, CXD9225GG_INTC_PIC_VIRT(i));
		/* interrupts treated as IRQ */
		writel(0, CXD9225GG_INTC_ISL_VIRT(i));
		/* disable interrupts */
		writel(~0, CXD9225GG_INTC_IEC_VIRT(i));
		/* level detection */
		writel(0, CXD9225GG_INTC_PLS_VIRT(i));
	}

	for (i = 0; i < 8; i++) {
		/* no detection (level) */
		writel(0, CXD9225GG_INTC_LVLC_VIRT(i));
		/* no detection (edge) */
		writel(0, CXD9225GG_INTC_EDGC_VIRT(i));
	}
	/* even priorities and clear VAD */
	for (i = 0; i < 128; i++) {
		writel(0, CXD9225GG_INTC_PRLX_VIRT(i));
		writel(0, CXD9225GG_INTC_VADX_VIRT(i));
	}

	/* clear prioriy mask */
	writel(0x0000ffff, CXD9225GG_INTC_PRLC_VIRT);

	for (i = 0; i < NR_IRQS; i++) {
		if (!cxd9225gg_irq_valid(i))
			continue;

		/* set the vector same as irq number + 0x0001000 */
		writel(i + 0x00010000, CXD9225GG_INTC_VADX_VIRT(i));

		/* set interrupt sense mode */
		cxd9225gg_set_irq_detection(i);

		if (cxd9225gg_irq_level(i)) {
			set_irq_chip(i, &cxd9225gg_irq_chip_level);
			set_irq_handler(i, handle_fasteoi_irq);
		} else {
			set_irq_chip(i, &cxd9225gg_irq_chip_edge);
			set_irq_handler(i, handle_fasteoi_irq);
		}
		set_irq_flags(i, IRQF_VALID);
	}
}

struct vic_regs {
	u32 ISL[4];
	u32 IEN[4];
	u32 SWI[4];
	u32 PLS[4];
	u32 PIC[4];

	u32 EDGC[8];
	u32 LVLC[8];

	u32 VADX[128];
	u32 PRLX[128];
};

static struct vic_regs vic;

extern void cxd9225gg_suspend_irq(void)
{
	unsigned int i;

	for (i = 0; i < 4; i++) {
		vic.ISL[i] = readl(CXD9225GG_INTC_ISL_VIRT(i));
		vic.IEN[i] = readl(CXD9225GG_INTC_IEN_VIRT(i));
		vic.SWI[i] = readl(CXD9225GG_INTC_SWI_VIRT(i));
		vic.PLS[i] = readl(CXD9225GG_INTC_PLS_VIRT(i));
		vic.PIC[i] = readl(CXD9225GG_INTC_PIC_VIRT(i));
	}

	for (i = 0; i < 8; i++) {
		vic.EDGC[i] = readl(CXD9225GG_INTC_EDGC_VIRT(i));
		vic.LVLC[i] = readl(CXD9225GG_INTC_LVLC_VIRT(i));
	}

	for (i = 0; i < 128; i++) {
		vic.VADX[i] = readl(CXD9225GG_INTC_VADX_VIRT(i));
		vic.PRLX[i] = readl(CXD9225GG_INTC_PRLX_VIRT(i));
	}

	/* disable interrupts */
	for (i = 0; i < 4; i++)
		writel(~0, CXD9225GG_INTC_IEC_VIRT(i));
}

extern void cxd9225gg_resume_irq(void)
{
	unsigned int i;

	/* disable interrupts */
	for (i = 0; i < 4; i++)
		writel(~0, CXD9225GG_INTC_IEC_VIRT(i));

	for (i = 0; i < 4; i++) {
		writel(vic.ISL[i], CXD9225GG_INTC_ISL_VIRT(i));
		writel(vic.SWI[i], CXD9225GG_INTC_SWI_VIRT(i));
		writel(vic.PLS[i], CXD9225GG_INTC_PLS_VIRT(i));
		writel(vic.PIC[i], CXD9225GG_INTC_PIC_VIRT(i));
	}

	for (i = 0; i < 8; i++) {
		writel(vic.EDGC[i], CXD9225GG_INTC_EDGC_VIRT(i));
		writel(vic.LVLC[i], CXD9225GG_INTC_LVLC_VIRT(i));
	}

	for (i = 0; i < 128; i++) {
		writel(vic.VADX[i], CXD9225GG_INTC_VADX_VIRT(i));
		writel(vic.PRLX[i], CXD9225GG_INTC_PRLX_VIRT(i));
	}

	/* clear prioriy mask */
	writel(0x0000ffff, CXD9225GG_INTC_PRLC_VIRT);

	/* enable interrupts */
	for (i = 0; i < 4; i++)
		writel(vic.IEN[i], CXD9225GG_INTC_IEN_VIRT(i));
}
