/* 2014-08-28: File changed by Sony Corporation */
/*
 * Copyright (C) 2014 Sony Corporation
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>
#include <mach/gpio.h>

#define EPR(fmt, args...) printk(KERN_ERR "[PCIEDEV]: " fmt, ##args)

struct cxd9225gg_gpio {
	cxd9225gg_gpio_pin_name_t pin;
	s32 active; /* 0: LACT, 1: HACT, -1: invalid */
};

#define PCIE_DEV_NAME_LEN 16
#define PCIE_DEV_GPIO_NUM 2
struct cxd9225gg_pcie_device {
	char name[PCIE_DEV_NAME_LEN];
	struct cxd9225gg_gpio gpio[PCIE_DEV_GPIO_NUM];
};

static struct cxd9225gg_pcie_device pciex4_devices[] = {
	{ .name = "ssd",  .gpio = {{ .pin = CXD9225GG_BGPIO_15, .active = 0 }, { .active = -1 }}},
	{ .name = "bne",  .gpio = {{ .pin = CXD9225GG_BGPIO_23, .active = 1 }, { .pin = CXD9225GG_GPINT_09, .active = 0 }}},
	{ .name = "sstp", .gpio = {{ .pin = CXD9225GG_BGPIO_14, .active = 1 }, { .pin = CXD9225GG_GPINT_15, .active = 0 }}},
	{ .name = "usb3", .gpio = {{ .pin = CXD9225GG_GPINT_14, .active = 0 }, { .active = -1 }}},
};

static void __cxd9225gg_pcie_set_device_power(struct cxd9225gg_pcie_device *dev, int on)
{
	int i;
	struct cxd9225gg_gpio *gpio;

	for (i = 0; i < ARRAY_SIZE(dev->gpio); i++) {
		gpio = &dev->gpio[i];
		if (gpio->active < 0)
			continue;

		cxd9225gg_set_gpio_dataout(gpio->pin, on ? gpio->active : !gpio->active);
	}
}

static int cxd9225gg_pcie_get_device_power(struct cxd9225gg_pcie_device *dev)
{
	int i;
	struct cxd9225gg_gpio *gpio;
	int ret = 0;

	for (i = 0; i < ARRAY_SIZE(dev->gpio); i++) {
		gpio = &dev->gpio[i];
		if (gpio->active < 0)
			continue;

		if (gpio->active == cxd9225gg_get_gpio_datain(gpio->pin)) {
			ret |= 1;
		} else {
			ret |= 2;
		}
	}

	if (ret == 1) /* on */
		return 1;

	if (ret == 2) /* off */
		return 0;

	return -1; /* invalid */
}

/* check if some devieces are already power on except for dev */
static int cxd9225gg_pcie_device_is_power_on(struct cxd9225gg_pcie_device *dev)
{
	int i;
	struct cxd9225gg_pcie_device *p;

	for (i = 0; i < ARRAY_SIZE(pciex4_devices); i++) {
		p = &pciex4_devices[i];
		if (cxd9225gg_pcie_get_device_power(p) == 1 && p != dev) {
			EPR("still power on: %s\n", p->name);
			return 1; /* another device is still power on */
		}
	}

	return 0;
}

static int cxd9225gg_pcie_set_device_power(struct cxd9225gg_pcie_device *dev, int on)
{
	if (on && cxd9225gg_pcie_device_is_power_on(dev))
		return -1;

	__cxd9225gg_pcie_set_device_power(dev, on);

	return 0;
}

static struct cxd9225gg_pcie_device *cxd9225gg_pcie_find_device(const char *name)
{
	int i;
	struct cxd9225gg_pcie_device *dev;

	for (i = 0; i < ARRAY_SIZE(pciex4_devices); i++) {
		dev = &pciex4_devices[i];
		if (!strncmp(dev->name, name, PCIE_DEV_NAME_LEN - 1))
			return dev;
	}

	return 0;
}

static int pcie_proc_read_devices(char *page, char **start, off_t off,
			 int count, int *eof, void *data)
{
	int i, stat;
	int len = 0;
	struct cxd9225gg_pcie_device *dev;

	for (i = 0; i < ARRAY_SIZE(pciex4_devices); i++) {
		dev = &pciex4_devices[i];
		stat = cxd9225gg_pcie_get_device_power(dev);
		len += snprintf(page + len, count - len, "%s: %s\n", dev->name, (stat < 0) ? "invalid" : (stat ? "on" : "off"));
	}

	return len;
}

static int pcie_proc_write_devices(struct file *file, const char __user *user_buf,
			  unsigned long count, void *data)
{
	char buf[PCIE_DEV_NAME_LEN + 5];
	char *str = buf;
	char delim[] = " ";
	char *name, *power;
	struct cxd9225gg_pcie_device *dev;
	int on;

	if (count >= sizeof(buf))
		return -EINVAL;

	if (copy_from_user(str, user_buf, count))
		return -EFAULT;

	buf[count] = '\0';
	if (buf[count - 1] == '\n') {
		buf[count - 1] = '\0';
	}

	name = strsep(&str, delim);
	if (!name) {
		EPR("invalid param: %s\n", buf);
		return -EINVAL;
	}

	dev = cxd9225gg_pcie_find_device(name);
	if (!dev) {
		EPR("not found device: %s\n", name);
		return -EINVAL;
	}

	power = strsep(&str, delim);
	if (!power) {
		EPR("invalid param: %s\n", buf);
		return -EINVAL;
	}

	if (!strncmp(power, "on", 2))
		on = 1;
	else if (!strncmp(power, "off", 3))
		on = 0;
	else {
		EPR("invalid param: %s\n", buf);
		return -EINVAL;
	}

	if (cxd9225gg_pcie_set_device_power(dev, on))
		return -EIO;

	return count;
}

#define PCIE_DEV_NODE "pcie_devices"

int __init pcie_proc_init_devices(void)
{
	struct proc_dir_entry *ent = create_proc_entry(PCIE_DEV_NODE, 0, NULL);

	if (!ent) {
		EPR("failed create: %s\n", PCIE_DEV_NODE);
		return -ENOMEM;
	}

	ent->read_proc  = pcie_proc_read_devices;
	ent->write_proc = pcie_proc_write_devices;

	return 0;
}

void __exit pcie_proc_exit_devices(void)
{
	remove_proc_entry(PCIE_DEV_NODE, NULL);
}

module_init(pcie_proc_init_devices);
module_exit(pcie_proc_exit_devices);
