/* 2014-08-28: File changed by Sony Corporation */
/*
 * Copyright (C) 2012 Sony Corporation
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/syscalls.h>

#include <mach/gpio.h>

#define CONFIG_PCIDBG_DEBUG

//==============================================================================
static __used char * __f_ = NULL;
static __used inline char * __basename_(char * str) {
	char *p;
	for(p=str; *p && ((*p=='/')? (str=p+1, 1): 1); p++);
	return str;
}
#define PROMPT  "[DBG] "
#define cmn_printk(level, format, arg...) \
	printk(level PROMPT "%s %s() @%d :> " format, \
		(__f_? __f_: (__f_=__basename_(__FILE__))), \
		__FUNCTION__, __LINE__, ## arg)

#ifdef  CONFIG_PCIDBG_DEBUG

#define dbg_debugl(format, arg...) \
	cmn_printk(KERN_DEBUG, format "\n" , ## arg)/*KERN_INFO*/

#define dbg_infol(format, arg...) \
	cmn_printk(KERN_NOTICE, format "\n" , ## arg)/*KERN_NOTICE*/

#define dbg_warningl(format, arg...) \
	cmn_printk(KERN_WARNING, format "\n" , ## arg)

#define dbg_errorl(format, arg...) \
	cmn_printk(KERN_ERR, "Error! " format "\n" , ## arg)

#else /* CONFIG_PCIDBG_DEBUG */

#define dbg_debugl(format, arg...)

#define dbg_infol(format, arg...)

#define dbg_warningl(format, arg...)

#define dbg_errorl(format, arg...) \
	cmn_printk(KERN_ERR, "Error! " format "\n" , ## arg)

#endif /* CONFIG_PCIDBG_DEBUG */
//==============================================================================
#define OP_R 'r'
#define OP_W 'w'

enum {
	DBG_OK = 0,
	DBG_ERR_FORMAT,
	DBG_ERR_FAIL,
	DBG_ERR_NOT_SUPPORT,
};

struct dbg_cmd_entry_t
{
	char *name;
	int (*pfn_do_cmd)(const char* cmdbuf);
	ssize_t (*pfn_output)(char __user *buf, loff_t *ppos);
	char *help;
	char *help_detail;
};
static struct dbg_cmd_entry_t cmd_list[];
static int cmd_list_size;

/*
 * record output for what command.
 * some commands is supposed to output large quality of data to the stdout,
 * (perhaps the upper user needs to save the data as a log file)
 * so we keep a buffer for doing this.
 * when user use 'cat /proc/cxd9225gg_debug' command to get data from us, we
 * just hand out this buffer.
 *
 * output_cmd : is the index of the command in cmd_list array.
*/
static int   output_cmd  = -1;

/*------------------------------------------------------------------------------
 * show resources of a pci_dev.
*/
static int dbg_pci_show_res(struct device *dev, void *data)
{
	struct pci_dev *pdev = to_pci_dev(dev);
	int i;
	printk(KERN_CRIT"%s\n", (strchr(pci_name(pdev),':')+1));
	printk(KERN_CRIT"\tVendorID(0x%04x)    DeviceID(0x%04x)    Class(0x%08x)\n",
			pdev->vendor, pdev->device, pdev->class);
	for (i=0; i<6; i++) {
		if (pci_resource_flags(pdev, i)) {
			printk(KERN_CRIT"\tRES[%d]: addr=0x%08x  "
				"len=0x%08x  flags=0x%08x\n", i,
				(unsigned)pci_resource_start(pdev, i),
				(unsigned)pci_resource_len(pdev, i),
				(unsigned)pci_resource_flags(pdev, i));
		}
	}
	/* always return 0 so that we can traversal all devices. */
	return 0;
}

/*
 * format: help
 * show help informations.
*/
static int dbg_do_help(const char * cmdbuf)
{
	int i;
	printk("***********************************************************\n");
	printk("* Note, this is the usage of /proc/cxd9225gg_debug node.     *\n");
	printk("* If you have a helper script, then see the script first. *\n");
	printk("***********************************************************\n\n");
	printk("Supported commands:\n");
	printk("-------------------\n");
	for (i=0; i<cmd_list_size; i++) {
		printk("%s :\n", cmd_list[i].name);
		if (cmd_list[i].help)
			printk("\t\t%s\n", cmd_list[i].help);
		if (cmd_list[i].help_detail)
			printk("\t\t%s\n", cmd_list[i].help_detail);
		printk("\n");
	}
	return 0;
}

/*
 * format: oc
 * output check, show what command is the next output for.
*/
static int dbg_do_oc(const char * cmdbuf)
{
	if (output_cmd >= 0) {
		printk("Next output is for command '%s'.\n", cmd_list[output_cmd].name);
	} else {
		printk("No data is to be output.\n");
	}
	return 0;
}

/*
 * format: mr <phyaddr:hex> <size:hex> <option>
 *         mw <phyaddr:hex> <value:hex>
 * in mr case, <size> means how many bytes to read and print.
 *             <option> can be '-q', means quiet, no printk.
 * in mw case, <value> means the value to write.
 *             Note: <value> has to be aligned against 4 bytes when write.
*/
static char* mem_buf = NULL;
static int   mem_buf_sz = 0;
static char _fpath[1024] = "";
static int dbg_do_mem_rw(const char * cmdbuf)
{
	int ret = 0;
	unsigned int phyaddr = 0;
	unsigned int value = 0;
	char op = 0;
	void * pmap = NULL;
	unsigned int nbytes = 0;
	char param1[32] = "";
	char *param2 = _fpath;
	int quiet = 0;
	int fdump = 0;

	/* parse the command parameters. */
	ret = sscanf(cmdbuf, "m%c 0x%x 0x%x %s %s", &op, &phyaddr, &value, param1, param2);
	if (!((op==OP_R && (ret==5 || ret==4 || ret==3 || ret==2))
			|| (op==OP_W && ret==3))) {
		dbg_errorl("mr command format is wrong!");
		return -DBG_ERR_FORMAT;
	}
	phyaddr &= ~0x3; //align against 32-bit.

	if (op==OP_R && ret >= 4) {
		if (!strcmp(param1, "-q")) { quiet = 1; }
		else
		if (!strcmp(param1, "-f")) {
			if (0!=strcmp(param2, "")) {
				fdump = 1;
			} else {
				dbg_errorl("-f is set but no filepath is given.");
				return -DBG_ERR_FORMAT;
			}
		}
	}

	/* default byte num for READ is 4 bytes */
	if (op==OP_R) {
		nbytes = (value==0)? sizeof(int): value;
	} else {
		/* WRITE does not use nbytes variable */
	}

	/* memory remap */
	pmap = (void*)ioremap_nocache(phyaddr, nbytes);
	if (!pmap) {
		dbg_errorl("ioremap_nocache(0x%08x, 0x%x) failed!", phyaddr, nbytes);
		return -DBG_ERR_FAIL;
	}

	/* do read/write */
	if (op == OP_R) {
		if (fdump) {
			//file dump, param2 is filepath
			char *fname = param2;
			int fd;
			unsigned int offset = 0;
			mm_segment_t old_fs;
			
			old_fs = get_fs();
			set_fs(KERNEL_DS);
			
			fd = sys_open(fname, O_WRONLY|O_CREAT, 0644);
			if (fd < 0) {
				printk("open %s for O_WRONLY, failed!\n", fname);
				return -DBG_ERR_FAIL;
			}

			while((nbytes-offset)>0) {
				int sz = min((u32)SZ_8M, (u32)(nbytes-offset));
				if (sz != sys_write(fd, pmap+offset, sz)) {
					dbg_errorl("sys_write() failed!");
					sys_close(fd);
					goto _getout;
				}
				offset += sz;
				dbg_infol("%d (0x%x) bytes are written.", offset, offset);
			}
			sys_close(fd);
			
			set_fs(old_fs);
		} else {
			//buffer dump
			/* prepare output buffer */
			mem_buf = kmalloc(nbytes, GFP_KERNEL);
			mem_buf_sz = nbytes;
			if (!mem_buf) {
				dbg_errorl("kmalloc(%d) failed!", nbytes);
				mem_buf_sz = 0;
				goto _getout;
			}
			
			/* do output */
			memcpy(mem_buf, pmap, mem_buf_sz);
			/* print the result as we want */
			if (!quiet) {
				unsigned int *p = pmap;
				int i;
				for (i=0; i<(nbytes/4); i++, p++) {
					if (!(i%4)) {
						if (i) { printk("\n"); }
						printk("0x%08x:  ", (u32)((u32)p-(u32)pmap+(u32)phyaddr));
					}
					printk("0x%08x ", *p);
				}
				printk("\n");
			}
		}
	} else {
		*((unsigned int*)pmap) = value;
		printk("mem[0x%08x] = 0x%08x\n", phyaddr, *((unsigned int*)pmap));
	}

_getout:
	// unmap the region.
	if (pmap) { iounmap(pmap); }

	return 0;
}

static ssize_t dbg_do_mem_out(char __user *buf, loff_t *ppos)
{
	int dummy; //shutup the warning
	int bytes_left;

	if (!mem_buf || !mem_buf_sz) { return 0; }

	bytes_left = mem_buf_sz - (int)*ppos;
	if (bytes_left >= SZ_4K) {
		/* directly output a 4K page */
		dummy = copy_to_user(buf, mem_buf+*ppos, SZ_4K);
		*ppos += SZ_4K;
		return SZ_4K;
	} else
	if (bytes_left > 0) {
		/* output the left bytes */
		dummy = copy_to_user(buf, mem_buf+*ppos, bytes_left);
		*ppos += bytes_left;

		/* because last data is copied to user, we can just
		   free the buffer now. */
		kfree(mem_buf);
		mem_buf = NULL;
		mem_buf_sz = 0;

		return bytes_left;
	}
	/* at last, nothing to be read from us */
	return 0;
}

/*
 * read file from kernel space, then print the content as string.
 * just for test.
 * format: kread <filename>
*/
static int dbg_do_kread(const char * cmdbuf)
{
	int ret = 0;
	char fname[256];
	char *buf = NULL;
	int fd;
	int n;
	mm_segment_t old_fs;

	ret = sscanf(cmdbuf, "kread %s", fname);
	if (ret != 1) {return -DBG_ERR_FORMAT;}

	old_fs = get_fs();
	set_fs(KERNEL_DS);

	fd = sys_open(fname, O_RDONLY, 0);
	if (fd >= 0) {
		buf = (void*)__get_free_page(GFP_KERNEL);
		if(!buf)
		{
			return -ENOMEM;
		}
		while ((n=sys_read(fd, buf, SZ_4K-1))>0) {
			buf[n] = '\0';
			printk("%s", buf);
		}
		printk("\n");
		sys_close(fd);
		free_page((unsigned long)buf);
	} else {
		printk("open %s for O_RDONLY, failed!\n", fname);
		return -DBG_ERR_FAIL;
	}

	set_fs(old_fs);
	return 0;
}

/*
 * write file from kernel space. write the absolute path into the file.
 * just for test.
 * format: kwrite <filename>
*/
static int dbg_do_kwrite(const char * cmdbuf)
{
	int ret = 0;
	char fname[256];
	int fd;
	mm_segment_t old_fs;

	ret = sscanf(cmdbuf, "kwrite %s", fname);
	if (ret != 1) {return -DBG_ERR_FORMAT;}

	old_fs = get_fs();
	set_fs(KERNEL_DS);

	fd = sys_open(fname, O_WRONLY|O_CREAT, 0644);
	if (fd >= 0) {
		sys_write(fd, fname, strlen(fname));
		sys_close(fd);
	} else {
		printk("open %s for O_WRONLY, failed!\n", fname);
		return -DBG_ERR_FAIL;
	}

	set_fs(old_fs);
	return 0;
}

static int dbg_do_cpfileto(const char * cmdbuf)
{
	int ret = 0;
	char fname[256];
	u32 physaddr = 0;
	char *buf = NULL;
	int fd;
	mm_segment_t old_fs;
	int failed = 0;

	ret = sscanf(cmdbuf, "cpfileto %s 0x%08x", fname, &physaddr);
	if (ret != 2) {return -DBG_ERR_FORMAT;}


	old_fs = get_fs();
	set_fs(KERNEL_DS);

	fd = sys_open(fname, O_RDONLY, 0);
	if (fd >= 0) {
		long long cpsize = 0;
		int n = 0;
		while (1) {
			buf = (void *)ioremap(physaddr, SZ_1M);
			if (!buf) {
				dbg_errorl("ioremap(0x%08x, 0x%x) failed!", physaddr, SZ_1M);
				failed = 1;
				break;
			}

			n=sys_read(fd, buf, SZ_1M);
			if (n<0) {
				//error
				dbg_errorl("sys_read() failed, ret=%d", n);
				failed = 1;
				break;
			}
			if (n==0) {
				//reach the end of the file.
				break;
			}

			iounmap(buf); buf=NULL;

			cpsize += n;
			physaddr += n;

			if (cpsize && !(cpsize%SZ_8M)) {
				printk("%lld MB has been copy to memory.\n", cpsize>>20);
			}
		}
		
		if (buf) iounmap(buf);
		sys_close(fd);

		printk("%llu bytes has been written to memory.\n", cpsize);
	} else {
		printk("open %s for O_RDONLY, failed!\n", fname);
		failed = 1;
	}

	set_fs(old_fs);

	if (failed) {return -DBG_ERR_FAIL;}


	return 0;
}

/*
 * format: swi <IRQ>
 * generate soft interrupt of given IRQ number.
 * some useful IRQ: (see irqs.h)
 * 22 - PCIEx1L0 INTA
 * 31 - PCIEx1L1 INTA
 * 40 - PCIEx4L  INTA
*/
static int dbg_do_swi(const char * cmdbuf)
{
	int ret;
	int irq;
	extern int cxd9225gg_generate_irq(unsigned int irq);
	ret = sscanf(cmdbuf, "swi %d", &irq);
	if (!(ret == 1)) { return -DBG_ERR_FORMAT; }
	ret = cxd9225gg_generate_irq(irq);
	if (ret) { return -DBG_ERR_FAIL; }
	return 0;
}

/*
 * format: gpio <bgpio> [in|out|<value>]
 * 1) gpio <bgpio>
 *	read value of bgpio value.
 * 2) gpio <bgpio> <value>
 *	write value of bgpio.
 *	bgpio direction will be changed to Output Mode first.
 *	value is 0(low) or 1(high).
 * 3) gpio <bgpio> in|out
 *	configure direction of bgpio.
*/
static int dbg_do_gpio_set(const char * cmdbuf)
{
	int ret;
	int bgpio, value;
	int gpio_id;
	char dir[8] = {0};
	ret = sscanf(cmdbuf, "gpio %d %d", &bgpio, &value);
	if (ret < 1) {return -DBG_ERR_FORMAT;}

	gpio_id = CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_10+(bgpio/8), (bgpio%8));
	if (ret == 2) {
		//set direction as output
		cxd9225gg_set_gpio_direction(gpio_id, 0);
		//write
		cxd9225gg_set_gpio_dataout(gpio_id, value!=0);
		//read back
		value = cxd9225gg_get_gpio_datain(gpio_id);
		printk("BGPIO[%d]=%d\n", bgpio, value);
	} else
	if (ret == 1) {
		ret = sscanf(cmdbuf, "gpio %d %s", &bgpio, dir);
		if (ret == 2) {
			// config direction
			if (!strncmp(dir, "in", strlen("in"))) {
				cxd9225gg_set_gpio_direction(gpio_id, 1);
			} else
			if (!strncmp(dir, "out", strlen("out"))) {
				cxd9225gg_set_gpio_direction(gpio_id, 0);
			} else {
				return -DBG_ERR_FORMAT;
			}
		} else {
			// read
			value = cxd9225gg_get_gpio_datain(gpio_id);
			printk("BGPIO[%d]=%d\n", bgpio, value);
		}
	}
	return 0;
}

/*
 * format: showpci
 * print general informations of all pci/pcie devices.
*/
static int dbg_do_pci_show(const char * cmdbuff)
{
	extern struct bus_type pci_bus_type;
	(void)bus_find_device(&pci_bus_type, NULL, NULL, dbg_pci_show_res);
	return 0;
}

/*
 * format: cr <bus_id> <reg_offset:hex>
 *         cw <bus_id> <reg_offset:hex> <value:hex>
 * read or write config space registers of given pci device.
*/
static int dbg_do_pci_cfg_rw(const char * cmdbuf)
{
	char op=0;
	int ret=0, bus=0, dev=0, func=0, off=0, val=0;
	struct pci_bus *pbus;

	ret = sscanf(cmdbuf, "c%c %x:%x.%x 0x%x 0x%x", &op, &bus, &dev, &func, &off, &val);
	if (!((op==OP_W && ret==6) || (op==OP_R && ret==5))) {
		return -DBG_ERR_FORMAT;
	}
	off &= ~0x3;

	/* need to check parameters */

	pbus = pci_find_bus(0, bus);
	if (pbus == NULL) {
		// error: no such pci_bus.
		return -DBG_ERR_FAIL;
	}

	ret = 0;
	if (op == OP_W) {
		ret = pci_bus_write_config_dword(pbus, (dev<<3|func), off, val);
	}
	if (!ret) {
		ret = pci_bus_read_config_dword(pbus, (dev<<3|func), off, &val);
		printk("%02x:%02x.%x: REG[0x%x]=0x%08x\n", bus, dev, func, off, val);
	}

	if (ret) {
		/*
		 * Error, pci/pcie hc driver can not read/write config
		 * space of the given device.
		*/
		printk("%02x:%02x.%x: %s config space REG[0x%x] failed! ret=%d",
			bus, dev, func, (op==OP_R)?"read":"write", off, ret);
		return -DBG_ERR_FAIL;
	}

	return 0;
}

/*
 * format: enable <bus_id>
 * enable given pci device.
*/
static int dbg_do_pci_enable(const char * cmdbuf)
{
	int ret, bus, dev, func;
	struct pci_dev *pdev;

	ret = sscanf(cmdbuf, "enable %x:%x.%x", &bus, &dev, &func);
	if (ret != 3) { return -DBG_ERR_FORMAT; }

	pdev = pci_get_bus_and_slot(bus, (dev<<3|func));
	if (!pdev) {
		printk("No device at %x:%x.%x\n", bus, dev, func);
		return -DBG_ERR_FAIL;
	}
	if (pci_enable_device(pdev)) { return -DBG_ERR_FAIL; }

	return 0;
}

/*
 * format: disable <bus_id>
 * disable given pci device.
*/
static int dbg_do_pci_disable(const char * cmdbuf)
{
	int ret, bus, dev, func;
	struct pci_dev *pdev;

	ret = sscanf(cmdbuf, "disable %x:%x.%x", &bus, &dev, &func);
	if (ret != 3) { return -DBG_ERR_FORMAT; }

	pdev = pci_get_bus_and_slot(bus, (dev<<3|func));
	if (pdev) {
		u16 old_cmd=0;
		u16 new_cmd=0;
		pci_disable_device(pdev);
		/*
		 * but, pci_disable_device() does not really disable the
		 * device, so we do the real job here now.
		*/
		pci_read_config_word(pdev, PCI_COMMAND, &old_cmd);
		new_cmd = old_cmd & ~(PCI_COMMAND_MEMORY|PCI_COMMAND_IO);
		if (new_cmd != old_cmd) {
			pci_write_config_word(pdev, PCI_COMMAND, new_cmd);
			printk("PCI: disabling device %s (%04x -> %04x)\n",
			       pci_name(pdev), old_cmd, new_cmd);
		}

	}

	return 0;
}


/*------------------------------------------------------------------------------
 * This is the list of all commands which this module supports.
 * Feel free to add any useful of funny command here.
 * All the command line parameters which are sent to the /proc node
 * will be pass to the handler function, including the 'cmd' itself.
*/
static struct dbg_cmd_entry_t cmd_list[] = {
	/* general commands */
	{"help", 		dbg_do_help,
	 0,			"Show this help.", NULL},
	{"oc",  		dbg_do_oc,
	 0,			"Output command check.", NULL},
	{"mr",  		dbg_do_mem_rw,
	dbg_do_mem_out,		"Memory read.",
	 			"format: mr <phyaddr> <n_32bit> [-q]"},
	{"mw",  		dbg_do_mem_rw,
	 0,			"Memory write.",
	 			"format: mw <phyaddr> <value>"},
	{"kread", 		dbg_do_kread,
	 0, 			"Test reading file from kernel space.", NULL},
	{"kwrite", 		dbg_do_kwrite,
	 0, 			"Test writing file from kernel space.", NULL},

	{"cpfileto", 		dbg_do_cpfileto,
	 0, 			"Copy file to physical address.",
	 			"format: cpfileto <filename> <physaddr>"},
	/* pci commands */
	{"showpci", 		dbg_do_pci_show,
	 0,			"Show brief of all PCI/PCIe devices, "
	 			"including bus_id.", NULL},
	{"cr",  		dbg_do_pci_cfg_rw,
	 0,			"Read PCI/PCIe device config space registers.",
	 			"format: cr <bus_id> <reg_offset>"},
	{"cw",  		dbg_do_pci_cfg_rw,
	 0,			"Write PCI/PCIe device config space registers.",
	 			"format: cw <bus_id> <reg_offset> <value>"},
	{"enable", 		dbg_do_pci_enable,
	 0,			"Enable PCI/PCIe device.",
	 			"format: enable <bus_id>"},
	{"disable", 		dbg_do_pci_disable,
	 0,			"Disable PCI/PCIe device.",
	 			"format: disable <bus_id>"},
	/* cxd9225gg specific commands */
	{"swi", 		dbg_do_swi,
	 0,			"Generate a swi.",
	 			"format: swi <IRQ>"},
	{"gpio", 		dbg_do_gpio_set,
	 0,			"Set, get or configure the given BGPIO.",
	 			"format: gpio <bgpio> [in|out|<value>]"},

};
static int cmd_list_size = ARRAY_SIZE(cmd_list);

static int dbg_parse_and_do_cmd(const char * cmdbuf)
{
	int ret;
	char cmd[32] = {0};
	int i;

	/* scan command name */
	ret = sscanf(cmdbuf, "%s ", cmd);
	if (ret != 1 || !strlen(cmd)) {
		/* match command failed */
		dbg_errorl("no command input!");
		return -1;
	} else
	if (32 < strlen(cmd)+1) {
		/* already overflow */
		dbg_errorl("command must be within 31 byts!");
		return -1;
	}

	/* find out the command entry, and do the command */
	for (i=0; i<cmd_list_size; i++) {
		if (!strcmp(cmd_list[i].name, cmd)) {
			/*
			 * If the command declare itself an outputing command,
			 * we make sure the mem_buf and mem_buf_sz are
			 * re-initialized before we call its handler func.
			*/
			if (cmd_list[i].pfn_output) {
				output_cmd=i;
			}
			if (cmd_list[i].pfn_do_cmd) {
				ret = cmd_list[i].pfn_do_cmd(cmdbuf);
			} else {
				dbg_warningl("No handler for Command %s.", cmd);
				ret = -1;
			}
			break;
		}
	}

	/* report failure if necessary */
	if (i < cmd_list_size) {
		if (ret == -DBG_ERR_FORMAT) {
			dbg_errorl("Invalid parameters!");
		} else
		if (ret) {
			dbg_errorl("Command %s failed!", cmd);
		}
	} else {
		/* no such command */
		ret = -1;
		dbg_errorl("Command %s is not supported!", cmd);
	}

	return ret;
}

/*
 * return nbytes to tell the caller there may be more data.
 * return 0 to tell the caller that it reaches the end.
 * return -1 to tell caller that we do not support this operation.
*/
static ssize_t cxd9225gg_procfs_dbg_read(struct file *file, char __user *buf,
					size_t nbytes, loff_t *ppos)
{
	if (nbytes < SZ_4K) {
		dbg_errorl("nbytes=0x%x, while page_size=0x%x", nbytes, SZ_4K);
		return -1;
	}

	if (output_cmd>=0 && cmd_list[output_cmd].pfn_output) {
		nbytes = cmd_list[output_cmd].pfn_output(buf, ppos);
		/* output is done, reset output_cmd. */
		if (!nbytes) {output_cmd = -1;}
		return nbytes;
	} else {
		printk("No data is to be output.\n");
	}

	return 0;
}

/*
 * return nbytes to tell caller that all data is written.
 * return -1 to tell caller that we do not support this operation.
*/
static ssize_t cxd9225gg_procfs_dbg_write(struct file *file, const char __user *buf_usr,
					size_t nbytes, loff_t *ppos)
{
	int ret = 0;
	char * bufpage = NULL;

	/* let tty preempt print first */
	msleep(1);

	/* allocate a page buffer and save the user data */
	bufpage = (char*)__get_free_page(GFP_KERNEL);
	if (!bufpage) {
		printk("__get_free_page() failed!!\n");
		return -1;
	}
	if ((ret = copy_from_user(bufpage, buf_usr, nbytes)) < 0) {
		printk("copy_from_user() failed!!\n");
		return -1;
	}
	bufpage[nbytes] = 0; //truncate the string.

	/* parse and do command */
	dbg_parse_and_do_cmd(bufpage);

	/* release the memory that allocated */
	if (bufpage) { free_page((unsigned long)bufpage); }

	/* tell the fs that the 'write' operation is done */
	return nbytes;
}

static const struct file_operations procfs_pci_dbg_ops = {
	.read	= cxd9225gg_procfs_dbg_read,
	.write	= cxd9225gg_procfs_dbg_write,
};

#define PROC_CXD9225GG_DEBUG "cxd9225gg_debug"

static int __init cxd9225gg_procfs_dbg_init(void)
{
	struct proc_dir_entry *e;

	/* give a NULL as parent, to create entry in /proc */
	e = proc_create(PROC_CXD9225GG_DEBUG, S_IFREG|S_IRUGO|S_IWUSR, NULL,
			&procfs_pci_dbg_ops);
	
	return 0;
}

static void __exit cxd9225gg_procfs_dbg_exit(void)
{
	remove_proc_entry(PROC_CXD9225GG_DEBUG, NULL);
}

late_initcall(cxd9225gg_procfs_dbg_init);
module_exit(cxd9225gg_procfs_dbg_exit);
