/* 2014-08-28: File changed by Sony Corporation */
/*
 * Copyright (C) 2014 Sony Corporation
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/device.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/io.h>
#include <linux/spi/spi.h>
#include <linux/snsc_boot_time.h>
#include <mach/plat_flash.h>

#define REG_SFMSMD 0x00
# define SFMSMD        0x72 /* PFE || SE forever || First Read Dual Output */
#define REG_SFMSSC 0x04
# define SFMSSC        0x37 /* SLD || SHD || SW 8*SCK */
#define REG_SFMSKC 0x08
# define SFMSKC        0x00 /* DV 2*CLK */
#define REG_SFMSST 0x0c
#define REG_SFMCOM 0x10
#define REG_SFMCMD 0x14
# define SFMCMD_ROM    0x0
# define SFMCMD_DIRECT 0x1
#define REG_SFMCST 0x18

#define WR(sfm, off, val) writeb(val, (sfm)->base + (off))
#define RD(sfm, off)      readb((sfm)->base + (off))
#define RD_W(sfm, off)    readw((sfm)->base + (off))

struct sfmemc {
	struct workqueue_struct *wq;
	struct work_struct	work;
	spinlock_t		lock;
	struct list_head	msg_queue;
	struct spi_master	*master;
	void __iomem		*base;
	unsigned int		memsize;
};

#if 0 /* for debug */
static void sfmemc_dump_reg(struct sfmemc *sfm)
{
	printk("SFMSMD: 0x%04x, SFMSSC: 0x%02x, SFMSKC: 0x%02x\n",
		RD_W(sfm, REG_SFMSMD),
		RD(sfm, REG_SFMSSC),
		RD(sfm, REG_SFMSKC)
	);
	printk("SFMSST: 0x%02x, SFMCOM: 0x%02x, SFMCMD: 0x%02x, SFMCST: 0x%02x\n",
		RD(sfm, REG_SFMSST),
		RD(sfm, REG_SFMCOM),
		RD(sfm, REG_SFMCMD),
		RD(sfm, REG_SFMCST)
	);
}
#endif

#if 0 /* bootloader has already initialized */
static inline void sfmemc_device_init(struct sfmemc *sfm)
{
	WR(sfm, REG_SFMSMD, SFMSMD);
	WR(sfm, REG_SFMSSC, SFMSSC);
	WR(sfm, REG_SFMSKC, SFMSKC);
}
#endif

static inline void sfmemc_cmd(struct sfmemc *sfm, u8 cmd)
{
	WR(sfm, REG_SFMCMD, cmd);
}

static inline u8 sfmemc_read(struct sfmemc *sfm)
{
	return RD(sfm, REG_SFMCOM);
}

static inline void sfmemc_write(struct sfmemc *sfm, u8 dat)
{
	WR(sfm, REG_SFMCOM, dat);
}

static int sfmemc_xfer_pio(struct sfmemc *sfm, struct spi_transfer *trans,
		unsigned *xfered)
{
	unsigned i;
	unsigned len = trans->len;
	const u8 *tx = trans->tx_buf;
	u8 *rx       = trans->rx_buf;

	if (!tx && !rx && len)
		return -EINVAL;

	for (i = 0; i < len; i++) {
		if (tx)
			sfmemc_write(sfm, *tx++);
		if (rx)
			*rx++ = sfmemc_read(sfm);
	}

	*xfered += len;

	return 0;
}

static int sfmemc_chip_select(struct spi_device *spi, u32 cs)
{
	platform_chip_select_fn func = spi->controller_data;

	return func(cs);
}

static int sfmemc_xfer(struct sfmemc *sfm, struct spi_device *spi,
		struct list_head *transfers, unsigned *xfered)
{
	struct spi_transfer *trans;
	int err;

	err = sfmemc_chip_select(spi, spi->chip_select);
	if (err)
		return err;

	sfmemc_cmd(sfm, SFMCMD_DIRECT);

	list_for_each_entry(trans, transfers, transfer_list) {
		err = sfmemc_xfer_pio(sfm, trans, xfered);
		if (err)
			break;

		if (trans->delay_usecs)
			udelay(trans->delay_usecs);
	}

	sfmemc_cmd(sfm, SFMCMD_ROM);

	return err;
}

static void sfmemc_work(struct work_struct *work)
{
	struct sfmemc *sfm = container_of(work, struct sfmemc, work);

	spin_lock_irq(&sfm->lock);

	while (!list_empty(&sfm->msg_queue)) {
		struct spi_message *msg;
		unsigned xfered = 0;

		msg = container_of(sfm->msg_queue.next, struct spi_message,
				queue);

		list_del_init(&msg->queue);
		spin_unlock_irq(&sfm->lock);

		msg->status = sfmemc_xfer(sfm, msg->spi, &msg->transfers,
				&xfered);
		msg->actual_length = xfered;
		msg->complete(msg->context);

		spin_lock_irq(&sfm->lock);
	}

	spin_unlock_irq(&sfm->lock);
}

static int sfmemc_transfer(struct spi_device *spi, struct spi_message *msg)
{
	struct sfmemc *sfm;
	unsigned long flags;

	msg->actual_length = 0;
	msg->status = 0;

	/* reject invalid messages and transfers */
	if (list_empty(&msg->transfers) || !msg->complete)
		return -EINVAL;

	sfm = spi_master_get_devdata(spi->master);

	spin_lock_irqsave(&sfm->lock, flags);
	list_add_tail(&msg->queue, &sfm->msg_queue);
	queue_work(sfm->wq, &sfm->work);
	spin_unlock_irqrestore(&sfm->lock, flags);

	return 0;
}

static int sfmemc_setup(struct spi_device *spi)
{
	return 0;
}

static void sfmemc_cleanup(struct spi_device *spi)
{
}

static int __init sfmemc_probe(struct platform_device *pdev)
{
	struct sfmemc *sfm;
	struct spi_master *master;
	struct resource *res;
	int ret = 0;

	BOOT_TIME_ADD1("sfmemc_probe: start");

	master = spi_alloc_master(&pdev->dev, sizeof(*sfm));
	if (master == NULL) {
		dev_err(&pdev->dev, "master allocation failed\n");
		return -ENOMEM;
	}

	master->setup    = sfmemc_setup;
	master->cleanup  = sfmemc_cleanup;
	master->transfer = sfmemc_transfer;
	master->bus_num  = SPI_BUS_NUM;
	master->num_chipselect = SPI_CHIP_NUM;

	dev_set_drvdata(&pdev->dev, master);

	sfm = spi_master_get_devdata(master);
	sfm->master = master;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		dev_err(&pdev->dev, "can't get I/O resource\n");
		ret = -ENXIO;
		goto master_put;
	}

	if (!request_mem_region(res->start, (res->end - res->start) + 1,
				pdev->dev.bus_id)) {
		dev_err(&pdev->dev, "can't get region I/O\n");
		ret = -EBUSY;
		goto master_put;
	}

	sfm->base = ioremap(res->start, res->end - res->start + 1);
	if (!sfm->base) {
		dev_err(&pdev->dev, "can't ioremap register I/O\n");
		ret = -ENOMEM;
		goto rel_region;
	}

	sfm->wq = create_singlethread_workqueue(pdev->name);
	if (!sfm->wq) {
		dev_err(&pdev->dev, "can't create workqueue\n");
		ret = -ENOMEM;
		goto reg_unmap;
	}

	spin_lock_init(&sfm->lock);
	INIT_LIST_HEAD(&sfm->msg_queue);
	INIT_WORK(&sfm->work, sfmemc_work);

	BOOT_TIME_ADD1("sfmemc_probe: spi_register");

	ret = spi_register_master(master);
	if (ret < 0) {
		dev_err(&pdev->dev, "can't register register master\n");
		goto destroy_wq;
	}

	BOOT_TIME_ADD1("sfmemc_probe: end");

	return ret;

destroy_wq:
	destroy_workqueue(sfm->wq);
reg_unmap:
	iounmap(sfm->base);
rel_region:
	release_mem_region(res->start, (res->end - res->start) + 1);
master_put:
	spi_master_put(master);

	return ret;
}

static int sfmemc_suspend(struct platform_device *pdev, pm_message_t state)
{
	/* nothing to do */
	return 0;
}

static int sfmemc_resume(struct platform_device *pdev)
{
	/* nothing to do */
	return 0;
}

#ifdef MODULE
static int __exit sfmemc_remove(struct platform_device *pdev)
{
	struct spi_master *master;
	struct sfmemc *sfm;
	struct resource *res;

	master = dev_get_drvdata(&pdev->dev);
	sfm = spi_master_get_devdata(master);

	spi_unregister_master(master);
	destroy_workqueue(sfm->wq);

	iounmap(sfm->base);
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res)
		release_mem_region(res->start, (res->end - res->start) + 1);

	spi_master_put(master);

	return 0;
}
#endif

static struct platform_driver sfmemc_driver = {
	.driver = {
		.name  = "sfmemc",
		.owner = THIS_MODULE,
	},
	.suspend = sfmemc_suspend,
	.resume  = sfmemc_resume,
	.remove  = __exit_p(sfmemc_remove),
};

static int __init sfmemc_init(void)
{
	return platform_driver_probe(&sfmemc_driver, sfmemc_probe);
}

#ifdef MODULE
module_init(sfmemc_init);
#else
rootfs_initcall(sfmemc_init);
#endif

#ifdef MODULE
static void __exit sfmemc_exit(void)
{
	platform_driver_unregister(&sfmemc_driver);
}
module_exit(sfmemc_exit);
#endif

MODULE_LICENSE("GPL");
