/*
 * Copyright (C) 2012 Sony Corporation 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License, version 2, as 
 * published by the Free Software Foundation. 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef __INCusbEvalLibh
#define __INCusbEvalLibh


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/mutex.h>
#include <linux/completion.h>
#include <linux/irq.h>

/* attention: for linux 2.6.23 */
#include <linux/semaphore.h>

#include "cxd9225gg_usb.h"


#define USB_CONF_INDEX_SIZE    16
#define USB_DESC_TYPE_CONFIG   0x0200
#define USB_DESC_TYPE_MASK     0x0700
#define USB_DESC_INDEX_MASK    0x000F
#define USB_1MILLI_SEC         1000000


typedef enum {
    USB_CONNECT_UNDEF          = 0,
    USB_DISCONNECTED_STATUS,
    USB_CONNECTED_STATUS
} USB_CONNECT_STATUS;

typedef enum {
    USB_TRANSFER_TYPE_PIO  = 0,
    USB_TRANSFER_TYPE_DMA,
    USB_TRANSFER_TYPE_NULL
} USB_TRANSFER_TYPE;


typedef enum {
    USB_DMA_OUT_PH_IDLE        = 0,
    USB_DMA_OUT_PH_STARTED,
    USB_DMA_OUT_PH_LESS3B_END,
    USB_DMA_OUT_PH_SHORTPACKET,
    USB_DMA_OUT_PH_MULTIPACKETS,
    USB_DMA_OUT_PH_EPDMA_END_SHORT,
    USB_DMA_OUT_PH_EPDMA_END_MULTI,
    USB_DMA_OUT_PH_ALL_END,
    USB_DMA_OUT_PH_WAIT_BRIDGEDMA_END,
    USB_DMA_OUT_PH_END_WITH_FRAGMENT
} USB_DMA_OUT_PHASE;

typedef enum {
    USB_DMA_IN_PH_IDLE         = 0,
    USB_DMA_IN_PH_BRIDGEDMA_END,
    USB_DMA_IN_PH_EPCDMA_END,
    USB_DMA_IN_PH_ALL_END,
} USB_DMA_IN_PHASE;

typedef enum {
    USB_CTRL_TRANS_PH_IDLE     = 0,
    USB_CTRL_TRANS_PH_SETUP_DATA_STAGE,
    USB_CTRL_TRANS_PH_SETUP_DATA_STAGE_PENDING,
    USB_CTRL_TRANS_PH_SETUP_STS_STAGE,
    USB_CTRL_TRANS_PH_SETUP_STS_STAGE_PENDING,
} USB_CTRL_TRANS_PHASE;

typedef enum {
    USB_DRV_STATE_UNKNOWN  = 0,
    USB_DRV_STATE_OFF,             /* Direct Power Down */
    USB_DRV_STATE_INACTIVE,        /* DPD req. occured but connected(Suspend) */
    USB_DRV_STATE_ACTIVE,          /* Enabled and Disconnected */
} USB_DRV_STATE;

/* 2009.07.17 */
typedef enum {
    USB_POW_ST_DEFAULT = 0,
    USB_POW_ST_POWER_DOWN,
    USB_POW_ST_DIRECT_PD,
    USB_POW_ST_DPD_ENABLED,
    USB_POW_ST_POWER_ON,
    USB_POW_ST_UNDEF
} USB_POW_ST;

typedef enum {
    USB_POW_EVENT_UNDEF = 0,
    USB_POW_EVENT_INIT_ON,
    USB_POW_EVENT_INIT_OFF,
    USB_POW_EVENT_ENABLE,
    USB_POW_EVENT_DISABLE,
    USB_POW_EVENT_PLUGIN,
    USB_POW_EVENT_PLUGOUT,
    USB_POW_EVENT_SUSPEND,
    USB_POW_EVENT_RESUME
} USB_POW_EVENT;

typedef enum {
    USB_DPD_RET_DIRECT_PD  = 0,
    USB_DPD_RET_POWER_DOWN,
    USB_DPD_RET_FAIL,
} USB_DPD_RET_CODE;

typedef enum {
    USB_DEV_ST_UNDEF   = 0,
    USB_DEV_ST_DEFAULT,
    USB_DEV_ST_ADDRESS,
    USB_DEV_ST_CONFIGURED,
} USB_DEV_ST;

typedef enum {
    USB_CTRL_TRANS_TYPE_UNDEF = 0,
    USB_CTRL_TRANS_NO_DATA,
    USB_CTRL_TRANS_WRITE,
    USB_CTRL_TRANS_READ,
} USB_CTRL_TRANS_TYPE;


/* forward declaration. */
struct cxd9225gg_ep;
struct cxd9225gg_req;
struct cxd9225gg_udc;


/** USB___ENDPOINT */
typedef struct usbEndpoint
{
    unsigned char    endpointIndex;               /** Endpoint Index */
    unsigned char    direction;                   /** Direction */
    unsigned short   maxPacketSize;               /** Max Packet Size */

    unsigned long   transferType;                /** Endpoint Transfer Type */


    unsigned long  uEndpointInterruptStatus;    /* EndPoint interrupt status */
    USB_TRANSFER_TYPE ulPipeTransferType;

    /* no timeout-func for semaphore in linux 2.6.23 */
    struct completion hSem4PipeSerialize;

    /* attention: no timeout-function for semaphore in linux 2.6.23
     *
     */
    struct completion hSem4WaitForInterrupt;

    /* EPC DMA END semahpore.
     * no timeout-func for semaphore in linux 2.6.23
     */
    struct completion hSemEpcDmaEndInt;

    /* Bridge DMA END semahpore.
     * no timeout-func for semaphore in linux 2.6.23
     */
    struct completion hSemBridgeDmaEndInt;

    USB_DMA_IN_PHASE   ulDmaInPhase;
    int    bIsAssigned;                /* enable or disable */
    int    bIntInformPending;
    int    bCswCompleteIntFlag;        /* from IN transfer complete to next out transfer req is Active */
    int    bInRequestIntFlag;          /* from IN transfer req to IN transfer complete is Active */
    int    bOutRequestIntFlag;         /* from OUT transfer req to OUT transfer complete is Active */
    int    bBulkOutEnd;                /* 2009.08.05 */
    int    bBulkInEnd;                 /* 2009.08.05 */

    int    op_flag;
} USBENDPOINT, *PUSBENDPOINT;

typedef struct usbTarget
{
    spinlock_t lock;

    int        bulkInNakPending;           /** Bulk In NAK Pending Report */
    int        usbEnableFlg;            /** accept enable/disable flag */
    int        usbPlugFlg;              /** Pulg flag */
    int        bIsPllUnlocked;                 /** PowerDown flag */
    int        bSuspendRequested;          /** Suspend Request Occured */
    int        bVbusIntOccured;            /** VBUS Int Occured */
    int        bResumeRequested;           /** Resume Request Occured */
    int        bResetRequested;            /** Bus Reset Request Occured */
    int        bSemNow;                    /** DMA sem wait */

    USB_POW_ST ulPowState;                 /** Power State */

    USB_CONNECT_STATUS usbConnectStatus;
    USB_DRV_STATE  ulUsbDrvState;       /** USB Func Driver State */

    struct completion sUsbIntSemId;

    /** PlugIn/PlugOut Sem ID
     * since, there is no 'timeout-func' for semaphore in linux 2.6.23.
     */
    struct completion sUsbPlugInOutSemId;

    struct mutex     hSemApiExecuting;

    unsigned short      uInterruptAssign;           /** Interrupt assortment */
    unsigned short      uTransitVersion;            /** Transit Version */

    unsigned long      uUsbIntStatus;              /** USB Interrupt Status */
    unsigned long      uUsbIntEnableSts;           /** USB Interrupt Enable Status */
    unsigned long      uEp0IntStatus;              /** EP0 Status Register */
    unsigned long      uEp1IntStatus;              /** EP1 Status Register */
    unsigned long      uEp2IntStatus;              /** EP2 Status Register */
    USB_DEV_ST             ulDeviceState;           /* default/address/configured */
    int                        bIsCtrlTransFinished;
    int                        bIsSetupRequested;
    int                        bIsCtrlTransOnGoing;        /** for control transfer */
    USB_CTRL_TRANS_PHASE   ulCtrlTransPhase;
    USB_CTRL_TRANS_TYPE    ulCtrlTransType;

    /* Control sync semaphore.
     * no timeout-func for semaphore in linux 2.6.23
     */
    struct completion hSem4SyncCtrlTransfer;

    int                        bEp2IntInformPending;   /* EP2 INT Pending flag */

      /* same structure PS */
    unsigned long      uInterruptStatus;        /* interrupt  status in driver */
                                                /* same bitmap PS bitmap */
    /* relaitionship Handle and EndPoint */
    PUSBENDPOINT     *pEndPointInfoTop;

    /* semaphore to notify VBUS-wait Task . */
    struct completion hSem4VbusWaitTask;

    /* Vbus Wait Task */
    struct task_struct *hThd4VbusWait;

    struct task_struct *hThd_UsbIsr;

    int            bDoPlugIn;                  /* reserve to call PlugIn */
    int            bDoPlugOut;                 /* reserve to call PlugOut */

    unsigned long      ulUsbIntEnBackup;           /* free Mask */
    unsigned long      ulAhbbIntEnBackup;          /* free Mask */

    int            bIsBusError;

    unsigned long      uAhbBIntStatus;             /** AHBBINT Register */
    unsigned long      uAhbBIntEnableSts;          /** AHBBINTEN Register */

    unsigned long      uEp0IntEnableSts;           /** EP0 Interrupt Enable Register */
    unsigned long      uEp1IntEnableSts;           /** EP1 Interrupt Enable Register */
    unsigned long      uEp2IntEnableSts;           /** EP2 Interrupt Enable Register */

    int        bDmaStop;                   /* DMA Stop flag */

    int        bBulkOutResult;             /** BulkOut DMA Result */
    unsigned long      uBulkOutRemainSize;         /** BulkOut DMA Remain Size */
    int        bBulkInResult;              /** BulkIn DMA Result */
    unsigned long      uBulkInRemainSize;          /** BulkIn DMA Remain Size */

	/** for Mac format error */
    int        bDmaSwStopFlag;             /** DMA Software Stop Flag */
    unsigned long      uDmaBfrRemainSize;          /** DMA before Remain Size */

    /** for SetConfiguration */
    unsigned long      ulConfigValue[USB_CONF_INDEX_SIZE];
    unsigned long      ulCurrentConfigValue;
    unsigned long      ulGetDescriptorValue;

    struct cxd9225gg_udc *udc_dev;
    atomic_t thread_exit;

    unsigned long      ulDebug[256];
    unsigned long      ulDebugCnt;
} USBTARGET, *PUSBTARGET;


extern int usbFncAddressSet(int dev_addr);
extern int usbFncSignalResume(struct cxd9225gg_udc *udc_dev);
extern int usbFncDeviceFeatureSet(int feature, int index);
extern int usbFncDeviceFeatureClear(int feature);
extern int usbFncEndpointAssign(struct cxd9225gg_udc *udc_dev,
                                           int ep_num);
extern int usbFncEndpointRelease(struct cxd9225gg_udc *udc_dev,
                                            int ep_num);
extern int usbFncEndpointStateSet(int endpointIndex, int stall);
int usbFncEndpointStatusGet(int endpointIndex, int *p_stall);
extern int usbFncIsBufferEmpty(int endpointIndex, int *p_buf_empty);
extern int usbFncCopyDataFromEpBuf(struct cxd9225gg_ep  *udc_ep,
                                              struct cxd9225gg_req *udc_req);
extern int usbFncCopyDataToEpBuf(struct cxd9225gg_ep  *udc_ep,
                                            struct cxd9225gg_req *udc_req);
extern int usbFncAttach(struct cxd9225gg_udc *udc_dev);
extern int usbFncDetach(struct cxd9225gg_udc *udc_dev);
extern int usbFncEnable(PUSBTARGET pTarget);
extern int usbFncDisable(PUSBTARGET pTarget);
extern int usbFncInterruptStatusGet(PUSBTARGET pTarget);
extern int usbFncInterruptStatusClear(PUSBTARGET pTarget);
extern int usbFncEndpointIntStatusGet(PUSBTARGET pTarget,
                                                        int ep_num);
extern int usbFncEndpointIntStatusClear(PUSBTARGET pTarget,
                                                          int ep_num);
extern int usbFncHandleDisconnectInterrupt(PUSBTARGET pTarget);
extern int usbFncHandleResumeInterrupt(PUSBTARGET pTarget);
extern int usbFncHandleResetInterrupt(PUSBTARGET pTarget);
extern int usbFncHandleSuspendInterrupt(PUSBTARGET pTarget);
extern int usbPowerStateControl(USB_POW_ST *pulPowState,
                                           USB_POW_EVENT ulPowEvent,
                                           int *pbPowerDownFlag);
extern int usbPowerResume(void);
extern int usbDirectPowerDown(void);
extern int usbReturnfromDirectPowerDown(void);
extern int usbConnectionNotify(int plug_flag);

/** Self Diagnosis */
extern int usbDiagReg(void);
extern int usbDiagPower(void);

/** Interrupt */
extern irqreturn_t usbBridgeIsr(int irq, void *dev_id);
extern irqreturn_t usbEpcIsr(int irq, void *dev_id);
extern int usbIsrExec(void *data);
extern int usbVbusWaitTask(void *data);

/** InitExit */
extern int usbIntcSetting(void);
extern int usbPlugIn(void);
extern int usbPlugOut(USB_POW_EVENT ulEvent);


int usbPioDataRecieve(struct cxd9225gg_ep  *udc_ep,
                                 struct cxd9225gg_req *udc_req);
int usbPioDataSend(struct cxd9225gg_ep  *udc_ep,
                             struct cxd9225gg_req *udc_req);
int usbSetupPioDataRecieve(struct cxd9225gg_ep  *udc_ep,
                                      struct cxd9225gg_req *udc_req);


#endif    /* __INCusbEvalLibh */
