/*
 * Copyright (C) 2012 Sony Corporation
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/pci.h>

#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>

#include <linux/errno.h>
#include <linux/init.h>

#include <linux/usb.h>

#include <linux/debugfs.h>

#include "../core/hcd.h"


#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>


extern struct usb_hcd		*usb_hcd_global;

#define EHCI_IAA_MSECS		10		/* arbitrary */
#define EHCI_IO_JIFFIES		(HZ/10)		/* io watchdog > irq_thresh */
#define EHCI_ASYNC_JIFFIES	(HZ/20)		/* async idle timeout */
#define EHCI_SHRINK_FRAMES	5		/* async qh unlink delay */

#include "ehci.h"
int usbHostTestMode(u32 port, u32 mode);
/**
 * @fn void  usbHostSetTestModeInitialize(void)
 *
 * USB Host Test Mode Initialize Function
 *
 * @param[in]   None
 *
 * @retval      0 or -1
 *
 * @note
 */
static int usbHostSetTestModeInitialize(void) {
	u32 ret= -1; /* return value */
	volatile u32 regData = 0x00000000; /* register buffer */
	s32 index = 0; /* index */
	s32 chkCount = 0; /* check counter */
	u32 portData = 0x00000000; /* Enable Port Number */
	
	struct ehci_hcd	*ehci = hcd_to_ehci(usb_hcd_global);

	/* USBCMD : Asynch Schedule Enable, Periodic Schedule Enable set to Disable */
	regData = ehci_readl(ehci, &ehci->regs->command);
	regData	= (regData & ~(CMD_PSE | CMD_ASE));
	ehci_writel(ehci, regData, &ehci->regs->command);

	/* Get N_POTS(HCSPARAMS : Number of Ports) */
	portData = HCS_N_PORTS(ehci->hcs_params);
	
	/* Check Port Number */
	if (portData > 0) {
		/* POTSC(1-5) : Suspend set to Suspend */
		for (index = 0; portData > (u32)index; index++) {
 			regData = ehci_readl(ehci, &ehci->regs->port_status[index]);
			regData = (regData | PORT_SUSPEND);
			ehci_writel(ehci, regData, &ehci->regs->port_status[index]);
		}

		/* USBCMD : RS set to Stop */
		regData = ehci_readl(ehci, &ehci->regs->command);
		regData = (regData & ~(CMD_RUN));
		ehci_writel(ehci, regData, &ehci->regs->command);

		/* USBSTS : Waiti until change the HCHalted */
		while (1) {
			regData = ehci_readl(ehci, &ehci->regs->status);
			if (STS_HALT == ( regData & STS_HALT)) {
				dev_dbg(ehci_to_hcd(ehci)->self.controller, "TEST MODE : Initialize Success.\n");
				ret = 0;
				break;
			} else {
				chkCount ++;
				if (25 < (u32)chkCount) {
					dev_dbg(ehci_to_hcd(ehci)->self.controller,
						"TEST MODE : USB_EHCD_USBSTS HCHalted Not Change.\n");
					break;
				}
				//taskDelay(2);
				mdelay(2*16);
			}
		}
	} else {
		dev_dbg(ehci_to_hcd(ehci)->self.controller,
			"TEST MODE : USB_EHCD_HCSPARAMS N_POTS[%d] Change POTSC None.\n",portData);
		
	}
	return (ret);
}

/**
 * @fn void  usbHostSetTestModeRegister(void)
 *
 * USB Host Test Mode Set Registers Function
 *
 * @param[in]   u32 port
 *              u32 mode
 *
 * @retval      0 or -1
 *
 * @note
 */
static int usbHostSetTestModeRegister(u32 port, u32 mode) {
	u32 ret= -1; /* return value */
	u32 regData = 0x00000000; /* register buffer */
	s32 index = 0; /* index */
	u32 portData = 0x00000000; /* Enable Port Number */
	
	struct ehci_hcd	*ehci = hcd_to_ehci(usb_hcd_global);
	
	switch (mode<<16) {
	case PORT_TEST_OFF:
	case PORT_TEST_J_STATE:
	case PORT_TEST_K_STATE:
	case PORT_TEST_SE0_NAK:
	case PORT_TEST_PACKET:
	case PORT_TEST_FORCE_ENABLE:

		/* Get N_POTS(HCSPARAMS : Number of Ports) */
		portData = HCS_N_PORTS(ehci->hcs_params);

		/* Check Port Number */
		if (portData > 0) {
			switch ((u8)port) {
			case 0: /* All Port specified */

				/* POTSC(1-5) : Port Test Control set to specified mode */
				for (index = 0; portData > (u32)index; index++) {
					regData = ehci_readl(ehci, &ehci->regs->port_status[index]);
					regData = (regData & ~PORT_TEST_MASK);
					regData = (regData | (mode << 16));
					ehci_writel(ehci, regData, &ehci->regs->port_status[index]);
					dev_dbg(ehci_to_hcd(ehci)->self.controller,
						"TEST MODE : PORTSC%d USB_EHCD_PORTSC_PORT_TEST_CONTROL[%d] Set.\n",(index + 1), mode);
				}

				ret = 0;
				break;

			default: /* Else */

				/* Specified POTSC : Port Test Control set to specified mode */
				if (port <= (u8)portData) {
					regData = ehci_readl(ehci, &ehci->regs->port_status[port - 1]);
					regData = (regData & ~PORT_TEST_MASK);
					regData = (regData | (mode << 16));
					ehci_writel(ehci, regData, &ehci->regs->port_status[port - 1]);
					dev_dbg(ehci_to_hcd(ehci)->self.controller,
						"TEST MODE : PORTSC%d USB_EHCD_PORTSC_PORT_TEST_CONTROL[%d] Set.\n",port, mode);
					ret = 0;
				} else {
					dev_dbg(ehci_to_hcd(ehci)->self.controller,
						"TEST MODE : N_POTS[%d] Illegal Port[%d] Error.\n",portData, port);
				}
				break;
			}
		} else {
			dev_dbg(ehci_to_hcd(ehci)->self.controller,
						"TEST MODE : USB_EHCD_HCSPARAMS N_POTS[%d] Change POTSC None.\n",portData);
		}
		break;

	default:

		dev_dbg(ehci_to_hcd(ehci)->self.controller,
						"TEST MODE : Illegal Mode[%d] Error.\n", mode);
		break;
	}
	return (ret);
}


/**
 * @fn      STATUS usbHostSelfDiagnose(UINT32 SDFlg)
 * @brief   USB HOST Test Mode Shift Function
 *
 * @retval  OK      0
 *          ERROR   -1
 *
 * @note
 */
int usbHostTestMode(u32 port, u32 mode)
{
    int      ret = -1;    /* return value */


    /* Initialize for Test Mode */
    if( 0 == (ret = usbHostSetTestModeInitialize()))
    {
        /* Set register for Test Mode */
        ret = usbHostSetTestModeRegister(port, mode);
    }
    return(ret);
}

