/*
 * Copyright (C) 2012 Sony Corporation
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/proc_fs.h>
#include <linux/ktime.h>
#include <asm/uaccess.h>

#define UNKNOWN_FUNCNAME "UNKNOWN_INITCALL"

unsigned long long s_bootup_timestamp = 0;

#define TMP_BUF_LEN 0x10000 /* 64 KB */
unsigned s_buf_len = TMP_BUF_LEN;
unsigned s_buf_stride = 0x80;
char s_temp_buf[TMP_BUF_LEN] = {0};
static char* s_temp_ptr = &s_temp_buf[0];

static const char* log_elapsed_time(ktime_t *tmstart, ktime_t *tmend, initcall_t *call)
{
	if ((s_temp_ptr - &s_temp_buf[0]) < s_buf_len) {
		unsigned long address = (unsigned long)__builtin_extract_return_addr((void *)((unsigned long) *call));
		char *modname;
		const char *name;
		unsigned long offset, size;
		char namebuf[KSYM_NAME_LEN];
		ktime_t delta;
		int i = 0;
		
		name = kallsyms_lookup(address, &size, &offset, &modname, namebuf);
		name = name ? name : UNKNOWN_FUNCNAME;

		delta = ktime_sub(*tmend, *tmstart);
		*((s32*)(s_temp_ptr+4*i++)) = tmstart->tv.sec;
		*((s32*)(s_temp_ptr+4*i++)) = tmstart->tv.nsec;
		*((s32*)(s_temp_ptr+4*i++)) = delta.tv.sec;
		*((s32*)(s_temp_ptr+4*i++)) = delta.tv.nsec;
		sprintf(s_temp_ptr+4*i, "%s", name);
		s_temp_ptr[s_buf_stride-1] = '\0'; //truncate

		s_temp_ptr += s_buf_stride;

		return name;
	}
	return 0;
}

// return nbytes to tell the caller there may be more data.
// return 0 to tell the caller that it reaches the end.
ssize_t test_proc_read(struct file *file, char __user *buf, size_t nbytes, loff_t *ppos)
{
	char * psrc = &s_temp_buf[0] + (*ppos)*s_buf_stride;
	char * const bufpage = (char*)__get_free_page(GFP_KERNEL);
	char * pdst = bufpage;
	int i = 0;
	int n = 0;
	if (0==*ppos) {
		unsigned long long _nsec = s_bootup_timestamp;
		unsigned long long _under_sec = do_div(_nsec, 1000000000);
		n = snprintf(pdst, (PAGE_SIZE-(pdst-bufpage)), "Kernel Start At [%llu.%09llu] Second After Power On.\n", _nsec, _under_sec);
		pdst += n;
		n = snprintf(pdst, (PAGE_SIZE-(pdst-bufpage)), "NO. : START TIME      : ELAPSED TIME    : FUNCTION NAME\n");
		pdst += n; /* the last char is NUL */
	}
	for (i=*ppos; (psrc-&s_temp_buf[0])<s_buf_len; psrc+=s_buf_stride, i++) {
		int x = 0;
		ktime_t start, elaps;
		start.tv.sec 	= *((int*)(psrc+4*x++));
		start.tv.nsec 	= *((int*)(psrc+4*x++));
		elaps.tv.sec 	= *((int*)(psrc+4*x++));
		elaps.tv.nsec 	= *((int*)(psrc+4*x++));
		if (! *(psrc+4*x)) {
			/* no more logs */
			break;
		}
		n = snprintf(pdst, (PAGE_SIZE-(pdst-bufpage)), "%3u : %5d.%09d : %5d.%09d : %s()\n",
			i+1, /* print index from 1, not 0 */
			start.tv.sec, start.tv.nsec, elaps.tv.sec, elaps.tv.nsec, psrc+4*x);
		if (n >= (PAGE_SIZE-(pdst-bufpage))) {
			/* bufpage is not enough, string is truncated. */
			break;
		}
		pdst += n; /* the last char is NUL */
	}

	n = pdst-bufpage;
	*ppos += i;

	if (copy_to_user(buf, bufpage, n)) {return -1;}
	free_page((unsigned long)bufpage);

	return n;
}

//return -1 to tell caller that we do not support write operation.
ssize_t test_proc_write(struct file *file, const char __user *buf, size_t nbytes, loff_t *ppos)
{
	return -1;
}

static const struct file_operations test_ops = {
	.read		= test_proc_read,
	.write		= test_proc_write,
};

static int test_proc_init(void)
{
	struct proc_dir_entry *e;
	// give a NULL as parent, to create entry in /proc/
	e = create_proc_entry("bootup", S_IFREG|S_IRUGO|S_IWUSR, NULL);
	e->proc_fops = &test_ops;

	return 0;
}
late_initcall(test_proc_init);